from nose.tools import assert_equal
from nose import SkipTest

class TestPypeObject:
    def TestRDFXML(self):
        # pype_object = PypeObject(URL, **attributes)
        # assert_equal(expected, pype_object.RDFXML())
        raise SkipTest # TODO: implement your test here

    def test___init__(self):
        # pype_object = PypeObject(URL, **attributes)
        raise SkipTest # TODO: implement your test here

class TestRunShellCmd:
    def test_run_shell_cmd(self):
        # assert_equal(expected, runShellCmd(args, **kwargs))
        raise SkipTest # TODO: implement your test here

class TestRunSgeSyncJob:
    def test_run_sge_sync_job(self):
        # assert_equal(expected, runSgeSyncJob(args))
        raise SkipTest # TODO: implement your test here

