daligner -v -k18 -h70 -t14 -H7000 -e0.75 raw_reads.92 raw_reads.1 raw_reads.2 raw_reads.3 raw_reads.4 raw_reads.5 raw_reads.6 raw_reads.7 raw_reads.8 raw_reads.9 raw_reads.10 raw_reads.11 raw_reads.12 raw_reads.13 raw_reads.14 raw_reads.15 raw_reads.16 raw_reads.17 raw_reads.18 raw_reads.19 raw_reads.20 raw_reads.21 raw_reads.22 raw_reads.23
LAcheck -v raw_reads *.las
LAsort -v raw_reads.1.raw_reads.92.C0 raw_reads.1.raw_reads.92.N0 raw_reads.1.raw_reads.92.C1 raw_reads.1.raw_reads.92.N1 raw_reads.1.raw_reads.92.C2 raw_reads.1.raw_reads.92.N2 raw_reads.1.raw_reads.92.C3 raw_reads.1.raw_reads.92.N3 && LAmerge -v L1.1.92 raw_reads.1.raw_reads.92.C0.S raw_reads.1.raw_reads.92.N0.S raw_reads.1.raw_reads.92.C1.S raw_reads.1.raw_reads.92.N1.S raw_reads.1.raw_reads.92.C2.S raw_reads.1.raw_reads.92.N2.S raw_reads.1.raw_reads.92.C3.S raw_reads.1.raw_reads.92.N3.S
LAsort -v raw_reads.2.raw_reads.92.C0 raw_reads.2.raw_reads.92.N0 raw_reads.2.raw_reads.92.C1 raw_reads.2.raw_reads.92.N1 raw_reads.2.raw_reads.92.C2 raw_reads.2.raw_reads.92.N2 raw_reads.2.raw_reads.92.C3 raw_reads.2.raw_reads.92.N3 && LAmerge -v L1.2.92 raw_reads.2.raw_reads.92.C0.S raw_reads.2.raw_reads.92.N0.S raw_reads.2.raw_reads.92.C1.S raw_reads.2.raw_reads.92.N1.S raw_reads.2.raw_reads.92.C2.S raw_reads.2.raw_reads.92.N2.S raw_reads.2.raw_reads.92.C3.S raw_reads.2.raw_reads.92.N3.S
LAsort -v raw_reads.3.raw_reads.92.C0 raw_reads.3.raw_reads.92.N0 raw_reads.3.raw_reads.92.C1 raw_reads.3.raw_reads.92.N1 raw_reads.3.raw_reads.92.C2 raw_reads.3.raw_reads.92.N2 raw_reads.3.raw_reads.92.C3 raw_reads.3.raw_reads.92.N3 && LAmerge -v L1.3.92 raw_reads.3.raw_reads.92.C0.S raw_reads.3.raw_reads.92.N0.S raw_reads.3.raw_reads.92.C1.S raw_reads.3.raw_reads.92.N1.S raw_reads.3.raw_reads.92.C2.S raw_reads.3.raw_reads.92.N2.S raw_reads.3.raw_reads.92.C3.S raw_reads.3.raw_reads.92.N3.S
LAsort -v raw_reads.4.raw_reads.92.C0 raw_reads.4.raw_reads.92.N0 raw_reads.4.raw_reads.92.C1 raw_reads.4.raw_reads.92.N1 raw_reads.4.raw_reads.92.C2 raw_reads.4.raw_reads.92.N2 raw_reads.4.raw_reads.92.C3 raw_reads.4.raw_reads.92.N3 && LAmerge -v L1.4.92 raw_reads.4.raw_reads.92.C0.S raw_reads.4.raw_reads.92.N0.S raw_reads.4.raw_reads.92.C1.S raw_reads.4.raw_reads.92.N1.S raw_reads.4.raw_reads.92.C2.S raw_reads.4.raw_reads.92.N2.S raw_reads.4.raw_reads.92.C3.S raw_reads.4.raw_reads.92.N3.S
LAsort -v raw_reads.5.raw_reads.92.C0 raw_reads.5.raw_reads.92.N0 raw_reads.5.raw_reads.92.C1 raw_reads.5.raw_reads.92.N1 raw_reads.5.raw_reads.92.C2 raw_reads.5.raw_reads.92.N2 raw_reads.5.raw_reads.92.C3 raw_reads.5.raw_reads.92.N3 && LAmerge -v L1.5.92 raw_reads.5.raw_reads.92.C0.S raw_reads.5.raw_reads.92.N0.S raw_reads.5.raw_reads.92.C1.S raw_reads.5.raw_reads.92.N1.S raw_reads.5.raw_reads.92.C2.S raw_reads.5.raw_reads.92.N2.S raw_reads.5.raw_reads.92.C3.S raw_reads.5.raw_reads.92.N3.S
LAsort -v raw_reads.6.raw_reads.92.C0 raw_reads.6.raw_reads.92.N0 raw_reads.6.raw_reads.92.C1 raw_reads.6.raw_reads.92.N1 raw_reads.6.raw_reads.92.C2 raw_reads.6.raw_reads.92.N2 raw_reads.6.raw_reads.92.C3 raw_reads.6.raw_reads.92.N3 && LAmerge -v L1.6.92 raw_reads.6.raw_reads.92.C0.S raw_reads.6.raw_reads.92.N0.S raw_reads.6.raw_reads.92.C1.S raw_reads.6.raw_reads.92.N1.S raw_reads.6.raw_reads.92.C2.S raw_reads.6.raw_reads.92.N2.S raw_reads.6.raw_reads.92.C3.S raw_reads.6.raw_reads.92.N3.S
LAsort -v raw_reads.7.raw_reads.92.C0 raw_reads.7.raw_reads.92.N0 raw_reads.7.raw_reads.92.C1 raw_reads.7.raw_reads.92.N1 raw_reads.7.raw_reads.92.C2 raw_reads.7.raw_reads.92.N2 raw_reads.7.raw_reads.92.C3 raw_reads.7.raw_reads.92.N3 && LAmerge -v L1.7.92 raw_reads.7.raw_reads.92.C0.S raw_reads.7.raw_reads.92.N0.S raw_reads.7.raw_reads.92.C1.S raw_reads.7.raw_reads.92.N1.S raw_reads.7.raw_reads.92.C2.S raw_reads.7.raw_reads.92.N2.S raw_reads.7.raw_reads.92.C3.S raw_reads.7.raw_reads.92.N3.S
LAsort -v raw_reads.8.raw_reads.92.C0 raw_reads.8.raw_reads.92.N0 raw_reads.8.raw_reads.92.C1 raw_reads.8.raw_reads.92.N1 raw_reads.8.raw_reads.92.C2 raw_reads.8.raw_reads.92.N2 raw_reads.8.raw_reads.92.C3 raw_reads.8.raw_reads.92.N3 && LAmerge -v L1.8.92 raw_reads.8.raw_reads.92.C0.S raw_reads.8.raw_reads.92.N0.S raw_reads.8.raw_reads.92.C1.S raw_reads.8.raw_reads.92.N1.S raw_reads.8.raw_reads.92.C2.S raw_reads.8.raw_reads.92.N2.S raw_reads.8.raw_reads.92.C3.S raw_reads.8.raw_reads.92.N3.S
LAsort -v raw_reads.9.raw_reads.92.C0 raw_reads.9.raw_reads.92.N0 raw_reads.9.raw_reads.92.C1 raw_reads.9.raw_reads.92.N1 raw_reads.9.raw_reads.92.C2 raw_reads.9.raw_reads.92.N2 raw_reads.9.raw_reads.92.C3 raw_reads.9.raw_reads.92.N3 && LAmerge -v L1.9.92 raw_reads.9.raw_reads.92.C0.S raw_reads.9.raw_reads.92.N0.S raw_reads.9.raw_reads.92.C1.S raw_reads.9.raw_reads.92.N1.S raw_reads.9.raw_reads.92.C2.S raw_reads.9.raw_reads.92.N2.S raw_reads.9.raw_reads.92.C3.S raw_reads.9.raw_reads.92.N3.S
LAsort -v raw_reads.10.raw_reads.92.C0 raw_reads.10.raw_reads.92.N0 raw_reads.10.raw_reads.92.C1 raw_reads.10.raw_reads.92.N1 raw_reads.10.raw_reads.92.C2 raw_reads.10.raw_reads.92.N2 raw_reads.10.raw_reads.92.C3 raw_reads.10.raw_reads.92.N3 && LAmerge -v L1.10.92 raw_reads.10.raw_reads.92.C0.S raw_reads.10.raw_reads.92.N0.S raw_reads.10.raw_reads.92.C1.S raw_reads.10.raw_reads.92.N1.S raw_reads.10.raw_reads.92.C2.S raw_reads.10.raw_reads.92.N2.S raw_reads.10.raw_reads.92.C3.S raw_reads.10.raw_reads.92.N3.S
LAsort -v raw_reads.11.raw_reads.92.C0 raw_reads.11.raw_reads.92.N0 raw_reads.11.raw_reads.92.C1 raw_reads.11.raw_reads.92.N1 raw_reads.11.raw_reads.92.C2 raw_reads.11.raw_reads.92.N2 raw_reads.11.raw_reads.92.C3 raw_reads.11.raw_reads.92.N3 && LAmerge -v L1.11.92 raw_reads.11.raw_reads.92.C0.S raw_reads.11.raw_reads.92.N0.S raw_reads.11.raw_reads.92.C1.S raw_reads.11.raw_reads.92.N1.S raw_reads.11.raw_reads.92.C2.S raw_reads.11.raw_reads.92.N2.S raw_reads.11.raw_reads.92.C3.S raw_reads.11.raw_reads.92.N3.S
LAsort -v raw_reads.12.raw_reads.92.C0 raw_reads.12.raw_reads.92.N0 raw_reads.12.raw_reads.92.C1 raw_reads.12.raw_reads.92.N1 raw_reads.12.raw_reads.92.C2 raw_reads.12.raw_reads.92.N2 raw_reads.12.raw_reads.92.C3 raw_reads.12.raw_reads.92.N3 && LAmerge -v L1.12.92 raw_reads.12.raw_reads.92.C0.S raw_reads.12.raw_reads.92.N0.S raw_reads.12.raw_reads.92.C1.S raw_reads.12.raw_reads.92.N1.S raw_reads.12.raw_reads.92.C2.S raw_reads.12.raw_reads.92.N2.S raw_reads.12.raw_reads.92.C3.S raw_reads.12.raw_reads.92.N3.S
LAsort -v raw_reads.13.raw_reads.92.C0 raw_reads.13.raw_reads.92.N0 raw_reads.13.raw_reads.92.C1 raw_reads.13.raw_reads.92.N1 raw_reads.13.raw_reads.92.C2 raw_reads.13.raw_reads.92.N2 raw_reads.13.raw_reads.92.C3 raw_reads.13.raw_reads.92.N3 && LAmerge -v L1.13.92 raw_reads.13.raw_reads.92.C0.S raw_reads.13.raw_reads.92.N0.S raw_reads.13.raw_reads.92.C1.S raw_reads.13.raw_reads.92.N1.S raw_reads.13.raw_reads.92.C2.S raw_reads.13.raw_reads.92.N2.S raw_reads.13.raw_reads.92.C3.S raw_reads.13.raw_reads.92.N3.S
LAsort -v raw_reads.14.raw_reads.92.C0 raw_reads.14.raw_reads.92.N0 raw_reads.14.raw_reads.92.C1 raw_reads.14.raw_reads.92.N1 raw_reads.14.raw_reads.92.C2 raw_reads.14.raw_reads.92.N2 raw_reads.14.raw_reads.92.C3 raw_reads.14.raw_reads.92.N3 && LAmerge -v L1.14.92 raw_reads.14.raw_reads.92.C0.S raw_reads.14.raw_reads.92.N0.S raw_reads.14.raw_reads.92.C1.S raw_reads.14.raw_reads.92.N1.S raw_reads.14.raw_reads.92.C2.S raw_reads.14.raw_reads.92.N2.S raw_reads.14.raw_reads.92.C3.S raw_reads.14.raw_reads.92.N3.S
LAsort -v raw_reads.15.raw_reads.92.C0 raw_reads.15.raw_reads.92.N0 raw_reads.15.raw_reads.92.C1 raw_reads.15.raw_reads.92.N1 raw_reads.15.raw_reads.92.C2 raw_reads.15.raw_reads.92.N2 raw_reads.15.raw_reads.92.C3 raw_reads.15.raw_reads.92.N3 && LAmerge -v L1.15.92 raw_reads.15.raw_reads.92.C0.S raw_reads.15.raw_reads.92.N0.S raw_reads.15.raw_reads.92.C1.S raw_reads.15.raw_reads.92.N1.S raw_reads.15.raw_reads.92.C2.S raw_reads.15.raw_reads.92.N2.S raw_reads.15.raw_reads.92.C3.S raw_reads.15.raw_reads.92.N3.S
LAsort -v raw_reads.16.raw_reads.92.C0 raw_reads.16.raw_reads.92.N0 raw_reads.16.raw_reads.92.C1 raw_reads.16.raw_reads.92.N1 raw_reads.16.raw_reads.92.C2 raw_reads.16.raw_reads.92.N2 raw_reads.16.raw_reads.92.C3 raw_reads.16.raw_reads.92.N3 && LAmerge -v L1.16.92 raw_reads.16.raw_reads.92.C0.S raw_reads.16.raw_reads.92.N0.S raw_reads.16.raw_reads.92.C1.S raw_reads.16.raw_reads.92.N1.S raw_reads.16.raw_reads.92.C2.S raw_reads.16.raw_reads.92.N2.S raw_reads.16.raw_reads.92.C3.S raw_reads.16.raw_reads.92.N3.S
LAsort -v raw_reads.17.raw_reads.92.C0 raw_reads.17.raw_reads.92.N0 raw_reads.17.raw_reads.92.C1 raw_reads.17.raw_reads.92.N1 raw_reads.17.raw_reads.92.C2 raw_reads.17.raw_reads.92.N2 raw_reads.17.raw_reads.92.C3 raw_reads.17.raw_reads.92.N3 && LAmerge -v L1.17.92 raw_reads.17.raw_reads.92.C0.S raw_reads.17.raw_reads.92.N0.S raw_reads.17.raw_reads.92.C1.S raw_reads.17.raw_reads.92.N1.S raw_reads.17.raw_reads.92.C2.S raw_reads.17.raw_reads.92.N2.S raw_reads.17.raw_reads.92.C3.S raw_reads.17.raw_reads.92.N3.S
LAsort -v raw_reads.18.raw_reads.92.C0 raw_reads.18.raw_reads.92.N0 raw_reads.18.raw_reads.92.C1 raw_reads.18.raw_reads.92.N1 raw_reads.18.raw_reads.92.C2 raw_reads.18.raw_reads.92.N2 raw_reads.18.raw_reads.92.C3 raw_reads.18.raw_reads.92.N3 && LAmerge -v L1.18.92 raw_reads.18.raw_reads.92.C0.S raw_reads.18.raw_reads.92.N0.S raw_reads.18.raw_reads.92.C1.S raw_reads.18.raw_reads.92.N1.S raw_reads.18.raw_reads.92.C2.S raw_reads.18.raw_reads.92.N2.S raw_reads.18.raw_reads.92.C3.S raw_reads.18.raw_reads.92.N3.S
LAsort -v raw_reads.19.raw_reads.92.C0 raw_reads.19.raw_reads.92.N0 raw_reads.19.raw_reads.92.C1 raw_reads.19.raw_reads.92.N1 raw_reads.19.raw_reads.92.C2 raw_reads.19.raw_reads.92.N2 raw_reads.19.raw_reads.92.C3 raw_reads.19.raw_reads.92.N3 && LAmerge -v L1.19.92 raw_reads.19.raw_reads.92.C0.S raw_reads.19.raw_reads.92.N0.S raw_reads.19.raw_reads.92.C1.S raw_reads.19.raw_reads.92.N1.S raw_reads.19.raw_reads.92.C2.S raw_reads.19.raw_reads.92.N2.S raw_reads.19.raw_reads.92.C3.S raw_reads.19.raw_reads.92.N3.S
LAsort -v raw_reads.20.raw_reads.92.C0 raw_reads.20.raw_reads.92.N0 raw_reads.20.raw_reads.92.C1 raw_reads.20.raw_reads.92.N1 raw_reads.20.raw_reads.92.C2 raw_reads.20.raw_reads.92.N2 raw_reads.20.raw_reads.92.C3 raw_reads.20.raw_reads.92.N3 && LAmerge -v L1.20.92 raw_reads.20.raw_reads.92.C0.S raw_reads.20.raw_reads.92.N0.S raw_reads.20.raw_reads.92.C1.S raw_reads.20.raw_reads.92.N1.S raw_reads.20.raw_reads.92.C2.S raw_reads.20.raw_reads.92.N2.S raw_reads.20.raw_reads.92.C3.S raw_reads.20.raw_reads.92.N3.S
LAsort -v raw_reads.21.raw_reads.92.C0 raw_reads.21.raw_reads.92.N0 raw_reads.21.raw_reads.92.C1 raw_reads.21.raw_reads.92.N1 raw_reads.21.raw_reads.92.C2 raw_reads.21.raw_reads.92.N2 raw_reads.21.raw_reads.92.C3 raw_reads.21.raw_reads.92.N3 && LAmerge -v L1.21.92 raw_reads.21.raw_reads.92.C0.S raw_reads.21.raw_reads.92.N0.S raw_reads.21.raw_reads.92.C1.S raw_reads.21.raw_reads.92.N1.S raw_reads.21.raw_reads.92.C2.S raw_reads.21.raw_reads.92.N2.S raw_reads.21.raw_reads.92.C3.S raw_reads.21.raw_reads.92.N3.S
LAsort -v raw_reads.22.raw_reads.92.C0 raw_reads.22.raw_reads.92.N0 raw_reads.22.raw_reads.92.C1 raw_reads.22.raw_reads.92.N1 raw_reads.22.raw_reads.92.C2 raw_reads.22.raw_reads.92.N2 raw_reads.22.raw_reads.92.C3 raw_reads.22.raw_reads.92.N3 && LAmerge -v L1.22.92 raw_reads.22.raw_reads.92.C0.S raw_reads.22.raw_reads.92.N0.S raw_reads.22.raw_reads.92.C1.S raw_reads.22.raw_reads.92.N1.S raw_reads.22.raw_reads.92.C2.S raw_reads.22.raw_reads.92.N2.S raw_reads.22.raw_reads.92.C3.S raw_reads.22.raw_reads.92.N3.S
LAsort -v raw_reads.23.raw_reads.92.C0 raw_reads.23.raw_reads.92.N0 raw_reads.23.raw_reads.92.C1 raw_reads.23.raw_reads.92.N1 raw_reads.23.raw_reads.92.C2 raw_reads.23.raw_reads.92.N2 raw_reads.23.raw_reads.92.C3 raw_reads.23.raw_reads.92.N3 && LAmerge -v L1.23.92 raw_reads.23.raw_reads.92.C0.S raw_reads.23.raw_reads.92.N0.S raw_reads.23.raw_reads.92.C1.S raw_reads.23.raw_reads.92.N1.S raw_reads.23.raw_reads.92.C2.S raw_reads.23.raw_reads.92.N2.S raw_reads.23.raw_reads.92.C3.S raw_reads.23.raw_reads.92.N3.S
LAsort -v raw_reads.92.raw_reads.1.C0 raw_reads.92.raw_reads.1.N0 raw_reads.92.raw_reads.1.C1 raw_reads.92.raw_reads.1.N1 raw_reads.92.raw_reads.1.C2 raw_reads.92.raw_reads.1.N2 raw_reads.92.raw_reads.1.C3 raw_reads.92.raw_reads.1.N3 && LAmerge -v L1.92.1 raw_reads.92.raw_reads.1.C0.S raw_reads.92.raw_reads.1.N0.S raw_reads.92.raw_reads.1.C1.S raw_reads.92.raw_reads.1.N1.S raw_reads.92.raw_reads.1.C2.S raw_reads.92.raw_reads.1.N2.S raw_reads.92.raw_reads.1.C3.S raw_reads.92.raw_reads.1.N3.S
LAsort -v raw_reads.92.raw_reads.2.C0 raw_reads.92.raw_reads.2.N0 raw_reads.92.raw_reads.2.C1 raw_reads.92.raw_reads.2.N1 raw_reads.92.raw_reads.2.C2 raw_reads.92.raw_reads.2.N2 raw_reads.92.raw_reads.2.C3 raw_reads.92.raw_reads.2.N3 && LAmerge -v L1.92.2 raw_reads.92.raw_reads.2.C0.S raw_reads.92.raw_reads.2.N0.S raw_reads.92.raw_reads.2.C1.S raw_reads.92.raw_reads.2.N1.S raw_reads.92.raw_reads.2.C2.S raw_reads.92.raw_reads.2.N2.S raw_reads.92.raw_reads.2.C3.S raw_reads.92.raw_reads.2.N3.S
LAsort -v raw_reads.92.raw_reads.3.C0 raw_reads.92.raw_reads.3.N0 raw_reads.92.raw_reads.3.C1 raw_reads.92.raw_reads.3.N1 raw_reads.92.raw_reads.3.C2 raw_reads.92.raw_reads.3.N2 raw_reads.92.raw_reads.3.C3 raw_reads.92.raw_reads.3.N3 && LAmerge -v L1.92.3 raw_reads.92.raw_reads.3.C0.S raw_reads.92.raw_reads.3.N0.S raw_reads.92.raw_reads.3.C1.S raw_reads.92.raw_reads.3.N1.S raw_reads.92.raw_reads.3.C2.S raw_reads.92.raw_reads.3.N2.S raw_reads.92.raw_reads.3.C3.S raw_reads.92.raw_reads.3.N3.S
LAsort -v raw_reads.92.raw_reads.4.C0 raw_reads.92.raw_reads.4.N0 raw_reads.92.raw_reads.4.C1 raw_reads.92.raw_reads.4.N1 raw_reads.92.raw_reads.4.C2 raw_reads.92.raw_reads.4.N2 raw_reads.92.raw_reads.4.C3 raw_reads.92.raw_reads.4.N3 && LAmerge -v L1.92.4 raw_reads.92.raw_reads.4.C0.S raw_reads.92.raw_reads.4.N0.S raw_reads.92.raw_reads.4.C1.S raw_reads.92.raw_reads.4.N1.S raw_reads.92.raw_reads.4.C2.S raw_reads.92.raw_reads.4.N2.S raw_reads.92.raw_reads.4.C3.S raw_reads.92.raw_reads.4.N3.S
LAsort -v raw_reads.92.raw_reads.5.C0 raw_reads.92.raw_reads.5.N0 raw_reads.92.raw_reads.5.C1 raw_reads.92.raw_reads.5.N1 raw_reads.92.raw_reads.5.C2 raw_reads.92.raw_reads.5.N2 raw_reads.92.raw_reads.5.C3 raw_reads.92.raw_reads.5.N3 && LAmerge -v L1.92.5 raw_reads.92.raw_reads.5.C0.S raw_reads.92.raw_reads.5.N0.S raw_reads.92.raw_reads.5.C1.S raw_reads.92.raw_reads.5.N1.S raw_reads.92.raw_reads.5.C2.S raw_reads.92.raw_reads.5.N2.S raw_reads.92.raw_reads.5.C3.S raw_reads.92.raw_reads.5.N3.S
LAsort -v raw_reads.92.raw_reads.6.C0 raw_reads.92.raw_reads.6.N0 raw_reads.92.raw_reads.6.C1 raw_reads.92.raw_reads.6.N1 raw_reads.92.raw_reads.6.C2 raw_reads.92.raw_reads.6.N2 raw_reads.92.raw_reads.6.C3 raw_reads.92.raw_reads.6.N3 && LAmerge -v L1.92.6 raw_reads.92.raw_reads.6.C0.S raw_reads.92.raw_reads.6.N0.S raw_reads.92.raw_reads.6.C1.S raw_reads.92.raw_reads.6.N1.S raw_reads.92.raw_reads.6.C2.S raw_reads.92.raw_reads.6.N2.S raw_reads.92.raw_reads.6.C3.S raw_reads.92.raw_reads.6.N3.S
LAsort -v raw_reads.92.raw_reads.7.C0 raw_reads.92.raw_reads.7.N0 raw_reads.92.raw_reads.7.C1 raw_reads.92.raw_reads.7.N1 raw_reads.92.raw_reads.7.C2 raw_reads.92.raw_reads.7.N2 raw_reads.92.raw_reads.7.C3 raw_reads.92.raw_reads.7.N3 && LAmerge -v L1.92.7 raw_reads.92.raw_reads.7.C0.S raw_reads.92.raw_reads.7.N0.S raw_reads.92.raw_reads.7.C1.S raw_reads.92.raw_reads.7.N1.S raw_reads.92.raw_reads.7.C2.S raw_reads.92.raw_reads.7.N2.S raw_reads.92.raw_reads.7.C3.S raw_reads.92.raw_reads.7.N3.S
LAsort -v raw_reads.92.raw_reads.8.C0 raw_reads.92.raw_reads.8.N0 raw_reads.92.raw_reads.8.C1 raw_reads.92.raw_reads.8.N1 raw_reads.92.raw_reads.8.C2 raw_reads.92.raw_reads.8.N2 raw_reads.92.raw_reads.8.C3 raw_reads.92.raw_reads.8.N3 && LAmerge -v L1.92.8 raw_reads.92.raw_reads.8.C0.S raw_reads.92.raw_reads.8.N0.S raw_reads.92.raw_reads.8.C1.S raw_reads.92.raw_reads.8.N1.S raw_reads.92.raw_reads.8.C2.S raw_reads.92.raw_reads.8.N2.S raw_reads.92.raw_reads.8.C3.S raw_reads.92.raw_reads.8.N3.S
LAsort -v raw_reads.92.raw_reads.9.C0 raw_reads.92.raw_reads.9.N0 raw_reads.92.raw_reads.9.C1 raw_reads.92.raw_reads.9.N1 raw_reads.92.raw_reads.9.C2 raw_reads.92.raw_reads.9.N2 raw_reads.92.raw_reads.9.C3 raw_reads.92.raw_reads.9.N3 && LAmerge -v L1.92.9 raw_reads.92.raw_reads.9.C0.S raw_reads.92.raw_reads.9.N0.S raw_reads.92.raw_reads.9.C1.S raw_reads.92.raw_reads.9.N1.S raw_reads.92.raw_reads.9.C2.S raw_reads.92.raw_reads.9.N2.S raw_reads.92.raw_reads.9.C3.S raw_reads.92.raw_reads.9.N3.S
LAsort -v raw_reads.92.raw_reads.10.C0 raw_reads.92.raw_reads.10.N0 raw_reads.92.raw_reads.10.C1 raw_reads.92.raw_reads.10.N1 raw_reads.92.raw_reads.10.C2 raw_reads.92.raw_reads.10.N2 raw_reads.92.raw_reads.10.C3 raw_reads.92.raw_reads.10.N3 && LAmerge -v L1.92.10 raw_reads.92.raw_reads.10.C0.S raw_reads.92.raw_reads.10.N0.S raw_reads.92.raw_reads.10.C1.S raw_reads.92.raw_reads.10.N1.S raw_reads.92.raw_reads.10.C2.S raw_reads.92.raw_reads.10.N2.S raw_reads.92.raw_reads.10.C3.S raw_reads.92.raw_reads.10.N3.S
LAsort -v raw_reads.92.raw_reads.11.C0 raw_reads.92.raw_reads.11.N0 raw_reads.92.raw_reads.11.C1 raw_reads.92.raw_reads.11.N1 raw_reads.92.raw_reads.11.C2 raw_reads.92.raw_reads.11.N2 raw_reads.92.raw_reads.11.C3 raw_reads.92.raw_reads.11.N3 && LAmerge -v L1.92.11 raw_reads.92.raw_reads.11.C0.S raw_reads.92.raw_reads.11.N0.S raw_reads.92.raw_reads.11.C1.S raw_reads.92.raw_reads.11.N1.S raw_reads.92.raw_reads.11.C2.S raw_reads.92.raw_reads.11.N2.S raw_reads.92.raw_reads.11.C3.S raw_reads.92.raw_reads.11.N3.S
LAsort -v raw_reads.92.raw_reads.12.C0 raw_reads.92.raw_reads.12.N0 raw_reads.92.raw_reads.12.C1 raw_reads.92.raw_reads.12.N1 raw_reads.92.raw_reads.12.C2 raw_reads.92.raw_reads.12.N2 raw_reads.92.raw_reads.12.C3 raw_reads.92.raw_reads.12.N3 && LAmerge -v L1.92.12 raw_reads.92.raw_reads.12.C0.S raw_reads.92.raw_reads.12.N0.S raw_reads.92.raw_reads.12.C1.S raw_reads.92.raw_reads.12.N1.S raw_reads.92.raw_reads.12.C2.S raw_reads.92.raw_reads.12.N2.S raw_reads.92.raw_reads.12.C3.S raw_reads.92.raw_reads.12.N3.S
LAsort -v raw_reads.92.raw_reads.13.C0 raw_reads.92.raw_reads.13.N0 raw_reads.92.raw_reads.13.C1 raw_reads.92.raw_reads.13.N1 raw_reads.92.raw_reads.13.C2 raw_reads.92.raw_reads.13.N2 raw_reads.92.raw_reads.13.C3 raw_reads.92.raw_reads.13.N3 && LAmerge -v L1.92.13 raw_reads.92.raw_reads.13.C0.S raw_reads.92.raw_reads.13.N0.S raw_reads.92.raw_reads.13.C1.S raw_reads.92.raw_reads.13.N1.S raw_reads.92.raw_reads.13.C2.S raw_reads.92.raw_reads.13.N2.S raw_reads.92.raw_reads.13.C3.S raw_reads.92.raw_reads.13.N3.S
LAsort -v raw_reads.92.raw_reads.14.C0 raw_reads.92.raw_reads.14.N0 raw_reads.92.raw_reads.14.C1 raw_reads.92.raw_reads.14.N1 raw_reads.92.raw_reads.14.C2 raw_reads.92.raw_reads.14.N2 raw_reads.92.raw_reads.14.C3 raw_reads.92.raw_reads.14.N3 && LAmerge -v L1.92.14 raw_reads.92.raw_reads.14.C0.S raw_reads.92.raw_reads.14.N0.S raw_reads.92.raw_reads.14.C1.S raw_reads.92.raw_reads.14.N1.S raw_reads.92.raw_reads.14.C2.S raw_reads.92.raw_reads.14.N2.S raw_reads.92.raw_reads.14.C3.S raw_reads.92.raw_reads.14.N3.S
LAsort -v raw_reads.92.raw_reads.15.C0 raw_reads.92.raw_reads.15.N0 raw_reads.92.raw_reads.15.C1 raw_reads.92.raw_reads.15.N1 raw_reads.92.raw_reads.15.C2 raw_reads.92.raw_reads.15.N2 raw_reads.92.raw_reads.15.C3 raw_reads.92.raw_reads.15.N3 && LAmerge -v L1.92.15 raw_reads.92.raw_reads.15.C0.S raw_reads.92.raw_reads.15.N0.S raw_reads.92.raw_reads.15.C1.S raw_reads.92.raw_reads.15.N1.S raw_reads.92.raw_reads.15.C2.S raw_reads.92.raw_reads.15.N2.S raw_reads.92.raw_reads.15.C3.S raw_reads.92.raw_reads.15.N3.S
LAsort -v raw_reads.92.raw_reads.16.C0 raw_reads.92.raw_reads.16.N0 raw_reads.92.raw_reads.16.C1 raw_reads.92.raw_reads.16.N1 raw_reads.92.raw_reads.16.C2 raw_reads.92.raw_reads.16.N2 raw_reads.92.raw_reads.16.C3 raw_reads.92.raw_reads.16.N3 && LAmerge -v L1.92.16 raw_reads.92.raw_reads.16.C0.S raw_reads.92.raw_reads.16.N0.S raw_reads.92.raw_reads.16.C1.S raw_reads.92.raw_reads.16.N1.S raw_reads.92.raw_reads.16.C2.S raw_reads.92.raw_reads.16.N2.S raw_reads.92.raw_reads.16.C3.S raw_reads.92.raw_reads.16.N3.S
LAsort -v raw_reads.92.raw_reads.17.C0 raw_reads.92.raw_reads.17.N0 raw_reads.92.raw_reads.17.C1 raw_reads.92.raw_reads.17.N1 raw_reads.92.raw_reads.17.C2 raw_reads.92.raw_reads.17.N2 raw_reads.92.raw_reads.17.C3 raw_reads.92.raw_reads.17.N3 && LAmerge -v L1.92.17 raw_reads.92.raw_reads.17.C0.S raw_reads.92.raw_reads.17.N0.S raw_reads.92.raw_reads.17.C1.S raw_reads.92.raw_reads.17.N1.S raw_reads.92.raw_reads.17.C2.S raw_reads.92.raw_reads.17.N2.S raw_reads.92.raw_reads.17.C3.S raw_reads.92.raw_reads.17.N3.S
LAsort -v raw_reads.92.raw_reads.18.C0 raw_reads.92.raw_reads.18.N0 raw_reads.92.raw_reads.18.C1 raw_reads.92.raw_reads.18.N1 raw_reads.92.raw_reads.18.C2 raw_reads.92.raw_reads.18.N2 raw_reads.92.raw_reads.18.C3 raw_reads.92.raw_reads.18.N3 && LAmerge -v L1.92.18 raw_reads.92.raw_reads.18.C0.S raw_reads.92.raw_reads.18.N0.S raw_reads.92.raw_reads.18.C1.S raw_reads.92.raw_reads.18.N1.S raw_reads.92.raw_reads.18.C2.S raw_reads.92.raw_reads.18.N2.S raw_reads.92.raw_reads.18.C3.S raw_reads.92.raw_reads.18.N3.S
LAsort -v raw_reads.92.raw_reads.19.C0 raw_reads.92.raw_reads.19.N0 raw_reads.92.raw_reads.19.C1 raw_reads.92.raw_reads.19.N1 raw_reads.92.raw_reads.19.C2 raw_reads.92.raw_reads.19.N2 raw_reads.92.raw_reads.19.C3 raw_reads.92.raw_reads.19.N3 && LAmerge -v L1.92.19 raw_reads.92.raw_reads.19.C0.S raw_reads.92.raw_reads.19.N0.S raw_reads.92.raw_reads.19.C1.S raw_reads.92.raw_reads.19.N1.S raw_reads.92.raw_reads.19.C2.S raw_reads.92.raw_reads.19.N2.S raw_reads.92.raw_reads.19.C3.S raw_reads.92.raw_reads.19.N3.S
LAsort -v raw_reads.92.raw_reads.20.C0 raw_reads.92.raw_reads.20.N0 raw_reads.92.raw_reads.20.C1 raw_reads.92.raw_reads.20.N1 raw_reads.92.raw_reads.20.C2 raw_reads.92.raw_reads.20.N2 raw_reads.92.raw_reads.20.C3 raw_reads.92.raw_reads.20.N3 && LAmerge -v L1.92.20 raw_reads.92.raw_reads.20.C0.S raw_reads.92.raw_reads.20.N0.S raw_reads.92.raw_reads.20.C1.S raw_reads.92.raw_reads.20.N1.S raw_reads.92.raw_reads.20.C2.S raw_reads.92.raw_reads.20.N2.S raw_reads.92.raw_reads.20.C3.S raw_reads.92.raw_reads.20.N3.S
LAsort -v raw_reads.92.raw_reads.21.C0 raw_reads.92.raw_reads.21.N0 raw_reads.92.raw_reads.21.C1 raw_reads.92.raw_reads.21.N1 raw_reads.92.raw_reads.21.C2 raw_reads.92.raw_reads.21.N2 raw_reads.92.raw_reads.21.C3 raw_reads.92.raw_reads.21.N3 && LAmerge -v L1.92.21 raw_reads.92.raw_reads.21.C0.S raw_reads.92.raw_reads.21.N0.S raw_reads.92.raw_reads.21.C1.S raw_reads.92.raw_reads.21.N1.S raw_reads.92.raw_reads.21.C2.S raw_reads.92.raw_reads.21.N2.S raw_reads.92.raw_reads.21.C3.S raw_reads.92.raw_reads.21.N3.S
LAsort -v raw_reads.92.raw_reads.22.C0 raw_reads.92.raw_reads.22.N0 raw_reads.92.raw_reads.22.C1 raw_reads.92.raw_reads.22.N1 raw_reads.92.raw_reads.22.C2 raw_reads.92.raw_reads.22.N2 raw_reads.92.raw_reads.22.C3 raw_reads.92.raw_reads.22.N3 && LAmerge -v L1.92.22 raw_reads.92.raw_reads.22.C0.S raw_reads.92.raw_reads.22.N0.S raw_reads.92.raw_reads.22.C1.S raw_reads.92.raw_reads.22.N1.S raw_reads.92.raw_reads.22.C2.S raw_reads.92.raw_reads.22.N2.S raw_reads.92.raw_reads.22.C3.S raw_reads.92.raw_reads.22.N3.S
LAsort -v raw_reads.92.raw_reads.23.C0 raw_reads.92.raw_reads.23.N0 raw_reads.92.raw_reads.23.C1 raw_reads.92.raw_reads.23.N1 raw_reads.92.raw_reads.23.C2 raw_reads.92.raw_reads.23.N2 raw_reads.92.raw_reads.23.C3 raw_reads.92.raw_reads.23.N3 && LAmerge -v L1.92.23 raw_reads.92.raw_reads.23.C0.S raw_reads.92.raw_reads.23.N0.S raw_reads.92.raw_reads.23.C1.S raw_reads.92.raw_reads.23.N1.S raw_reads.92.raw_reads.23.C2.S raw_reads.92.raw_reads.23.N2.S raw_reads.92.raw_reads.23.C3.S raw_reads.92.raw_reads.23.N3.S
LAcheck -vS raw_reads L1.1.92
LAcheck -vS raw_reads L1.2.92
LAcheck -vS raw_reads L1.3.92
LAcheck -vS raw_reads L1.4.92
LAcheck -vS raw_reads L1.5.92
LAcheck -vS raw_reads L1.6.92
LAcheck -vS raw_reads L1.7.92
LAcheck -vS raw_reads L1.8.92
LAcheck -vS raw_reads L1.9.92
LAcheck -vS raw_reads L1.10.92
LAcheck -vS raw_reads L1.11.92
LAcheck -vS raw_reads L1.12.92
LAcheck -vS raw_reads L1.13.92
LAcheck -vS raw_reads L1.14.92
LAcheck -vS raw_reads L1.15.92
LAcheck -vS raw_reads L1.16.92
LAcheck -vS raw_reads L1.17.92
LAcheck -vS raw_reads L1.18.92
LAcheck -vS raw_reads L1.19.92
LAcheck -vS raw_reads L1.20.92
LAcheck -vS raw_reads L1.21.92
LAcheck -vS raw_reads L1.22.92
LAcheck -vS raw_reads L1.23.92
LAcheck -vS raw_reads L1.92.1
LAcheck -vS raw_reads L1.92.2
LAcheck -vS raw_reads L1.92.3
LAcheck -vS raw_reads L1.92.4
LAcheck -vS raw_reads L1.92.5
LAcheck -vS raw_reads L1.92.6
LAcheck -vS raw_reads L1.92.7
LAcheck -vS raw_reads L1.92.8
LAcheck -vS raw_reads L1.92.9
LAcheck -vS raw_reads L1.92.10
LAcheck -vS raw_reads L1.92.11
LAcheck -vS raw_reads L1.92.12
LAcheck -vS raw_reads L1.92.13
LAcheck -vS raw_reads L1.92.14
LAcheck -vS raw_reads L1.92.15
LAcheck -vS raw_reads L1.92.16
LAcheck -vS raw_reads L1.92.17
LAcheck -vS raw_reads L1.92.18
LAcheck -vS raw_reads L1.92.19
LAcheck -vS raw_reads L1.92.20
LAcheck -vS raw_reads L1.92.21
LAcheck -vS raw_reads L1.92.22
LAcheck -vS raw_reads L1.92.23
daligner -v -k18 -h70 -t14 -H7000 -e0.75 raw_reads.91 raw_reads.46 raw_reads.47 raw_reads.48 raw_reads.49 raw_reads.50 raw_reads.51 raw_reads.52 raw_reads.53 raw_reads.54 raw_reads.55 raw_reads.56 raw_reads.57 raw_reads.58 raw_reads.59 raw_reads.60 raw_reads.61 raw_reads.62 raw_reads.63 raw_reads.64 raw_reads.65 raw_reads.66 raw_reads.67 raw_reads.68
LAcheck -v raw_reads *.las
LAsort -v raw_reads.46.raw_reads.91.C0 raw_reads.46.raw_reads.91.N0 raw_reads.46.raw_reads.91.C1 raw_reads.46.raw_reads.91.N1 raw_reads.46.raw_reads.91.C2 raw_reads.46.raw_reads.91.N2 raw_reads.46.raw_reads.91.C3 raw_reads.46.raw_reads.91.N3 && LAmerge -v L1.46.91 raw_reads.46.raw_reads.91.C0.S raw_reads.46.raw_reads.91.N0.S raw_reads.46.raw_reads.91.C1.S raw_reads.46.raw_reads.91.N1.S raw_reads.46.raw_reads.91.C2.S raw_reads.46.raw_reads.91.N2.S raw_reads.46.raw_reads.91.C3.S raw_reads.46.raw_reads.91.N3.S
LAsort -v raw_reads.47.raw_reads.91.C0 raw_reads.47.raw_reads.91.N0 raw_reads.47.raw_reads.91.C1 raw_reads.47.raw_reads.91.N1 raw_reads.47.raw_reads.91.C2 raw_reads.47.raw_reads.91.N2 raw_reads.47.raw_reads.91.C3 raw_reads.47.raw_reads.91.N3 && LAmerge -v L1.47.91 raw_reads.47.raw_reads.91.C0.S raw_reads.47.raw_reads.91.N0.S raw_reads.47.raw_reads.91.C1.S raw_reads.47.raw_reads.91.N1.S raw_reads.47.raw_reads.91.C2.S raw_reads.47.raw_reads.91.N2.S raw_reads.47.raw_reads.91.C3.S raw_reads.47.raw_reads.91.N3.S
LAsort -v raw_reads.48.raw_reads.91.C0 raw_reads.48.raw_reads.91.N0 raw_reads.48.raw_reads.91.C1 raw_reads.48.raw_reads.91.N1 raw_reads.48.raw_reads.91.C2 raw_reads.48.raw_reads.91.N2 raw_reads.48.raw_reads.91.C3 raw_reads.48.raw_reads.91.N3 && LAmerge -v L1.48.91 raw_reads.48.raw_reads.91.C0.S raw_reads.48.raw_reads.91.N0.S raw_reads.48.raw_reads.91.C1.S raw_reads.48.raw_reads.91.N1.S raw_reads.48.raw_reads.91.C2.S raw_reads.48.raw_reads.91.N2.S raw_reads.48.raw_reads.91.C3.S raw_reads.48.raw_reads.91.N3.S
LAsort -v raw_reads.49.raw_reads.91.C0 raw_reads.49.raw_reads.91.N0 raw_reads.49.raw_reads.91.C1 raw_reads.49.raw_reads.91.N1 raw_reads.49.raw_reads.91.C2 raw_reads.49.raw_reads.91.N2 raw_reads.49.raw_reads.91.C3 raw_reads.49.raw_reads.91.N3 && LAmerge -v L1.49.91 raw_reads.49.raw_reads.91.C0.S raw_reads.49.raw_reads.91.N0.S raw_reads.49.raw_reads.91.C1.S raw_reads.49.raw_reads.91.N1.S raw_reads.49.raw_reads.91.C2.S raw_reads.49.raw_reads.91.N2.S raw_reads.49.raw_reads.91.C3.S raw_reads.49.raw_reads.91.N3.S
LAsort -v raw_reads.50.raw_reads.91.C0 raw_reads.50.raw_reads.91.N0 raw_reads.50.raw_reads.91.C1 raw_reads.50.raw_reads.91.N1 raw_reads.50.raw_reads.91.C2 raw_reads.50.raw_reads.91.N2 raw_reads.50.raw_reads.91.C3 raw_reads.50.raw_reads.91.N3 && LAmerge -v L1.50.91 raw_reads.50.raw_reads.91.C0.S raw_reads.50.raw_reads.91.N0.S raw_reads.50.raw_reads.91.C1.S raw_reads.50.raw_reads.91.N1.S raw_reads.50.raw_reads.91.C2.S raw_reads.50.raw_reads.91.N2.S raw_reads.50.raw_reads.91.C3.S raw_reads.50.raw_reads.91.N3.S
LAsort -v raw_reads.51.raw_reads.91.C0 raw_reads.51.raw_reads.91.N0 raw_reads.51.raw_reads.91.C1 raw_reads.51.raw_reads.91.N1 raw_reads.51.raw_reads.91.C2 raw_reads.51.raw_reads.91.N2 raw_reads.51.raw_reads.91.C3 raw_reads.51.raw_reads.91.N3 && LAmerge -v L1.51.91 raw_reads.51.raw_reads.91.C0.S raw_reads.51.raw_reads.91.N0.S raw_reads.51.raw_reads.91.C1.S raw_reads.51.raw_reads.91.N1.S raw_reads.51.raw_reads.91.C2.S raw_reads.51.raw_reads.91.N2.S raw_reads.51.raw_reads.91.C3.S raw_reads.51.raw_reads.91.N3.S
LAsort -v raw_reads.52.raw_reads.91.C0 raw_reads.52.raw_reads.91.N0 raw_reads.52.raw_reads.91.C1 raw_reads.52.raw_reads.91.N1 raw_reads.52.raw_reads.91.C2 raw_reads.52.raw_reads.91.N2 raw_reads.52.raw_reads.91.C3 raw_reads.52.raw_reads.91.N3 && LAmerge -v L1.52.91 raw_reads.52.raw_reads.91.C0.S raw_reads.52.raw_reads.91.N0.S raw_reads.52.raw_reads.91.C1.S raw_reads.52.raw_reads.91.N1.S raw_reads.52.raw_reads.91.C2.S raw_reads.52.raw_reads.91.N2.S raw_reads.52.raw_reads.91.C3.S raw_reads.52.raw_reads.91.N3.S
LAsort -v raw_reads.53.raw_reads.91.C0 raw_reads.53.raw_reads.91.N0 raw_reads.53.raw_reads.91.C1 raw_reads.53.raw_reads.91.N1 raw_reads.53.raw_reads.91.C2 raw_reads.53.raw_reads.91.N2 raw_reads.53.raw_reads.91.C3 raw_reads.53.raw_reads.91.N3 && LAmerge -v L1.53.91 raw_reads.53.raw_reads.91.C0.S raw_reads.53.raw_reads.91.N0.S raw_reads.53.raw_reads.91.C1.S raw_reads.53.raw_reads.91.N1.S raw_reads.53.raw_reads.91.C2.S raw_reads.53.raw_reads.91.N2.S raw_reads.53.raw_reads.91.C3.S raw_reads.53.raw_reads.91.N3.S
LAsort -v raw_reads.54.raw_reads.91.C0 raw_reads.54.raw_reads.91.N0 raw_reads.54.raw_reads.91.C1 raw_reads.54.raw_reads.91.N1 raw_reads.54.raw_reads.91.C2 raw_reads.54.raw_reads.91.N2 raw_reads.54.raw_reads.91.C3 raw_reads.54.raw_reads.91.N3 && LAmerge -v L1.54.91 raw_reads.54.raw_reads.91.C0.S raw_reads.54.raw_reads.91.N0.S raw_reads.54.raw_reads.91.C1.S raw_reads.54.raw_reads.91.N1.S raw_reads.54.raw_reads.91.C2.S raw_reads.54.raw_reads.91.N2.S raw_reads.54.raw_reads.91.C3.S raw_reads.54.raw_reads.91.N3.S
LAsort -v raw_reads.55.raw_reads.91.C0 raw_reads.55.raw_reads.91.N0 raw_reads.55.raw_reads.91.C1 raw_reads.55.raw_reads.91.N1 raw_reads.55.raw_reads.91.C2 raw_reads.55.raw_reads.91.N2 raw_reads.55.raw_reads.91.C3 raw_reads.55.raw_reads.91.N3 && LAmerge -v L1.55.91 raw_reads.55.raw_reads.91.C0.S raw_reads.55.raw_reads.91.N0.S raw_reads.55.raw_reads.91.C1.S raw_reads.55.raw_reads.91.N1.S raw_reads.55.raw_reads.91.C2.S raw_reads.55.raw_reads.91.N2.S raw_reads.55.raw_reads.91.C3.S raw_reads.55.raw_reads.91.N3.S
LAsort -v raw_reads.56.raw_reads.91.C0 raw_reads.56.raw_reads.91.N0 raw_reads.56.raw_reads.91.C1 raw_reads.56.raw_reads.91.N1 raw_reads.56.raw_reads.91.C2 raw_reads.56.raw_reads.91.N2 raw_reads.56.raw_reads.91.C3 raw_reads.56.raw_reads.91.N3 && LAmerge -v L1.56.91 raw_reads.56.raw_reads.91.C0.S raw_reads.56.raw_reads.91.N0.S raw_reads.56.raw_reads.91.C1.S raw_reads.56.raw_reads.91.N1.S raw_reads.56.raw_reads.91.C2.S raw_reads.56.raw_reads.91.N2.S raw_reads.56.raw_reads.91.C3.S raw_reads.56.raw_reads.91.N3.S
LAsort -v raw_reads.57.raw_reads.91.C0 raw_reads.57.raw_reads.91.N0 raw_reads.57.raw_reads.91.C1 raw_reads.57.raw_reads.91.N1 raw_reads.57.raw_reads.91.C2 raw_reads.57.raw_reads.91.N2 raw_reads.57.raw_reads.91.C3 raw_reads.57.raw_reads.91.N3 && LAmerge -v L1.57.91 raw_reads.57.raw_reads.91.C0.S raw_reads.57.raw_reads.91.N0.S raw_reads.57.raw_reads.91.C1.S raw_reads.57.raw_reads.91.N1.S raw_reads.57.raw_reads.91.C2.S raw_reads.57.raw_reads.91.N2.S raw_reads.57.raw_reads.91.C3.S raw_reads.57.raw_reads.91.N3.S
LAsort -v raw_reads.58.raw_reads.91.C0 raw_reads.58.raw_reads.91.N0 raw_reads.58.raw_reads.91.C1 raw_reads.58.raw_reads.91.N1 raw_reads.58.raw_reads.91.C2 raw_reads.58.raw_reads.91.N2 raw_reads.58.raw_reads.91.C3 raw_reads.58.raw_reads.91.N3 && LAmerge -v L1.58.91 raw_reads.58.raw_reads.91.C0.S raw_reads.58.raw_reads.91.N0.S raw_reads.58.raw_reads.91.C1.S raw_reads.58.raw_reads.91.N1.S raw_reads.58.raw_reads.91.C2.S raw_reads.58.raw_reads.91.N2.S raw_reads.58.raw_reads.91.C3.S raw_reads.58.raw_reads.91.N3.S
LAsort -v raw_reads.59.raw_reads.91.C0 raw_reads.59.raw_reads.91.N0 raw_reads.59.raw_reads.91.C1 raw_reads.59.raw_reads.91.N1 raw_reads.59.raw_reads.91.C2 raw_reads.59.raw_reads.91.N2 raw_reads.59.raw_reads.91.C3 raw_reads.59.raw_reads.91.N3 && LAmerge -v L1.59.91 raw_reads.59.raw_reads.91.C0.S raw_reads.59.raw_reads.91.N0.S raw_reads.59.raw_reads.91.C1.S raw_reads.59.raw_reads.91.N1.S raw_reads.59.raw_reads.91.C2.S raw_reads.59.raw_reads.91.N2.S raw_reads.59.raw_reads.91.C3.S raw_reads.59.raw_reads.91.N3.S
LAsort -v raw_reads.60.raw_reads.91.C0 raw_reads.60.raw_reads.91.N0 raw_reads.60.raw_reads.91.C1 raw_reads.60.raw_reads.91.N1 raw_reads.60.raw_reads.91.C2 raw_reads.60.raw_reads.91.N2 raw_reads.60.raw_reads.91.C3 raw_reads.60.raw_reads.91.N3 && LAmerge -v L1.60.91 raw_reads.60.raw_reads.91.C0.S raw_reads.60.raw_reads.91.N0.S raw_reads.60.raw_reads.91.C1.S raw_reads.60.raw_reads.91.N1.S raw_reads.60.raw_reads.91.C2.S raw_reads.60.raw_reads.91.N2.S raw_reads.60.raw_reads.91.C3.S raw_reads.60.raw_reads.91.N3.S
LAsort -v raw_reads.61.raw_reads.91.C0 raw_reads.61.raw_reads.91.N0 raw_reads.61.raw_reads.91.C1 raw_reads.61.raw_reads.91.N1 raw_reads.61.raw_reads.91.C2 raw_reads.61.raw_reads.91.N2 raw_reads.61.raw_reads.91.C3 raw_reads.61.raw_reads.91.N3 && LAmerge -v L1.61.91 raw_reads.61.raw_reads.91.C0.S raw_reads.61.raw_reads.91.N0.S raw_reads.61.raw_reads.91.C1.S raw_reads.61.raw_reads.91.N1.S raw_reads.61.raw_reads.91.C2.S raw_reads.61.raw_reads.91.N2.S raw_reads.61.raw_reads.91.C3.S raw_reads.61.raw_reads.91.N3.S
LAsort -v raw_reads.62.raw_reads.91.C0 raw_reads.62.raw_reads.91.N0 raw_reads.62.raw_reads.91.C1 raw_reads.62.raw_reads.91.N1 raw_reads.62.raw_reads.91.C2 raw_reads.62.raw_reads.91.N2 raw_reads.62.raw_reads.91.C3 raw_reads.62.raw_reads.91.N3 && LAmerge -v L1.62.91 raw_reads.62.raw_reads.91.C0.S raw_reads.62.raw_reads.91.N0.S raw_reads.62.raw_reads.91.C1.S raw_reads.62.raw_reads.91.N1.S raw_reads.62.raw_reads.91.C2.S raw_reads.62.raw_reads.91.N2.S raw_reads.62.raw_reads.91.C3.S raw_reads.62.raw_reads.91.N3.S
LAsort -v raw_reads.63.raw_reads.91.C0 raw_reads.63.raw_reads.91.N0 raw_reads.63.raw_reads.91.C1 raw_reads.63.raw_reads.91.N1 raw_reads.63.raw_reads.91.C2 raw_reads.63.raw_reads.91.N2 raw_reads.63.raw_reads.91.C3 raw_reads.63.raw_reads.91.N3 && LAmerge -v L1.63.91 raw_reads.63.raw_reads.91.C0.S raw_reads.63.raw_reads.91.N0.S raw_reads.63.raw_reads.91.C1.S raw_reads.63.raw_reads.91.N1.S raw_reads.63.raw_reads.91.C2.S raw_reads.63.raw_reads.91.N2.S raw_reads.63.raw_reads.91.C3.S raw_reads.63.raw_reads.91.N3.S
LAsort -v raw_reads.64.raw_reads.91.C0 raw_reads.64.raw_reads.91.N0 raw_reads.64.raw_reads.91.C1 raw_reads.64.raw_reads.91.N1 raw_reads.64.raw_reads.91.C2 raw_reads.64.raw_reads.91.N2 raw_reads.64.raw_reads.91.C3 raw_reads.64.raw_reads.91.N3 && LAmerge -v L1.64.91 raw_reads.64.raw_reads.91.C0.S raw_reads.64.raw_reads.91.N0.S raw_reads.64.raw_reads.91.C1.S raw_reads.64.raw_reads.91.N1.S raw_reads.64.raw_reads.91.C2.S raw_reads.64.raw_reads.91.N2.S raw_reads.64.raw_reads.91.C3.S raw_reads.64.raw_reads.91.N3.S
LAsort -v raw_reads.65.raw_reads.91.C0 raw_reads.65.raw_reads.91.N0 raw_reads.65.raw_reads.91.C1 raw_reads.65.raw_reads.91.N1 raw_reads.65.raw_reads.91.C2 raw_reads.65.raw_reads.91.N2 raw_reads.65.raw_reads.91.C3 raw_reads.65.raw_reads.91.N3 && LAmerge -v L1.65.91 raw_reads.65.raw_reads.91.C0.S raw_reads.65.raw_reads.91.N0.S raw_reads.65.raw_reads.91.C1.S raw_reads.65.raw_reads.91.N1.S raw_reads.65.raw_reads.91.C2.S raw_reads.65.raw_reads.91.N2.S raw_reads.65.raw_reads.91.C3.S raw_reads.65.raw_reads.91.N3.S
LAsort -v raw_reads.66.raw_reads.91.C0 raw_reads.66.raw_reads.91.N0 raw_reads.66.raw_reads.91.C1 raw_reads.66.raw_reads.91.N1 raw_reads.66.raw_reads.91.C2 raw_reads.66.raw_reads.91.N2 raw_reads.66.raw_reads.91.C3 raw_reads.66.raw_reads.91.N3 && LAmerge -v L1.66.91 raw_reads.66.raw_reads.91.C0.S raw_reads.66.raw_reads.91.N0.S raw_reads.66.raw_reads.91.C1.S raw_reads.66.raw_reads.91.N1.S raw_reads.66.raw_reads.91.C2.S raw_reads.66.raw_reads.91.N2.S raw_reads.66.raw_reads.91.C3.S raw_reads.66.raw_reads.91.N3.S
LAsort -v raw_reads.67.raw_reads.91.C0 raw_reads.67.raw_reads.91.N0 raw_reads.67.raw_reads.91.C1 raw_reads.67.raw_reads.91.N1 raw_reads.67.raw_reads.91.C2 raw_reads.67.raw_reads.91.N2 raw_reads.67.raw_reads.91.C3 raw_reads.67.raw_reads.91.N3 && LAmerge -v L1.67.91 raw_reads.67.raw_reads.91.C0.S raw_reads.67.raw_reads.91.N0.S raw_reads.67.raw_reads.91.C1.S raw_reads.67.raw_reads.91.N1.S raw_reads.67.raw_reads.91.C2.S raw_reads.67.raw_reads.91.N2.S raw_reads.67.raw_reads.91.C3.S raw_reads.67.raw_reads.91.N3.S
LAsort -v raw_reads.68.raw_reads.91.C0 raw_reads.68.raw_reads.91.N0 raw_reads.68.raw_reads.91.C1 raw_reads.68.raw_reads.91.N1 raw_reads.68.raw_reads.91.C2 raw_reads.68.raw_reads.91.N2 raw_reads.68.raw_reads.91.C3 raw_reads.68.raw_reads.91.N3 && LAmerge -v L1.68.91 raw_reads.68.raw_reads.91.C0.S raw_reads.68.raw_reads.91.N0.S raw_reads.68.raw_reads.91.C1.S raw_reads.68.raw_reads.91.N1.S raw_reads.68.raw_reads.91.C2.S raw_reads.68.raw_reads.91.N2.S raw_reads.68.raw_reads.91.C3.S raw_reads.68.raw_reads.91.N3.S
LAsort -v raw_reads.91.raw_reads.46.C0 raw_reads.91.raw_reads.46.N0 raw_reads.91.raw_reads.46.C1 raw_reads.91.raw_reads.46.N1 raw_reads.91.raw_reads.46.C2 raw_reads.91.raw_reads.46.N2 raw_reads.91.raw_reads.46.C3 raw_reads.91.raw_reads.46.N3 && LAmerge -v L1.91.46 raw_reads.91.raw_reads.46.C0.S raw_reads.91.raw_reads.46.N0.S raw_reads.91.raw_reads.46.C1.S raw_reads.91.raw_reads.46.N1.S raw_reads.91.raw_reads.46.C2.S raw_reads.91.raw_reads.46.N2.S raw_reads.91.raw_reads.46.C3.S raw_reads.91.raw_reads.46.N3.S
LAsort -v raw_reads.91.raw_reads.47.C0 raw_reads.91.raw_reads.47.N0 raw_reads.91.raw_reads.47.C1 raw_reads.91.raw_reads.47.N1 raw_reads.91.raw_reads.47.C2 raw_reads.91.raw_reads.47.N2 raw_reads.91.raw_reads.47.C3 raw_reads.91.raw_reads.47.N3 && LAmerge -v L1.91.47 raw_reads.91.raw_reads.47.C0.S raw_reads.91.raw_reads.47.N0.S raw_reads.91.raw_reads.47.C1.S raw_reads.91.raw_reads.47.N1.S raw_reads.91.raw_reads.47.C2.S raw_reads.91.raw_reads.47.N2.S raw_reads.91.raw_reads.47.C3.S raw_reads.91.raw_reads.47.N3.S
LAsort -v raw_reads.91.raw_reads.48.C0 raw_reads.91.raw_reads.48.N0 raw_reads.91.raw_reads.48.C1 raw_reads.91.raw_reads.48.N1 raw_reads.91.raw_reads.48.C2 raw_reads.91.raw_reads.48.N2 raw_reads.91.raw_reads.48.C3 raw_reads.91.raw_reads.48.N3 && LAmerge -v L1.91.48 raw_reads.91.raw_reads.48.C0.S raw_reads.91.raw_reads.48.N0.S raw_reads.91.raw_reads.48.C1.S raw_reads.91.raw_reads.48.N1.S raw_reads.91.raw_reads.48.C2.S raw_reads.91.raw_reads.48.N2.S raw_reads.91.raw_reads.48.C3.S raw_reads.91.raw_reads.48.N3.S
LAsort -v raw_reads.91.raw_reads.49.C0 raw_reads.91.raw_reads.49.N0 raw_reads.91.raw_reads.49.C1 raw_reads.91.raw_reads.49.N1 raw_reads.91.raw_reads.49.C2 raw_reads.91.raw_reads.49.N2 raw_reads.91.raw_reads.49.C3 raw_reads.91.raw_reads.49.N3 && LAmerge -v L1.91.49 raw_reads.91.raw_reads.49.C0.S raw_reads.91.raw_reads.49.N0.S raw_reads.91.raw_reads.49.C1.S raw_reads.91.raw_reads.49.N1.S raw_reads.91.raw_reads.49.C2.S raw_reads.91.raw_reads.49.N2.S raw_reads.91.raw_reads.49.C3.S raw_reads.91.raw_reads.49.N3.S
LAsort -v raw_reads.91.raw_reads.50.C0 raw_reads.91.raw_reads.50.N0 raw_reads.91.raw_reads.50.C1 raw_reads.91.raw_reads.50.N1 raw_reads.91.raw_reads.50.C2 raw_reads.91.raw_reads.50.N2 raw_reads.91.raw_reads.50.C3 raw_reads.91.raw_reads.50.N3 && LAmerge -v L1.91.50 raw_reads.91.raw_reads.50.C0.S raw_reads.91.raw_reads.50.N0.S raw_reads.91.raw_reads.50.C1.S raw_reads.91.raw_reads.50.N1.S raw_reads.91.raw_reads.50.C2.S raw_reads.91.raw_reads.50.N2.S raw_reads.91.raw_reads.50.C3.S raw_reads.91.raw_reads.50.N3.S
LAsort -v raw_reads.91.raw_reads.51.C0 raw_reads.91.raw_reads.51.N0 raw_reads.91.raw_reads.51.C1 raw_reads.91.raw_reads.51.N1 raw_reads.91.raw_reads.51.C2 raw_reads.91.raw_reads.51.N2 raw_reads.91.raw_reads.51.C3 raw_reads.91.raw_reads.51.N3 && LAmerge -v L1.91.51 raw_reads.91.raw_reads.51.C0.S raw_reads.91.raw_reads.51.N0.S raw_reads.91.raw_reads.51.C1.S raw_reads.91.raw_reads.51.N1.S raw_reads.91.raw_reads.51.C2.S raw_reads.91.raw_reads.51.N2.S raw_reads.91.raw_reads.51.C3.S raw_reads.91.raw_reads.51.N3.S
LAsort -v raw_reads.91.raw_reads.52.C0 raw_reads.91.raw_reads.52.N0 raw_reads.91.raw_reads.52.C1 raw_reads.91.raw_reads.52.N1 raw_reads.91.raw_reads.52.C2 raw_reads.91.raw_reads.52.N2 raw_reads.91.raw_reads.52.C3 raw_reads.91.raw_reads.52.N3 && LAmerge -v L1.91.52 raw_reads.91.raw_reads.52.C0.S raw_reads.91.raw_reads.52.N0.S raw_reads.91.raw_reads.52.C1.S raw_reads.91.raw_reads.52.N1.S raw_reads.91.raw_reads.52.C2.S raw_reads.91.raw_reads.52.N2.S raw_reads.91.raw_reads.52.C3.S raw_reads.91.raw_reads.52.N3.S
LAsort -v raw_reads.91.raw_reads.53.C0 raw_reads.91.raw_reads.53.N0 raw_reads.91.raw_reads.53.C1 raw_reads.91.raw_reads.53.N1 raw_reads.91.raw_reads.53.C2 raw_reads.91.raw_reads.53.N2 raw_reads.91.raw_reads.53.C3 raw_reads.91.raw_reads.53.N3 && LAmerge -v L1.91.53 raw_reads.91.raw_reads.53.C0.S raw_reads.91.raw_reads.53.N0.S raw_reads.91.raw_reads.53.C1.S raw_reads.91.raw_reads.53.N1.S raw_reads.91.raw_reads.53.C2.S raw_reads.91.raw_reads.53.N2.S raw_reads.91.raw_reads.53.C3.S raw_reads.91.raw_reads.53.N3.S
LAsort -v raw_reads.91.raw_reads.54.C0 raw_reads.91.raw_reads.54.N0 raw_reads.91.raw_reads.54.C1 raw_reads.91.raw_reads.54.N1 raw_reads.91.raw_reads.54.C2 raw_reads.91.raw_reads.54.N2 raw_reads.91.raw_reads.54.C3 raw_reads.91.raw_reads.54.N3 && LAmerge -v L1.91.54 raw_reads.91.raw_reads.54.C0.S raw_reads.91.raw_reads.54.N0.S raw_reads.91.raw_reads.54.C1.S raw_reads.91.raw_reads.54.N1.S raw_reads.91.raw_reads.54.C2.S raw_reads.91.raw_reads.54.N2.S raw_reads.91.raw_reads.54.C3.S raw_reads.91.raw_reads.54.N3.S
LAsort -v raw_reads.91.raw_reads.55.C0 raw_reads.91.raw_reads.55.N0 raw_reads.91.raw_reads.55.C1 raw_reads.91.raw_reads.55.N1 raw_reads.91.raw_reads.55.C2 raw_reads.91.raw_reads.55.N2 raw_reads.91.raw_reads.55.C3 raw_reads.91.raw_reads.55.N3 && LAmerge -v L1.91.55 raw_reads.91.raw_reads.55.C0.S raw_reads.91.raw_reads.55.N0.S raw_reads.91.raw_reads.55.C1.S raw_reads.91.raw_reads.55.N1.S raw_reads.91.raw_reads.55.C2.S raw_reads.91.raw_reads.55.N2.S raw_reads.91.raw_reads.55.C3.S raw_reads.91.raw_reads.55.N3.S
LAsort -v raw_reads.91.raw_reads.56.C0 raw_reads.91.raw_reads.56.N0 raw_reads.91.raw_reads.56.C1 raw_reads.91.raw_reads.56.N1 raw_reads.91.raw_reads.56.C2 raw_reads.91.raw_reads.56.N2 raw_reads.91.raw_reads.56.C3 raw_reads.91.raw_reads.56.N3 && LAmerge -v L1.91.56 raw_reads.91.raw_reads.56.C0.S raw_reads.91.raw_reads.56.N0.S raw_reads.91.raw_reads.56.C1.S raw_reads.91.raw_reads.56.N1.S raw_reads.91.raw_reads.56.C2.S raw_reads.91.raw_reads.56.N2.S raw_reads.91.raw_reads.56.C3.S raw_reads.91.raw_reads.56.N3.S
LAsort -v raw_reads.91.raw_reads.57.C0 raw_reads.91.raw_reads.57.N0 raw_reads.91.raw_reads.57.C1 raw_reads.91.raw_reads.57.N1 raw_reads.91.raw_reads.57.C2 raw_reads.91.raw_reads.57.N2 raw_reads.91.raw_reads.57.C3 raw_reads.91.raw_reads.57.N3 && LAmerge -v L1.91.57 raw_reads.91.raw_reads.57.C0.S raw_reads.91.raw_reads.57.N0.S raw_reads.91.raw_reads.57.C1.S raw_reads.91.raw_reads.57.N1.S raw_reads.91.raw_reads.57.C2.S raw_reads.91.raw_reads.57.N2.S raw_reads.91.raw_reads.57.C3.S raw_reads.91.raw_reads.57.N3.S
LAsort -v raw_reads.91.raw_reads.58.C0 raw_reads.91.raw_reads.58.N0 raw_reads.91.raw_reads.58.C1 raw_reads.91.raw_reads.58.N1 raw_reads.91.raw_reads.58.C2 raw_reads.91.raw_reads.58.N2 raw_reads.91.raw_reads.58.C3 raw_reads.91.raw_reads.58.N3 && LAmerge -v L1.91.58 raw_reads.91.raw_reads.58.C0.S raw_reads.91.raw_reads.58.N0.S raw_reads.91.raw_reads.58.C1.S raw_reads.91.raw_reads.58.N1.S raw_reads.91.raw_reads.58.C2.S raw_reads.91.raw_reads.58.N2.S raw_reads.91.raw_reads.58.C3.S raw_reads.91.raw_reads.58.N3.S
LAsort -v raw_reads.91.raw_reads.59.C0 raw_reads.91.raw_reads.59.N0 raw_reads.91.raw_reads.59.C1 raw_reads.91.raw_reads.59.N1 raw_reads.91.raw_reads.59.C2 raw_reads.91.raw_reads.59.N2 raw_reads.91.raw_reads.59.C3 raw_reads.91.raw_reads.59.N3 && LAmerge -v L1.91.59 raw_reads.91.raw_reads.59.C0.S raw_reads.91.raw_reads.59.N0.S raw_reads.91.raw_reads.59.C1.S raw_reads.91.raw_reads.59.N1.S raw_reads.91.raw_reads.59.C2.S raw_reads.91.raw_reads.59.N2.S raw_reads.91.raw_reads.59.C3.S raw_reads.91.raw_reads.59.N3.S
LAsort -v raw_reads.91.raw_reads.60.C0 raw_reads.91.raw_reads.60.N0 raw_reads.91.raw_reads.60.C1 raw_reads.91.raw_reads.60.N1 raw_reads.91.raw_reads.60.C2 raw_reads.91.raw_reads.60.N2 raw_reads.91.raw_reads.60.C3 raw_reads.91.raw_reads.60.N3 && LAmerge -v L1.91.60 raw_reads.91.raw_reads.60.C0.S raw_reads.91.raw_reads.60.N0.S raw_reads.91.raw_reads.60.C1.S raw_reads.91.raw_reads.60.N1.S raw_reads.91.raw_reads.60.C2.S raw_reads.91.raw_reads.60.N2.S raw_reads.91.raw_reads.60.C3.S raw_reads.91.raw_reads.60.N3.S
LAsort -v raw_reads.91.raw_reads.61.C0 raw_reads.91.raw_reads.61.N0 raw_reads.91.raw_reads.61.C1 raw_reads.91.raw_reads.61.N1 raw_reads.91.raw_reads.61.C2 raw_reads.91.raw_reads.61.N2 raw_reads.91.raw_reads.61.C3 raw_reads.91.raw_reads.61.N3 && LAmerge -v L1.91.61 raw_reads.91.raw_reads.61.C0.S raw_reads.91.raw_reads.61.N0.S raw_reads.91.raw_reads.61.C1.S raw_reads.91.raw_reads.61.N1.S raw_reads.91.raw_reads.61.C2.S raw_reads.91.raw_reads.61.N2.S raw_reads.91.raw_reads.61.C3.S raw_reads.91.raw_reads.61.N3.S
LAsort -v raw_reads.91.raw_reads.62.C0 raw_reads.91.raw_reads.62.N0 raw_reads.91.raw_reads.62.C1 raw_reads.91.raw_reads.62.N1 raw_reads.91.raw_reads.62.C2 raw_reads.91.raw_reads.62.N2 raw_reads.91.raw_reads.62.C3 raw_reads.91.raw_reads.62.N3 && LAmerge -v L1.91.62 raw_reads.91.raw_reads.62.C0.S raw_reads.91.raw_reads.62.N0.S raw_reads.91.raw_reads.62.C1.S raw_reads.91.raw_reads.62.N1.S raw_reads.91.raw_reads.62.C2.S raw_reads.91.raw_reads.62.N2.S raw_reads.91.raw_reads.62.C3.S raw_reads.91.raw_reads.62.N3.S
LAsort -v raw_reads.91.raw_reads.63.C0 raw_reads.91.raw_reads.63.N0 raw_reads.91.raw_reads.63.C1 raw_reads.91.raw_reads.63.N1 raw_reads.91.raw_reads.63.C2 raw_reads.91.raw_reads.63.N2 raw_reads.91.raw_reads.63.C3 raw_reads.91.raw_reads.63.N3 && LAmerge -v L1.91.63 raw_reads.91.raw_reads.63.C0.S raw_reads.91.raw_reads.63.N0.S raw_reads.91.raw_reads.63.C1.S raw_reads.91.raw_reads.63.N1.S raw_reads.91.raw_reads.63.C2.S raw_reads.91.raw_reads.63.N2.S raw_reads.91.raw_reads.63.C3.S raw_reads.91.raw_reads.63.N3.S
LAsort -v raw_reads.91.raw_reads.64.C0 raw_reads.91.raw_reads.64.N0 raw_reads.91.raw_reads.64.C1 raw_reads.91.raw_reads.64.N1 raw_reads.91.raw_reads.64.C2 raw_reads.91.raw_reads.64.N2 raw_reads.91.raw_reads.64.C3 raw_reads.91.raw_reads.64.N3 && LAmerge -v L1.91.64 raw_reads.91.raw_reads.64.C0.S raw_reads.91.raw_reads.64.N0.S raw_reads.91.raw_reads.64.C1.S raw_reads.91.raw_reads.64.N1.S raw_reads.91.raw_reads.64.C2.S raw_reads.91.raw_reads.64.N2.S raw_reads.91.raw_reads.64.C3.S raw_reads.91.raw_reads.64.N3.S
LAsort -v raw_reads.91.raw_reads.65.C0 raw_reads.91.raw_reads.65.N0 raw_reads.91.raw_reads.65.C1 raw_reads.91.raw_reads.65.N1 raw_reads.91.raw_reads.65.C2 raw_reads.91.raw_reads.65.N2 raw_reads.91.raw_reads.65.C3 raw_reads.91.raw_reads.65.N3 && LAmerge -v L1.91.65 raw_reads.91.raw_reads.65.C0.S raw_reads.91.raw_reads.65.N0.S raw_reads.91.raw_reads.65.C1.S raw_reads.91.raw_reads.65.N1.S raw_reads.91.raw_reads.65.C2.S raw_reads.91.raw_reads.65.N2.S raw_reads.91.raw_reads.65.C3.S raw_reads.91.raw_reads.65.N3.S
LAsort -v raw_reads.91.raw_reads.66.C0 raw_reads.91.raw_reads.66.N0 raw_reads.91.raw_reads.66.C1 raw_reads.91.raw_reads.66.N1 raw_reads.91.raw_reads.66.C2 raw_reads.91.raw_reads.66.N2 raw_reads.91.raw_reads.66.C3 raw_reads.91.raw_reads.66.N3 && LAmerge -v L1.91.66 raw_reads.91.raw_reads.66.C0.S raw_reads.91.raw_reads.66.N0.S raw_reads.91.raw_reads.66.C1.S raw_reads.91.raw_reads.66.N1.S raw_reads.91.raw_reads.66.C2.S raw_reads.91.raw_reads.66.N2.S raw_reads.91.raw_reads.66.C3.S raw_reads.91.raw_reads.66.N3.S
LAsort -v raw_reads.91.raw_reads.67.C0 raw_reads.91.raw_reads.67.N0 raw_reads.91.raw_reads.67.C1 raw_reads.91.raw_reads.67.N1 raw_reads.91.raw_reads.67.C2 raw_reads.91.raw_reads.67.N2 raw_reads.91.raw_reads.67.C3 raw_reads.91.raw_reads.67.N3 && LAmerge -v L1.91.67 raw_reads.91.raw_reads.67.C0.S raw_reads.91.raw_reads.67.N0.S raw_reads.91.raw_reads.67.C1.S raw_reads.91.raw_reads.67.N1.S raw_reads.91.raw_reads.67.C2.S raw_reads.91.raw_reads.67.N2.S raw_reads.91.raw_reads.67.C3.S raw_reads.91.raw_reads.67.N3.S
LAsort -v raw_reads.91.raw_reads.68.C0 raw_reads.91.raw_reads.68.N0 raw_reads.91.raw_reads.68.C1 raw_reads.91.raw_reads.68.N1 raw_reads.91.raw_reads.68.C2 raw_reads.91.raw_reads.68.N2 raw_reads.91.raw_reads.68.C3 raw_reads.91.raw_reads.68.N3 && LAmerge -v L1.91.68 raw_reads.91.raw_reads.68.C0.S raw_reads.91.raw_reads.68.N0.S raw_reads.91.raw_reads.68.C1.S raw_reads.91.raw_reads.68.N1.S raw_reads.91.raw_reads.68.C2.S raw_reads.91.raw_reads.68.N2.S raw_reads.91.raw_reads.68.C3.S raw_reads.91.raw_reads.68.N3.S
LAcheck -vS raw_reads L1.46.91
LAcheck -vS raw_reads L1.47.91
LAcheck -vS raw_reads L1.48.91
LAcheck -vS raw_reads L1.49.91
LAcheck -vS raw_reads L1.50.91
LAcheck -vS raw_reads L1.51.91
LAcheck -vS raw_reads L1.52.91
LAcheck -vS raw_reads L1.53.91
LAcheck -vS raw_reads L1.54.91
LAcheck -vS raw_reads L1.55.91
LAcheck -vS raw_reads L1.56.91
LAcheck -vS raw_reads L1.57.91
LAcheck -vS raw_reads L1.58.91
LAcheck -vS raw_reads L1.59.91
LAcheck -vS raw_reads L1.60.91
LAcheck -vS raw_reads L1.61.91
LAcheck -vS raw_reads L1.62.91
LAcheck -vS raw_reads L1.63.91
LAcheck -vS raw_reads L1.64.91
LAcheck -vS raw_reads L1.65.91
LAcheck -vS raw_reads L1.66.91
LAcheck -vS raw_reads L1.67.91
LAcheck -vS raw_reads L1.68.91
LAcheck -vS raw_reads L1.91.46
LAcheck -vS raw_reads L1.91.47
LAcheck -vS raw_reads L1.91.48
LAcheck -vS raw_reads L1.91.49
LAcheck -vS raw_reads L1.91.50
LAcheck -vS raw_reads L1.91.51
LAcheck -vS raw_reads L1.91.52
LAcheck -vS raw_reads L1.91.53
LAcheck -vS raw_reads L1.91.54
LAcheck -vS raw_reads L1.91.55
LAcheck -vS raw_reads L1.91.56
LAcheck -vS raw_reads L1.91.57
LAcheck -vS raw_reads L1.91.58
LAcheck -vS raw_reads L1.91.59
LAcheck -vS raw_reads L1.91.60
LAcheck -vS raw_reads L1.91.61
LAcheck -vS raw_reads L1.91.62
LAcheck -vS raw_reads L1.91.63
LAcheck -vS raw_reads L1.91.64
LAcheck -vS raw_reads L1.91.65
LAcheck -vS raw_reads L1.91.66
LAcheck -vS raw_reads L1.91.67
LAcheck -vS raw_reads L1.91.68
daligner -v -k18 -h70 -t14 -H7000 -e0.75 raw_reads.27 raw_reads.14 raw_reads.15 raw_reads.16 raw_reads.17 raw_reads.18 raw_reads.19 raw_reads.20 raw_reads.21 raw_reads.22 raw_reads.23 raw_reads.24 raw_reads.25 raw_reads.26 raw_reads.27
LAcheck -v raw_reads *.las
LAsort -v raw_reads.14.raw_reads.27.C0 raw_reads.14.raw_reads.27.N0 raw_reads.14.raw_reads.27.C1 raw_reads.14.raw_reads.27.N1 raw_reads.14.raw_reads.27.C2 raw_reads.14.raw_reads.27.N2 raw_reads.14.raw_reads.27.C3 raw_reads.14.raw_reads.27.N3 && LAmerge -v L1.14.27 raw_reads.14.raw_reads.27.C0.S raw_reads.14.raw_reads.27.N0.S raw_reads.14.raw_reads.27.C1.S raw_reads.14.raw_reads.27.N1.S raw_reads.14.raw_reads.27.C2.S raw_reads.14.raw_reads.27.N2.S raw_reads.14.raw_reads.27.C3.S raw_reads.14.raw_reads.27.N3.S
LAsort -v raw_reads.15.raw_reads.27.C0 raw_reads.15.raw_reads.27.N0 raw_reads.15.raw_reads.27.C1 raw_reads.15.raw_reads.27.N1 raw_reads.15.raw_reads.27.C2 raw_reads.15.raw_reads.27.N2 raw_reads.15.raw_reads.27.C3 raw_reads.15.raw_reads.27.N3 && LAmerge -v L1.15.27 raw_reads.15.raw_reads.27.C0.S raw_reads.15.raw_reads.27.N0.S raw_reads.15.raw_reads.27.C1.S raw_reads.15.raw_reads.27.N1.S raw_reads.15.raw_reads.27.C2.S raw_reads.15.raw_reads.27.N2.S raw_reads.15.raw_reads.27.C3.S raw_reads.15.raw_reads.27.N3.S
LAsort -v raw_reads.16.raw_reads.27.C0 raw_reads.16.raw_reads.27.N0 raw_reads.16.raw_reads.27.C1 raw_reads.16.raw_reads.27.N1 raw_reads.16.raw_reads.27.C2 raw_reads.16.raw_reads.27.N2 raw_reads.16.raw_reads.27.C3 raw_reads.16.raw_reads.27.N3 && LAmerge -v L1.16.27 raw_reads.16.raw_reads.27.C0.S raw_reads.16.raw_reads.27.N0.S raw_reads.16.raw_reads.27.C1.S raw_reads.16.raw_reads.27.N1.S raw_reads.16.raw_reads.27.C2.S raw_reads.16.raw_reads.27.N2.S raw_reads.16.raw_reads.27.C3.S raw_reads.16.raw_reads.27.N3.S
LAsort -v raw_reads.17.raw_reads.27.C0 raw_reads.17.raw_reads.27.N0 raw_reads.17.raw_reads.27.C1 raw_reads.17.raw_reads.27.N1 raw_reads.17.raw_reads.27.C2 raw_reads.17.raw_reads.27.N2 raw_reads.17.raw_reads.27.C3 raw_reads.17.raw_reads.27.N3 && LAmerge -v L1.17.27 raw_reads.17.raw_reads.27.C0.S raw_reads.17.raw_reads.27.N0.S raw_reads.17.raw_reads.27.C1.S raw_reads.17.raw_reads.27.N1.S raw_reads.17.raw_reads.27.C2.S raw_reads.17.raw_reads.27.N2.S raw_reads.17.raw_reads.27.C3.S raw_reads.17.raw_reads.27.N3.S
LAsort -v raw_reads.18.raw_reads.27.C0 raw_reads.18.raw_reads.27.N0 raw_reads.18.raw_reads.27.C1 raw_reads.18.raw_reads.27.N1 raw_reads.18.raw_reads.27.C2 raw_reads.18.raw_reads.27.N2 raw_reads.18.raw_reads.27.C3 raw_reads.18.raw_reads.27.N3 && LAmerge -v L1.18.27 raw_reads.18.raw_reads.27.C0.S raw_reads.18.raw_reads.27.N0.S raw_reads.18.raw_reads.27.C1.S raw_reads.18.raw_reads.27.N1.S raw_reads.18.raw_reads.27.C2.S raw_reads.18.raw_reads.27.N2.S raw_reads.18.raw_reads.27.C3.S raw_reads.18.raw_reads.27.N3.S
LAsort -v raw_reads.19.raw_reads.27.C0 raw_reads.19.raw_reads.27.N0 raw_reads.19.raw_reads.27.C1 raw_reads.19.raw_reads.27.N1 raw_reads.19.raw_reads.27.C2 raw_reads.19.raw_reads.27.N2 raw_reads.19.raw_reads.27.C3 raw_reads.19.raw_reads.27.N3 && LAmerge -v L1.19.27 raw_reads.19.raw_reads.27.C0.S raw_reads.19.raw_reads.27.N0.S raw_reads.19.raw_reads.27.C1.S raw_reads.19.raw_reads.27.N1.S raw_reads.19.raw_reads.27.C2.S raw_reads.19.raw_reads.27.N2.S raw_reads.19.raw_reads.27.C3.S raw_reads.19.raw_reads.27.N3.S
LAsort -v raw_reads.20.raw_reads.27.C0 raw_reads.20.raw_reads.27.N0 raw_reads.20.raw_reads.27.C1 raw_reads.20.raw_reads.27.N1 raw_reads.20.raw_reads.27.C2 raw_reads.20.raw_reads.27.N2 raw_reads.20.raw_reads.27.C3 raw_reads.20.raw_reads.27.N3 && LAmerge -v L1.20.27 raw_reads.20.raw_reads.27.C0.S raw_reads.20.raw_reads.27.N0.S raw_reads.20.raw_reads.27.C1.S raw_reads.20.raw_reads.27.N1.S raw_reads.20.raw_reads.27.C2.S raw_reads.20.raw_reads.27.N2.S raw_reads.20.raw_reads.27.C3.S raw_reads.20.raw_reads.27.N3.S
LAsort -v raw_reads.21.raw_reads.27.C0 raw_reads.21.raw_reads.27.N0 raw_reads.21.raw_reads.27.C1 raw_reads.21.raw_reads.27.N1 raw_reads.21.raw_reads.27.C2 raw_reads.21.raw_reads.27.N2 raw_reads.21.raw_reads.27.C3 raw_reads.21.raw_reads.27.N3 && LAmerge -v L1.21.27 raw_reads.21.raw_reads.27.C0.S raw_reads.21.raw_reads.27.N0.S raw_reads.21.raw_reads.27.C1.S raw_reads.21.raw_reads.27.N1.S raw_reads.21.raw_reads.27.C2.S raw_reads.21.raw_reads.27.N2.S raw_reads.21.raw_reads.27.C3.S raw_reads.21.raw_reads.27.N3.S
LAsort -v raw_reads.22.raw_reads.27.C0 raw_reads.22.raw_reads.27.N0 raw_reads.22.raw_reads.27.C1 raw_reads.22.raw_reads.27.N1 raw_reads.22.raw_reads.27.C2 raw_reads.22.raw_reads.27.N2 raw_reads.22.raw_reads.27.C3 raw_reads.22.raw_reads.27.N3 && LAmerge -v L1.22.27 raw_reads.22.raw_reads.27.C0.S raw_reads.22.raw_reads.27.N0.S raw_reads.22.raw_reads.27.C1.S raw_reads.22.raw_reads.27.N1.S raw_reads.22.raw_reads.27.C2.S raw_reads.22.raw_reads.27.N2.S raw_reads.22.raw_reads.27.C3.S raw_reads.22.raw_reads.27.N3.S
LAsort -v raw_reads.23.raw_reads.27.C0 raw_reads.23.raw_reads.27.N0 raw_reads.23.raw_reads.27.C1 raw_reads.23.raw_reads.27.N1 raw_reads.23.raw_reads.27.C2 raw_reads.23.raw_reads.27.N2 raw_reads.23.raw_reads.27.C3 raw_reads.23.raw_reads.27.N3 && LAmerge -v L1.23.27 raw_reads.23.raw_reads.27.C0.S raw_reads.23.raw_reads.27.N0.S raw_reads.23.raw_reads.27.C1.S raw_reads.23.raw_reads.27.N1.S raw_reads.23.raw_reads.27.C2.S raw_reads.23.raw_reads.27.N2.S raw_reads.23.raw_reads.27.C3.S raw_reads.23.raw_reads.27.N3.S
LAsort -v raw_reads.24.raw_reads.27.C0 raw_reads.24.raw_reads.27.N0 raw_reads.24.raw_reads.27.C1 raw_reads.24.raw_reads.27.N1 raw_reads.24.raw_reads.27.C2 raw_reads.24.raw_reads.27.N2 raw_reads.24.raw_reads.27.C3 raw_reads.24.raw_reads.27.N3 && LAmerge -v L1.24.27 raw_reads.24.raw_reads.27.C0.S raw_reads.24.raw_reads.27.N0.S raw_reads.24.raw_reads.27.C1.S raw_reads.24.raw_reads.27.N1.S raw_reads.24.raw_reads.27.C2.S raw_reads.24.raw_reads.27.N2.S raw_reads.24.raw_reads.27.C3.S raw_reads.24.raw_reads.27.N3.S
LAsort -v raw_reads.25.raw_reads.27.C0 raw_reads.25.raw_reads.27.N0 raw_reads.25.raw_reads.27.C1 raw_reads.25.raw_reads.27.N1 raw_reads.25.raw_reads.27.C2 raw_reads.25.raw_reads.27.N2 raw_reads.25.raw_reads.27.C3 raw_reads.25.raw_reads.27.N3 && LAmerge -v L1.25.27 raw_reads.25.raw_reads.27.C0.S raw_reads.25.raw_reads.27.N0.S raw_reads.25.raw_reads.27.C1.S raw_reads.25.raw_reads.27.N1.S raw_reads.25.raw_reads.27.C2.S raw_reads.25.raw_reads.27.N2.S raw_reads.25.raw_reads.27.C3.S raw_reads.25.raw_reads.27.N3.S
LAsort -v raw_reads.26.raw_reads.27.C0 raw_reads.26.raw_reads.27.N0 raw_reads.26.raw_reads.27.C1 raw_reads.26.raw_reads.27.N1 raw_reads.26.raw_reads.27.C2 raw_reads.26.raw_reads.27.N2 raw_reads.26.raw_reads.27.C3 raw_reads.26.raw_reads.27.N3 && LAmerge -v L1.26.27 raw_reads.26.raw_reads.27.C0.S raw_reads.26.raw_reads.27.N0.S raw_reads.26.raw_reads.27.C1.S raw_reads.26.raw_reads.27.N1.S raw_reads.26.raw_reads.27.C2.S raw_reads.26.raw_reads.27.N2.S raw_reads.26.raw_reads.27.C3.S raw_reads.26.raw_reads.27.N3.S
LAsort -v raw_reads.27.raw_reads.14.C0 raw_reads.27.raw_reads.14.N0 raw_reads.27.raw_reads.14.C1 raw_reads.27.raw_reads.14.N1 raw_reads.27.raw_reads.14.C2 raw_reads.27.raw_reads.14.N2 raw_reads.27.raw_reads.14.C3 raw_reads.27.raw_reads.14.N3 && LAmerge -v L1.27.14 raw_reads.27.raw_reads.14.C0.S raw_reads.27.raw_reads.14.N0.S raw_reads.27.raw_reads.14.C1.S raw_reads.27.raw_reads.14.N1.S raw_reads.27.raw_reads.14.C2.S raw_reads.27.raw_reads.14.N2.S raw_reads.27.raw_reads.14.C3.S raw_reads.27.raw_reads.14.N3.S
LAsort -v raw_reads.27.raw_reads.15.C0 raw_reads.27.raw_reads.15.N0 raw_reads.27.raw_reads.15.C1 raw_reads.27.raw_reads.15.N1 raw_reads.27.raw_reads.15.C2 raw_reads.27.raw_reads.15.N2 raw_reads.27.raw_reads.15.C3 raw_reads.27.raw_reads.15.N3 && LAmerge -v L1.27.15 raw_reads.27.raw_reads.15.C0.S raw_reads.27.raw_reads.15.N0.S raw_reads.27.raw_reads.15.C1.S raw_reads.27.raw_reads.15.N1.S raw_reads.27.raw_reads.15.C2.S raw_reads.27.raw_reads.15.N2.S raw_reads.27.raw_reads.15.C3.S raw_reads.27.raw_reads.15.N3.S
LAsort -v raw_reads.27.raw_reads.16.C0 raw_reads.27.raw_reads.16.N0 raw_reads.27.raw_reads.16.C1 raw_reads.27.raw_reads.16.N1 raw_reads.27.raw_reads.16.C2 raw_reads.27.raw_reads.16.N2 raw_reads.27.raw_reads.16.C3 raw_reads.27.raw_reads.16.N3 && LAmerge -v L1.27.16 raw_reads.27.raw_reads.16.C0.S raw_reads.27.raw_reads.16.N0.S raw_reads.27.raw_reads.16.C1.S raw_reads.27.raw_reads.16.N1.S raw_reads.27.raw_reads.16.C2.S raw_reads.27.raw_reads.16.N2.S raw_reads.27.raw_reads.16.C3.S raw_reads.27.raw_reads.16.N3.S
LAsort -v raw_reads.27.raw_reads.17.C0 raw_reads.27.raw_reads.17.N0 raw_reads.27.raw_reads.17.C1 raw_reads.27.raw_reads.17.N1 raw_reads.27.raw_reads.17.C2 raw_reads.27.raw_reads.17.N2 raw_reads.27.raw_reads.17.C3 raw_reads.27.raw_reads.17.N3 && LAmerge -v L1.27.17 raw_reads.27.raw_reads.17.C0.S raw_reads.27.raw_reads.17.N0.S raw_reads.27.raw_reads.17.C1.S raw_reads.27.raw_reads.17.N1.S raw_reads.27.raw_reads.17.C2.S raw_reads.27.raw_reads.17.N2.S raw_reads.27.raw_reads.17.C3.S raw_reads.27.raw_reads.17.N3.S
LAsort -v raw_reads.27.raw_reads.18.C0 raw_reads.27.raw_reads.18.N0 raw_reads.27.raw_reads.18.C1 raw_reads.27.raw_reads.18.N1 raw_reads.27.raw_reads.18.C2 raw_reads.27.raw_reads.18.N2 raw_reads.27.raw_reads.18.C3 raw_reads.27.raw_reads.18.N3 && LAmerge -v L1.27.18 raw_reads.27.raw_reads.18.C0.S raw_reads.27.raw_reads.18.N0.S raw_reads.27.raw_reads.18.C1.S raw_reads.27.raw_reads.18.N1.S raw_reads.27.raw_reads.18.C2.S raw_reads.27.raw_reads.18.N2.S raw_reads.27.raw_reads.18.C3.S raw_reads.27.raw_reads.18.N3.S
LAsort -v raw_reads.27.raw_reads.19.C0 raw_reads.27.raw_reads.19.N0 raw_reads.27.raw_reads.19.C1 raw_reads.27.raw_reads.19.N1 raw_reads.27.raw_reads.19.C2 raw_reads.27.raw_reads.19.N2 raw_reads.27.raw_reads.19.C3 raw_reads.27.raw_reads.19.N3 && LAmerge -v L1.27.19 raw_reads.27.raw_reads.19.C0.S raw_reads.27.raw_reads.19.N0.S raw_reads.27.raw_reads.19.C1.S raw_reads.27.raw_reads.19.N1.S raw_reads.27.raw_reads.19.C2.S raw_reads.27.raw_reads.19.N2.S raw_reads.27.raw_reads.19.C3.S raw_reads.27.raw_reads.19.N3.S
LAsort -v raw_reads.27.raw_reads.20.C0 raw_reads.27.raw_reads.20.N0 raw_reads.27.raw_reads.20.C1 raw_reads.27.raw_reads.20.N1 raw_reads.27.raw_reads.20.C2 raw_reads.27.raw_reads.20.N2 raw_reads.27.raw_reads.20.C3 raw_reads.27.raw_reads.20.N3 && LAmerge -v L1.27.20 raw_reads.27.raw_reads.20.C0.S raw_reads.27.raw_reads.20.N0.S raw_reads.27.raw_reads.20.C1.S raw_reads.27.raw_reads.20.N1.S raw_reads.27.raw_reads.20.C2.S raw_reads.27.raw_reads.20.N2.S raw_reads.27.raw_reads.20.C3.S raw_reads.27.raw_reads.20.N3.S
LAsort -v raw_reads.27.raw_reads.21.C0 raw_reads.27.raw_reads.21.N0 raw_reads.27.raw_reads.21.C1 raw_reads.27.raw_reads.21.N1 raw_reads.27.raw_reads.21.C2 raw_reads.27.raw_reads.21.N2 raw_reads.27.raw_reads.21.C3 raw_reads.27.raw_reads.21.N3 && LAmerge -v L1.27.21 raw_reads.27.raw_reads.21.C0.S raw_reads.27.raw_reads.21.N0.S raw_reads.27.raw_reads.21.C1.S raw_reads.27.raw_reads.21.N1.S raw_reads.27.raw_reads.21.C2.S raw_reads.27.raw_reads.21.N2.S raw_reads.27.raw_reads.21.C3.S raw_reads.27.raw_reads.21.N3.S
LAsort -v raw_reads.27.raw_reads.22.C0 raw_reads.27.raw_reads.22.N0 raw_reads.27.raw_reads.22.C1 raw_reads.27.raw_reads.22.N1 raw_reads.27.raw_reads.22.C2 raw_reads.27.raw_reads.22.N2 raw_reads.27.raw_reads.22.C3 raw_reads.27.raw_reads.22.N3 && LAmerge -v L1.27.22 raw_reads.27.raw_reads.22.C0.S raw_reads.27.raw_reads.22.N0.S raw_reads.27.raw_reads.22.C1.S raw_reads.27.raw_reads.22.N1.S raw_reads.27.raw_reads.22.C2.S raw_reads.27.raw_reads.22.N2.S raw_reads.27.raw_reads.22.C3.S raw_reads.27.raw_reads.22.N3.S
LAsort -v raw_reads.27.raw_reads.23.C0 raw_reads.27.raw_reads.23.N0 raw_reads.27.raw_reads.23.C1 raw_reads.27.raw_reads.23.N1 raw_reads.27.raw_reads.23.C2 raw_reads.27.raw_reads.23.N2 raw_reads.27.raw_reads.23.C3 raw_reads.27.raw_reads.23.N3 && LAmerge -v L1.27.23 raw_reads.27.raw_reads.23.C0.S raw_reads.27.raw_reads.23.N0.S raw_reads.27.raw_reads.23.C1.S raw_reads.27.raw_reads.23.N1.S raw_reads.27.raw_reads.23.C2.S raw_reads.27.raw_reads.23.N2.S raw_reads.27.raw_reads.23.C3.S raw_reads.27.raw_reads.23.N3.S
LAsort -v raw_reads.27.raw_reads.24.C0 raw_reads.27.raw_reads.24.N0 raw_reads.27.raw_reads.24.C1 raw_reads.27.raw_reads.24.N1 raw_reads.27.raw_reads.24.C2 raw_reads.27.raw_reads.24.N2 raw_reads.27.raw_reads.24.C3 raw_reads.27.raw_reads.24.N3 && LAmerge -v L1.27.24 raw_reads.27.raw_reads.24.C0.S raw_reads.27.raw_reads.24.N0.S raw_reads.27.raw_reads.24.C1.S raw_reads.27.raw_reads.24.N1.S raw_reads.27.raw_reads.24.C2.S raw_reads.27.raw_reads.24.N2.S raw_reads.27.raw_reads.24.C3.S raw_reads.27.raw_reads.24.N3.S
LAsort -v raw_reads.27.raw_reads.25.C0 raw_reads.27.raw_reads.25.N0 raw_reads.27.raw_reads.25.C1 raw_reads.27.raw_reads.25.N1 raw_reads.27.raw_reads.25.C2 raw_reads.27.raw_reads.25.N2 raw_reads.27.raw_reads.25.C3 raw_reads.27.raw_reads.25.N3 && LAmerge -v L1.27.25 raw_reads.27.raw_reads.25.C0.S raw_reads.27.raw_reads.25.N0.S raw_reads.27.raw_reads.25.C1.S raw_reads.27.raw_reads.25.N1.S raw_reads.27.raw_reads.25.C2.S raw_reads.27.raw_reads.25.N2.S raw_reads.27.raw_reads.25.C3.S raw_reads.27.raw_reads.25.N3.S
LAsort -v raw_reads.27.raw_reads.26.C0 raw_reads.27.raw_reads.26.N0 raw_reads.27.raw_reads.26.C1 raw_reads.27.raw_reads.26.N1 raw_reads.27.raw_reads.26.C2 raw_reads.27.raw_reads.26.N2 raw_reads.27.raw_reads.26.C3 raw_reads.27.raw_reads.26.N3 && LAmerge -v L1.27.26 raw_reads.27.raw_reads.26.C0.S raw_reads.27.raw_reads.26.N0.S raw_reads.27.raw_reads.26.C1.S raw_reads.27.raw_reads.26.N1.S raw_reads.27.raw_reads.26.C2.S raw_reads.27.raw_reads.26.N2.S raw_reads.27.raw_reads.26.C3.S raw_reads.27.raw_reads.26.N3.S
LAsort -v raw_reads.27.raw_reads.27.C0 raw_reads.27.raw_reads.27.N0 raw_reads.27.raw_reads.27.C1 raw_reads.27.raw_reads.27.N1 raw_reads.27.raw_reads.27.C2 raw_reads.27.raw_reads.27.N2 raw_reads.27.raw_reads.27.C3 raw_reads.27.raw_reads.27.N3 && LAmerge -v L1.27.27 raw_reads.27.raw_reads.27.C0.S raw_reads.27.raw_reads.27.N0.S raw_reads.27.raw_reads.27.C1.S raw_reads.27.raw_reads.27.N1.S raw_reads.27.raw_reads.27.C2.S raw_reads.27.raw_reads.27.N2.S raw_reads.27.raw_reads.27.C3.S raw_reads.27.raw_reads.27.N3.S
LAcheck -vS raw_reads L1.14.27
LAcheck -vS raw_reads L1.15.27
LAcheck -vS raw_reads L1.16.27
LAcheck -vS raw_reads L1.17.27
LAcheck -vS raw_reads L1.18.27
LAcheck -vS raw_reads L1.19.27
LAcheck -vS raw_reads L1.20.27
LAcheck -vS raw_reads L1.21.27
LAcheck -vS raw_reads L1.22.27
LAcheck -vS raw_reads L1.23.27
LAcheck -vS raw_reads L1.24.27
LAcheck -vS raw_reads L1.25.27
LAcheck -vS raw_reads L1.26.27
LAcheck -vS raw_reads L1.27.14
LAcheck -vS raw_reads L1.27.15
LAcheck -vS raw_reads L1.27.16
LAcheck -vS raw_reads L1.27.17
LAcheck -vS raw_reads L1.27.18
LAcheck -vS raw_reads L1.27.19
LAcheck -vS raw_reads L1.27.20
LAcheck -vS raw_reads L1.27.21
LAcheck -vS raw_reads L1.27.22
LAcheck -vS raw_reads L1.27.23
LAcheck -vS raw_reads L1.27.24
LAcheck -vS raw_reads L1.27.25
LAcheck -vS raw_reads L1.27.26
LAcheck -vS raw_reads L1.27.27
daligner -v -k18 -h70 -t14 -H7000 -e0.75 raw_reads.73 raw_reads.19 raw_reads.20 raw_reads.21 raw_reads.22 raw_reads.23 raw_reads.24 raw_reads.25 raw_reads.26 raw_reads.27 raw_reads.28 raw_reads.29 raw_reads.30 raw_reads.31 raw_reads.32 raw_reads.33 raw_reads.34 raw_reads.35 raw_reads.36
LAcheck -v raw_reads *.las
LAsort -v raw_reads.19.raw_reads.73.C0 raw_reads.19.raw_reads.73.N0 raw_reads.19.raw_reads.73.C1 raw_reads.19.raw_reads.73.N1 raw_reads.19.raw_reads.73.C2 raw_reads.19.raw_reads.73.N2 raw_reads.19.raw_reads.73.C3 raw_reads.19.raw_reads.73.N3 && LAmerge -v L1.19.73 raw_reads.19.raw_reads.73.C0.S raw_reads.19.raw_reads.73.N0.S raw_reads.19.raw_reads.73.C1.S raw_reads.19.raw_reads.73.N1.S raw_reads.19.raw_reads.73.C2.S raw_reads.19.raw_reads.73.N2.S raw_reads.19.raw_reads.73.C3.S raw_reads.19.raw_reads.73.N3.S
LAsort -v raw_reads.20.raw_reads.73.C0 raw_reads.20.raw_reads.73.N0 raw_reads.20.raw_reads.73.C1 raw_reads.20.raw_reads.73.N1 raw_reads.20.raw_reads.73.C2 raw_reads.20.raw_reads.73.N2 raw_reads.20.raw_reads.73.C3 raw_reads.20.raw_reads.73.N3 && LAmerge -v L1.20.73 raw_reads.20.raw_reads.73.C0.S raw_reads.20.raw_reads.73.N0.S raw_reads.20.raw_reads.73.C1.S raw_reads.20.raw_reads.73.N1.S raw_reads.20.raw_reads.73.C2.S raw_reads.20.raw_reads.73.N2.S raw_reads.20.raw_reads.73.C3.S raw_reads.20.raw_reads.73.N3.S
LAsort -v raw_reads.21.raw_reads.73.C0 raw_reads.21.raw_reads.73.N0 raw_reads.21.raw_reads.73.C1 raw_reads.21.raw_reads.73.N1 raw_reads.21.raw_reads.73.C2 raw_reads.21.raw_reads.73.N2 raw_reads.21.raw_reads.73.C3 raw_reads.21.raw_reads.73.N3 && LAmerge -v L1.21.73 raw_reads.21.raw_reads.73.C0.S raw_reads.21.raw_reads.73.N0.S raw_reads.21.raw_reads.73.C1.S raw_reads.21.raw_reads.73.N1.S raw_reads.21.raw_reads.73.C2.S raw_reads.21.raw_reads.73.N2.S raw_reads.21.raw_reads.73.C3.S raw_reads.21.raw_reads.73.N3.S
LAsort -v raw_reads.22.raw_reads.73.C0 raw_reads.22.raw_reads.73.N0 raw_reads.22.raw_reads.73.C1 raw_reads.22.raw_reads.73.N1 raw_reads.22.raw_reads.73.C2 raw_reads.22.raw_reads.73.N2 raw_reads.22.raw_reads.73.C3 raw_reads.22.raw_reads.73.N3 && LAmerge -v L1.22.73 raw_reads.22.raw_reads.73.C0.S raw_reads.22.raw_reads.73.N0.S raw_reads.22.raw_reads.73.C1.S raw_reads.22.raw_reads.73.N1.S raw_reads.22.raw_reads.73.C2.S raw_reads.22.raw_reads.73.N2.S raw_reads.22.raw_reads.73.C3.S raw_reads.22.raw_reads.73.N3.S
LAsort -v raw_reads.23.raw_reads.73.C0 raw_reads.23.raw_reads.73.N0 raw_reads.23.raw_reads.73.C1 raw_reads.23.raw_reads.73.N1 raw_reads.23.raw_reads.73.C2 raw_reads.23.raw_reads.73.N2 raw_reads.23.raw_reads.73.C3 raw_reads.23.raw_reads.73.N3 && LAmerge -v L1.23.73 raw_reads.23.raw_reads.73.C0.S raw_reads.23.raw_reads.73.N0.S raw_reads.23.raw_reads.73.C1.S raw_reads.23.raw_reads.73.N1.S raw_reads.23.raw_reads.73.C2.S raw_reads.23.raw_reads.73.N2.S raw_reads.23.raw_reads.73.C3.S raw_reads.23.raw_reads.73.N3.S
LAsort -v raw_reads.24.raw_reads.73.C0 raw_reads.24.raw_reads.73.N0 raw_reads.24.raw_reads.73.C1 raw_reads.24.raw_reads.73.N1 raw_reads.24.raw_reads.73.C2 raw_reads.24.raw_reads.73.N2 raw_reads.24.raw_reads.73.C3 raw_reads.24.raw_reads.73.N3 && LAmerge -v L1.24.73 raw_reads.24.raw_reads.73.C0.S raw_reads.24.raw_reads.73.N0.S raw_reads.24.raw_reads.73.C1.S raw_reads.24.raw_reads.73.N1.S raw_reads.24.raw_reads.73.C2.S raw_reads.24.raw_reads.73.N2.S raw_reads.24.raw_reads.73.C3.S raw_reads.24.raw_reads.73.N3.S
LAsort -v raw_reads.25.raw_reads.73.C0 raw_reads.25.raw_reads.73.N0 raw_reads.25.raw_reads.73.C1 raw_reads.25.raw_reads.73.N1 raw_reads.25.raw_reads.73.C2 raw_reads.25.raw_reads.73.N2 raw_reads.25.raw_reads.73.C3 raw_reads.25.raw_reads.73.N3 && LAmerge -v L1.25.73 raw_reads.25.raw_reads.73.C0.S raw_reads.25.raw_reads.73.N0.S raw_reads.25.raw_reads.73.C1.S raw_reads.25.raw_reads.73.N1.S raw_reads.25.raw_reads.73.C2.S raw_reads.25.raw_reads.73.N2.S raw_reads.25.raw_reads.73.C3.S raw_reads.25.raw_reads.73.N3.S
LAsort -v raw_reads.26.raw_reads.73.C0 raw_reads.26.raw_reads.73.N0 raw_reads.26.raw_reads.73.C1 raw_reads.26.raw_reads.73.N1 raw_reads.26.raw_reads.73.C2 raw_reads.26.raw_reads.73.N2 raw_reads.26.raw_reads.73.C3 raw_reads.26.raw_reads.73.N3 && LAmerge -v L1.26.73 raw_reads.26.raw_reads.73.C0.S raw_reads.26.raw_reads.73.N0.S raw_reads.26.raw_reads.73.C1.S raw_reads.26.raw_reads.73.N1.S raw_reads.26.raw_reads.73.C2.S raw_reads.26.raw_reads.73.N2.S raw_reads.26.raw_reads.73.C3.S raw_reads.26.raw_reads.73.N3.S
LAsort -v raw_reads.27.raw_reads.73.C0 raw_reads.27.raw_reads.73.N0 raw_reads.27.raw_reads.73.C1 raw_reads.27.raw_reads.73.N1 raw_reads.27.raw_reads.73.C2 raw_reads.27.raw_reads.73.N2 raw_reads.27.raw_reads.73.C3 raw_reads.27.raw_reads.73.N3 && LAmerge -v L1.27.73 raw_reads.27.raw_reads.73.C0.S raw_reads.27.raw_reads.73.N0.S raw_reads.27.raw_reads.73.C1.S raw_reads.27.raw_reads.73.N1.S raw_reads.27.raw_reads.73.C2.S raw_reads.27.raw_reads.73.N2.S raw_reads.27.raw_reads.73.C3.S raw_reads.27.raw_reads.73.N3.S
LAsort -v raw_reads.28.raw_reads.73.C0 raw_reads.28.raw_reads.73.N0 raw_reads.28.raw_reads.73.C1 raw_reads.28.raw_reads.73.N1 raw_reads.28.raw_reads.73.C2 raw_reads.28.raw_reads.73.N2 raw_reads.28.raw_reads.73.C3 raw_reads.28.raw_reads.73.N3 && LAmerge -v L1.28.73 raw_reads.28.raw_reads.73.C0.S raw_reads.28.raw_reads.73.N0.S raw_reads.28.raw_reads.73.C1.S raw_reads.28.raw_reads.73.N1.S raw_reads.28.raw_reads.73.C2.S raw_reads.28.raw_reads.73.N2.S raw_reads.28.raw_reads.73.C3.S raw_reads.28.raw_reads.73.N3.S
LAsort -v raw_reads.29.raw_reads.73.C0 raw_reads.29.raw_reads.73.N0 raw_reads.29.raw_reads.73.C1 raw_reads.29.raw_reads.73.N1 raw_reads.29.raw_reads.73.C2 raw_reads.29.raw_reads.73.N2 raw_reads.29.raw_reads.73.C3 raw_reads.29.raw_reads.73.N3 && LAmerge -v L1.29.73 raw_reads.29.raw_reads.73.C0.S raw_reads.29.raw_reads.73.N0.S raw_reads.29.raw_reads.73.C1.S raw_reads.29.raw_reads.73.N1.S raw_reads.29.raw_reads.73.C2.S raw_reads.29.raw_reads.73.N2.S raw_reads.29.raw_reads.73.C3.S raw_reads.29.raw_reads.73.N3.S
LAsort -v raw_reads.30.raw_reads.73.C0 raw_reads.30.raw_reads.73.N0 raw_reads.30.raw_reads.73.C1 raw_reads.30.raw_reads.73.N1 raw_reads.30.raw_reads.73.C2 raw_reads.30.raw_reads.73.N2 raw_reads.30.raw_reads.73.C3 raw_reads.30.raw_reads.73.N3 && LAmerge -v L1.30.73 raw_reads.30.raw_reads.73.C0.S raw_reads.30.raw_reads.73.N0.S raw_reads.30.raw_reads.73.C1.S raw_reads.30.raw_reads.73.N1.S raw_reads.30.raw_reads.73.C2.S raw_reads.30.raw_reads.73.N2.S raw_reads.30.raw_reads.73.C3.S raw_reads.30.raw_reads.73.N3.S
LAsort -v raw_reads.31.raw_reads.73.C0 raw_reads.31.raw_reads.73.N0 raw_reads.31.raw_reads.73.C1 raw_reads.31.raw_reads.73.N1 raw_reads.31.raw_reads.73.C2 raw_reads.31.raw_reads.73.N2 raw_reads.31.raw_reads.73.C3 raw_reads.31.raw_reads.73.N3 && LAmerge -v L1.31.73 raw_reads.31.raw_reads.73.C0.S raw_reads.31.raw_reads.73.N0.S raw_reads.31.raw_reads.73.C1.S raw_reads.31.raw_reads.73.N1.S raw_reads.31.raw_reads.73.C2.S raw_reads.31.raw_reads.73.N2.S raw_reads.31.raw_reads.73.C3.S raw_reads.31.raw_reads.73.N3.S
LAsort -v raw_reads.32.raw_reads.73.C0 raw_reads.32.raw_reads.73.N0 raw_reads.32.raw_reads.73.C1 raw_reads.32.raw_reads.73.N1 raw_reads.32.raw_reads.73.C2 raw_reads.32.raw_reads.73.N2 raw_reads.32.raw_reads.73.C3 raw_reads.32.raw_reads.73.N3 && LAmerge -v L1.32.73 raw_reads.32.raw_reads.73.C0.S raw_reads.32.raw_reads.73.N0.S raw_reads.32.raw_reads.73.C1.S raw_reads.32.raw_reads.73.N1.S raw_reads.32.raw_reads.73.C2.S raw_reads.32.raw_reads.73.N2.S raw_reads.32.raw_reads.73.C3.S raw_reads.32.raw_reads.73.N3.S
LAsort -v raw_reads.33.raw_reads.73.C0 raw_reads.33.raw_reads.73.N0 raw_reads.33.raw_reads.73.C1 raw_reads.33.raw_reads.73.N1 raw_reads.33.raw_reads.73.C2 raw_reads.33.raw_reads.73.N2 raw_reads.33.raw_reads.73.C3 raw_reads.33.raw_reads.73.N3 && LAmerge -v L1.33.73 raw_reads.33.raw_reads.73.C0.S raw_reads.33.raw_reads.73.N0.S raw_reads.33.raw_reads.73.C1.S raw_reads.33.raw_reads.73.N1.S raw_reads.33.raw_reads.73.C2.S raw_reads.33.raw_reads.73.N2.S raw_reads.33.raw_reads.73.C3.S raw_reads.33.raw_reads.73.N3.S
LAsort -v raw_reads.34.raw_reads.73.C0 raw_reads.34.raw_reads.73.N0 raw_reads.34.raw_reads.73.C1 raw_reads.34.raw_reads.73.N1 raw_reads.34.raw_reads.73.C2 raw_reads.34.raw_reads.73.N2 raw_reads.34.raw_reads.73.C3 raw_reads.34.raw_reads.73.N3 && LAmerge -v L1.34.73 raw_reads.34.raw_reads.73.C0.S raw_reads.34.raw_reads.73.N0.S raw_reads.34.raw_reads.73.C1.S raw_reads.34.raw_reads.73.N1.S raw_reads.34.raw_reads.73.C2.S raw_reads.34.raw_reads.73.N2.S raw_reads.34.raw_reads.73.C3.S raw_reads.34.raw_reads.73.N3.S
LAsort -v raw_reads.35.raw_reads.73.C0 raw_reads.35.raw_reads.73.N0 raw_reads.35.raw_reads.73.C1 raw_reads.35.raw_reads.73.N1 raw_reads.35.raw_reads.73.C2 raw_reads.35.raw_reads.73.N2 raw_reads.35.raw_reads.73.C3 raw_reads.35.raw_reads.73.N3 && LAmerge -v L1.35.73 raw_reads.35.raw_reads.73.C0.S raw_reads.35.raw_reads.73.N0.S raw_reads.35.raw_reads.73.C1.S raw_reads.35.raw_reads.73.N1.S raw_reads.35.raw_reads.73.C2.S raw_reads.35.raw_reads.73.N2.S raw_reads.35.raw_reads.73.C3.S raw_reads.35.raw_reads.73.N3.S
LAsort -v raw_reads.36.raw_reads.73.C0 raw_reads.36.raw_reads.73.N0 raw_reads.36.raw_reads.73.C1 raw_reads.36.raw_reads.73.N1 raw_reads.36.raw_reads.73.C2 raw_reads.36.raw_reads.73.N2 raw_reads.36.raw_reads.73.C3 raw_reads.36.raw_reads.73.N3 && LAmerge -v L1.36.73 raw_reads.36.raw_reads.73.C0.S raw_reads.36.raw_reads.73.N0.S raw_reads.36.raw_reads.73.C1.S raw_reads.36.raw_reads.73.N1.S raw_reads.36.raw_reads.73.C2.S raw_reads.36.raw_reads.73.N2.S raw_reads.36.raw_reads.73.C3.S raw_reads.36.raw_reads.73.N3.S
LAsort -v raw_reads.73.raw_reads.19.C0 raw_reads.73.raw_reads.19.N0 raw_reads.73.raw_reads.19.C1 raw_reads.73.raw_reads.19.N1 raw_reads.73.raw_reads.19.C2 raw_reads.73.raw_reads.19.N2 raw_reads.73.raw_reads.19.C3 raw_reads.73.raw_reads.19.N3 && LAmerge -v L1.73.19 raw_reads.73.raw_reads.19.C0.S raw_reads.73.raw_reads.19.N0.S raw_reads.73.raw_reads.19.C1.S raw_reads.73.raw_reads.19.N1.S raw_reads.73.raw_reads.19.C2.S raw_reads.73.raw_reads.19.N2.S raw_reads.73.raw_reads.19.C3.S raw_reads.73.raw_reads.19.N3.S
LAsort -v raw_reads.73.raw_reads.20.C0 raw_reads.73.raw_reads.20.N0 raw_reads.73.raw_reads.20.C1 raw_reads.73.raw_reads.20.N1 raw_reads.73.raw_reads.20.C2 raw_reads.73.raw_reads.20.N2 raw_reads.73.raw_reads.20.C3 raw_reads.73.raw_reads.20.N3 && LAmerge -v L1.73.20 raw_reads.73.raw_reads.20.C0.S raw_reads.73.raw_reads.20.N0.S raw_reads.73.raw_reads.20.C1.S raw_reads.73.raw_reads.20.N1.S raw_reads.73.raw_reads.20.C2.S raw_reads.73.raw_reads.20.N2.S raw_reads.73.raw_reads.20.C3.S raw_reads.73.raw_reads.20.N3.S
LAsort -v raw_reads.73.raw_reads.21.C0 raw_reads.73.raw_reads.21.N0 raw_reads.73.raw_reads.21.C1 raw_reads.73.raw_reads.21.N1 raw_reads.73.raw_reads.21.C2 raw_reads.73.raw_reads.21.N2 raw_reads.73.raw_reads.21.C3 raw_reads.73.raw_reads.21.N3 && LAmerge -v L1.73.21 raw_reads.73.raw_reads.21.C0.S raw_reads.73.raw_reads.21.N0.S raw_reads.73.raw_reads.21.C1.S raw_reads.73.raw_reads.21.N1.S raw_reads.73.raw_reads.21.C2.S raw_reads.73.raw_reads.21.N2.S raw_reads.73.raw_reads.21.C3.S raw_reads.73.raw_reads.21.N3.S
LAsort -v raw_reads.73.raw_reads.22.C0 raw_reads.73.raw_reads.22.N0 raw_reads.73.raw_reads.22.C1 raw_reads.73.raw_reads.22.N1 raw_reads.73.raw_reads.22.C2 raw_reads.73.raw_reads.22.N2 raw_reads.73.raw_reads.22.C3 raw_reads.73.raw_reads.22.N3 && LAmerge -v L1.73.22 raw_reads.73.raw_reads.22.C0.S raw_reads.73.raw_reads.22.N0.S raw_reads.73.raw_reads.22.C1.S raw_reads.73.raw_reads.22.N1.S raw_reads.73.raw_reads.22.C2.S raw_reads.73.raw_reads.22.N2.S raw_reads.73.raw_reads.22.C3.S raw_reads.73.raw_reads.22.N3.S
LAsort -v raw_reads.73.raw_reads.23.C0 raw_reads.73.raw_reads.23.N0 raw_reads.73.raw_reads.23.C1 raw_reads.73.raw_reads.23.N1 raw_reads.73.raw_reads.23.C2 raw_reads.73.raw_reads.23.N2 raw_reads.73.raw_reads.23.C3 raw_reads.73.raw_reads.23.N3 && LAmerge -v L1.73.23 raw_reads.73.raw_reads.23.C0.S raw_reads.73.raw_reads.23.N0.S raw_reads.73.raw_reads.23.C1.S raw_reads.73.raw_reads.23.N1.S raw_reads.73.raw_reads.23.C2.S raw_reads.73.raw_reads.23.N2.S raw_reads.73.raw_reads.23.C3.S raw_reads.73.raw_reads.23.N3.S
LAsort -v raw_reads.73.raw_reads.24.C0 raw_reads.73.raw_reads.24.N0 raw_reads.73.raw_reads.24.C1 raw_reads.73.raw_reads.24.N1 raw_reads.73.raw_reads.24.C2 raw_reads.73.raw_reads.24.N2 raw_reads.73.raw_reads.24.C3 raw_reads.73.raw_reads.24.N3 && LAmerge -v L1.73.24 raw_reads.73.raw_reads.24.C0.S raw_reads.73.raw_reads.24.N0.S raw_reads.73.raw_reads.24.C1.S raw_reads.73.raw_reads.24.N1.S raw_reads.73.raw_reads.24.C2.S raw_reads.73.raw_reads.24.N2.S raw_reads.73.raw_reads.24.C3.S raw_reads.73.raw_reads.24.N3.S
LAsort -v raw_reads.73.raw_reads.25.C0 raw_reads.73.raw_reads.25.N0 raw_reads.73.raw_reads.25.C1 raw_reads.73.raw_reads.25.N1 raw_reads.73.raw_reads.25.C2 raw_reads.73.raw_reads.25.N2 raw_reads.73.raw_reads.25.C3 raw_reads.73.raw_reads.25.N3 && LAmerge -v L1.73.25 raw_reads.73.raw_reads.25.C0.S raw_reads.73.raw_reads.25.N0.S raw_reads.73.raw_reads.25.C1.S raw_reads.73.raw_reads.25.N1.S raw_reads.73.raw_reads.25.C2.S raw_reads.73.raw_reads.25.N2.S raw_reads.73.raw_reads.25.C3.S raw_reads.73.raw_reads.25.N3.S
LAsort -v raw_reads.73.raw_reads.26.C0 raw_reads.73.raw_reads.26.N0 raw_reads.73.raw_reads.26.C1 raw_reads.73.raw_reads.26.N1 raw_reads.73.raw_reads.26.C2 raw_reads.73.raw_reads.26.N2 raw_reads.73.raw_reads.26.C3 raw_reads.73.raw_reads.26.N3 && LAmerge -v L1.73.26 raw_reads.73.raw_reads.26.C0.S raw_reads.73.raw_reads.26.N0.S raw_reads.73.raw_reads.26.C1.S raw_reads.73.raw_reads.26.N1.S raw_reads.73.raw_reads.26.C2.S raw_reads.73.raw_reads.26.N2.S raw_reads.73.raw_reads.26.C3.S raw_reads.73.raw_reads.26.N3.S
LAsort -v raw_reads.73.raw_reads.27.C0 raw_reads.73.raw_reads.27.N0 raw_reads.73.raw_reads.27.C1 raw_reads.73.raw_reads.27.N1 raw_reads.73.raw_reads.27.C2 raw_reads.73.raw_reads.27.N2 raw_reads.73.raw_reads.27.C3 raw_reads.73.raw_reads.27.N3 && LAmerge -v L1.73.27 raw_reads.73.raw_reads.27.C0.S raw_reads.73.raw_reads.27.N0.S raw_reads.73.raw_reads.27.C1.S raw_reads.73.raw_reads.27.N1.S raw_reads.73.raw_reads.27.C2.S raw_reads.73.raw_reads.27.N2.S raw_reads.73.raw_reads.27.C3.S raw_reads.73.raw_reads.27.N3.S
LAsort -v raw_reads.73.raw_reads.28.C0 raw_reads.73.raw_reads.28.N0 raw_reads.73.raw_reads.28.C1 raw_reads.73.raw_reads.28.N1 raw_reads.73.raw_reads.28.C2 raw_reads.73.raw_reads.28.N2 raw_reads.73.raw_reads.28.C3 raw_reads.73.raw_reads.28.N3 && LAmerge -v L1.73.28 raw_reads.73.raw_reads.28.C0.S raw_reads.73.raw_reads.28.N0.S raw_reads.73.raw_reads.28.C1.S raw_reads.73.raw_reads.28.N1.S raw_reads.73.raw_reads.28.C2.S raw_reads.73.raw_reads.28.N2.S raw_reads.73.raw_reads.28.C3.S raw_reads.73.raw_reads.28.N3.S
LAsort -v raw_reads.73.raw_reads.29.C0 raw_reads.73.raw_reads.29.N0 raw_reads.73.raw_reads.29.C1 raw_reads.73.raw_reads.29.N1 raw_reads.73.raw_reads.29.C2 raw_reads.73.raw_reads.29.N2 raw_reads.73.raw_reads.29.C3 raw_reads.73.raw_reads.29.N3 && LAmerge -v L1.73.29 raw_reads.73.raw_reads.29.C0.S raw_reads.73.raw_reads.29.N0.S raw_reads.73.raw_reads.29.C1.S raw_reads.73.raw_reads.29.N1.S raw_reads.73.raw_reads.29.C2.S raw_reads.73.raw_reads.29.N2.S raw_reads.73.raw_reads.29.C3.S raw_reads.73.raw_reads.29.N3.S
LAsort -v raw_reads.73.raw_reads.30.C0 raw_reads.73.raw_reads.30.N0 raw_reads.73.raw_reads.30.C1 raw_reads.73.raw_reads.30.N1 raw_reads.73.raw_reads.30.C2 raw_reads.73.raw_reads.30.N2 raw_reads.73.raw_reads.30.C3 raw_reads.73.raw_reads.30.N3 && LAmerge -v L1.73.30 raw_reads.73.raw_reads.30.C0.S raw_reads.73.raw_reads.30.N0.S raw_reads.73.raw_reads.30.C1.S raw_reads.73.raw_reads.30.N1.S raw_reads.73.raw_reads.30.C2.S raw_reads.73.raw_reads.30.N2.S raw_reads.73.raw_reads.30.C3.S raw_reads.73.raw_reads.30.N3.S
LAsort -v raw_reads.73.raw_reads.31.C0 raw_reads.73.raw_reads.31.N0 raw_reads.73.raw_reads.31.C1 raw_reads.73.raw_reads.31.N1 raw_reads.73.raw_reads.31.C2 raw_reads.73.raw_reads.31.N2 raw_reads.73.raw_reads.31.C3 raw_reads.73.raw_reads.31.N3 && LAmerge -v L1.73.31 raw_reads.73.raw_reads.31.C0.S raw_reads.73.raw_reads.31.N0.S raw_reads.73.raw_reads.31.C1.S raw_reads.73.raw_reads.31.N1.S raw_reads.73.raw_reads.31.C2.S raw_reads.73.raw_reads.31.N2.S raw_reads.73.raw_reads.31.C3.S raw_reads.73.raw_reads.31.N3.S
LAsort -v raw_reads.73.raw_reads.32.C0 raw_reads.73.raw_reads.32.N0 raw_reads.73.raw_reads.32.C1 raw_reads.73.raw_reads.32.N1 raw_reads.73.raw_reads.32.C2 raw_reads.73.raw_reads.32.N2 raw_reads.73.raw_reads.32.C3 raw_reads.73.raw_reads.32.N3 && LAmerge -v L1.73.32 raw_reads.73.raw_reads.32.C0.S raw_reads.73.raw_reads.32.N0.S raw_reads.73.raw_reads.32.C1.S raw_reads.73.raw_reads.32.N1.S raw_reads.73.raw_reads.32.C2.S raw_reads.73.raw_reads.32.N2.S raw_reads.73.raw_reads.32.C3.S raw_reads.73.raw_reads.32.N3.S
LAsort -v raw_reads.73.raw_reads.33.C0 raw_reads.73.raw_reads.33.N0 raw_reads.73.raw_reads.33.C1 raw_reads.73.raw_reads.33.N1 raw_reads.73.raw_reads.33.C2 raw_reads.73.raw_reads.33.N2 raw_reads.73.raw_reads.33.C3 raw_reads.73.raw_reads.33.N3 && LAmerge -v L1.73.33 raw_reads.73.raw_reads.33.C0.S raw_reads.73.raw_reads.33.N0.S raw_reads.73.raw_reads.33.C1.S raw_reads.73.raw_reads.33.N1.S raw_reads.73.raw_reads.33.C2.S raw_reads.73.raw_reads.33.N2.S raw_reads.73.raw_reads.33.C3.S raw_reads.73.raw_reads.33.N3.S
LAsort -v raw_reads.73.raw_reads.34.C0 raw_reads.73.raw_reads.34.N0 raw_reads.73.raw_reads.34.C1 raw_reads.73.raw_reads.34.N1 raw_reads.73.raw_reads.34.C2 raw_reads.73.raw_reads.34.N2 raw_reads.73.raw_reads.34.C3 raw_reads.73.raw_reads.34.N3 && LAmerge -v L1.73.34 raw_reads.73.raw_reads.34.C0.S raw_reads.73.raw_reads.34.N0.S raw_reads.73.raw_reads.34.C1.S raw_reads.73.raw_reads.34.N1.S raw_reads.73.raw_reads.34.C2.S raw_reads.73.raw_reads.34.N2.S raw_reads.73.raw_reads.34.C3.S raw_reads.73.raw_reads.34.N3.S
LAsort -v raw_reads.73.raw_reads.35.C0 raw_reads.73.raw_reads.35.N0 raw_reads.73.raw_reads.35.C1 raw_reads.73.raw_reads.35.N1 raw_reads.73.raw_reads.35.C2 raw_reads.73.raw_reads.35.N2 raw_reads.73.raw_reads.35.C3 raw_reads.73.raw_reads.35.N3 && LAmerge -v L1.73.35 raw_reads.73.raw_reads.35.C0.S raw_reads.73.raw_reads.35.N0.S raw_reads.73.raw_reads.35.C1.S raw_reads.73.raw_reads.35.N1.S raw_reads.73.raw_reads.35.C2.S raw_reads.73.raw_reads.35.N2.S raw_reads.73.raw_reads.35.C3.S raw_reads.73.raw_reads.35.N3.S
LAsort -v raw_reads.73.raw_reads.36.C0 raw_reads.73.raw_reads.36.N0 raw_reads.73.raw_reads.36.C1 raw_reads.73.raw_reads.36.N1 raw_reads.73.raw_reads.36.C2 raw_reads.73.raw_reads.36.N2 raw_reads.73.raw_reads.36.C3 raw_reads.73.raw_reads.36.N3 && LAmerge -v L1.73.36 raw_reads.73.raw_reads.36.C0.S raw_reads.73.raw_reads.36.N0.S raw_reads.73.raw_reads.36.C1.S raw_reads.73.raw_reads.36.N1.S raw_reads.73.raw_reads.36.C2.S raw_reads.73.raw_reads.36.N2.S raw_reads.73.raw_reads.36.C3.S raw_reads.73.raw_reads.36.N3.S
LAcheck -vS raw_reads L1.19.73
LAcheck -vS raw_reads L1.20.73
LAcheck -vS raw_reads L1.21.73
LAcheck -vS raw_reads L1.22.73
LAcheck -vS raw_reads L1.23.73
LAcheck -vS raw_reads L1.24.73
LAcheck -vS raw_reads L1.25.73
LAcheck -vS raw_reads L1.26.73
LAcheck -vS raw_reads L1.27.73
LAcheck -vS raw_reads L1.28.73
LAcheck -vS raw_reads L1.29.73
LAcheck -vS raw_reads L1.30.73
LAcheck -vS raw_reads L1.31.73
LAcheck -vS raw_reads L1.32.73
LAcheck -vS raw_reads L1.33.73
LAcheck -vS raw_reads L1.34.73
LAcheck -vS raw_reads L1.35.73
LAcheck -vS raw_reads L1.36.73
LAcheck -vS raw_reads L1.73.19
LAcheck -vS raw_reads L1.73.20
LAcheck -vS raw_reads L1.73.21
LAcheck -vS raw_reads L1.73.22
LAcheck -vS raw_reads L1.73.23
LAcheck -vS raw_reads L1.73.24
LAcheck -vS raw_reads L1.73.25
LAcheck -vS raw_reads L1.73.26
LAcheck -vS raw_reads L1.73.27
LAcheck -vS raw_reads L1.73.28
LAcheck -vS raw_reads L1.73.29
LAcheck -vS raw_reads L1.73.30
LAcheck -vS raw_reads L1.73.31
LAcheck -vS raw_reads L1.73.32
LAcheck -vS raw_reads L1.73.33
LAcheck -vS raw_reads L1.73.34
LAcheck -vS raw_reads L1.73.35
LAcheck -vS raw_reads L1.73.36
daligner -v -k18 -h70 -t14 -H7000 -e0.75 raw_reads.37 raw_reads.19 raw_reads.20 raw_reads.21 raw_reads.22 raw_reads.23 raw_reads.24 raw_reads.25 raw_reads.26 raw_reads.27 raw_reads.28 raw_reads.29 raw_reads.30 raw_reads.31 raw_reads.32 raw_reads.33 raw_reads.34 raw_reads.35 raw_reads.36 raw_reads.37
LAcheck -v raw_reads *.las
LAsort -v raw_reads.19.raw_reads.37.C0 raw_reads.19.raw_reads.37.N0 raw_reads.19.raw_reads.37.C1 raw_reads.19.raw_reads.37.N1 raw_reads.19.raw_reads.37.C2 raw_reads.19.raw_reads.37.N2 raw_reads.19.raw_reads.37.C3 raw_reads.19.raw_reads.37.N3 && LAmerge -v L1.19.37 raw_reads.19.raw_reads.37.C0.S raw_reads.19.raw_reads.37.N0.S raw_reads.19.raw_reads.37.C1.S raw_reads.19.raw_reads.37.N1.S raw_reads.19.raw_reads.37.C2.S raw_reads.19.raw_reads.37.N2.S raw_reads.19.raw_reads.37.C3.S raw_reads.19.raw_reads.37.N3.S
LAsort -v raw_reads.20.raw_reads.37.C0 raw_reads.20.raw_reads.37.N0 raw_reads.20.raw_reads.37.C1 raw_reads.20.raw_reads.37.N1 raw_reads.20.raw_reads.37.C2 raw_reads.20.raw_reads.37.N2 raw_reads.20.raw_reads.37.C3 raw_reads.20.raw_reads.37.N3 && LAmerge -v L1.20.37 raw_reads.20.raw_reads.37.C0.S raw_reads.20.raw_reads.37.N0.S raw_reads.20.raw_reads.37.C1.S raw_reads.20.raw_reads.37.N1.S raw_reads.20.raw_reads.37.C2.S raw_reads.20.raw_reads.37.N2.S raw_reads.20.raw_reads.37.C3.S raw_reads.20.raw_reads.37.N3.S
LAsort -v raw_reads.21.raw_reads.37.C0 raw_reads.21.raw_reads.37.N0 raw_reads.21.raw_reads.37.C1 raw_reads.21.raw_reads.37.N1 raw_reads.21.raw_reads.37.C2 raw_reads.21.raw_reads.37.N2 raw_reads.21.raw_reads.37.C3 raw_reads.21.raw_reads.37.N3 && LAmerge -v L1.21.37 raw_reads.21.raw_reads.37.C0.S raw_reads.21.raw_reads.37.N0.S raw_reads.21.raw_reads.37.C1.S raw_reads.21.raw_reads.37.N1.S raw_reads.21.raw_reads.37.C2.S raw_reads.21.raw_reads.37.N2.S raw_reads.21.raw_reads.37.C3.S raw_reads.21.raw_reads.37.N3.S
LAsort -v raw_reads.22.raw_reads.37.C0 raw_reads.22.raw_reads.37.N0 raw_reads.22.raw_reads.37.C1 raw_reads.22.raw_reads.37.N1 raw_reads.22.raw_reads.37.C2 raw_reads.22.raw_reads.37.N2 raw_reads.22.raw_reads.37.C3 raw_reads.22.raw_reads.37.N3 && LAmerge -v L1.22.37 raw_reads.22.raw_reads.37.C0.S raw_reads.22.raw_reads.37.N0.S raw_reads.22.raw_reads.37.C1.S raw_reads.22.raw_reads.37.N1.S raw_reads.22.raw_reads.37.C2.S raw_reads.22.raw_reads.37.N2.S raw_reads.22.raw_reads.37.C3.S raw_reads.22.raw_reads.37.N3.S
LAsort -v raw_reads.23.raw_reads.37.C0 raw_reads.23.raw_reads.37.N0 raw_reads.23.raw_reads.37.C1 raw_reads.23.raw_reads.37.N1 raw_reads.23.raw_reads.37.C2 raw_reads.23.raw_reads.37.N2 raw_reads.23.raw_reads.37.C3 raw_reads.23.raw_reads.37.N3 && LAmerge -v L1.23.37 raw_reads.23.raw_reads.37.C0.S raw_reads.23.raw_reads.37.N0.S raw_reads.23.raw_reads.37.C1.S raw_reads.23.raw_reads.37.N1.S raw_reads.23.raw_reads.37.C2.S raw_reads.23.raw_reads.37.N2.S raw_reads.23.raw_reads.37.C3.S raw_reads.23.raw_reads.37.N3.S
LAsort -v raw_reads.24.raw_reads.37.C0 raw_reads.24.raw_reads.37.N0 raw_reads.24.raw_reads.37.C1 raw_reads.24.raw_reads.37.N1 raw_reads.24.raw_reads.37.C2 raw_reads.24.raw_reads.37.N2 raw_reads.24.raw_reads.37.C3 raw_reads.24.raw_reads.37.N3 && LAmerge -v L1.24.37 raw_reads.24.raw_reads.37.C0.S raw_reads.24.raw_reads.37.N0.S raw_reads.24.raw_reads.37.C1.S raw_reads.24.raw_reads.37.N1.S raw_reads.24.raw_reads.37.C2.S raw_reads.24.raw_reads.37.N2.S raw_reads.24.raw_reads.37.C3.S raw_reads.24.raw_reads.37.N3.S
LAsort -v raw_reads.25.raw_reads.37.C0 raw_reads.25.raw_reads.37.N0 raw_reads.25.raw_reads.37.C1 raw_reads.25.raw_reads.37.N1 raw_reads.25.raw_reads.37.C2 raw_reads.25.raw_reads.37.N2 raw_reads.25.raw_reads.37.C3 raw_reads.25.raw_reads.37.N3 && LAmerge -v L1.25.37 raw_reads.25.raw_reads.37.C0.S raw_reads.25.raw_reads.37.N0.S raw_reads.25.raw_reads.37.C1.S raw_reads.25.raw_reads.37.N1.S raw_reads.25.raw_reads.37.C2.S raw_reads.25.raw_reads.37.N2.S raw_reads.25.raw_reads.37.C3.S raw_reads.25.raw_reads.37.N3.S
LAsort -v raw_reads.26.raw_reads.37.C0 raw_reads.26.raw_reads.37.N0 raw_reads.26.raw_reads.37.C1 raw_reads.26.raw_reads.37.N1 raw_reads.26.raw_reads.37.C2 raw_reads.26.raw_reads.37.N2 raw_reads.26.raw_reads.37.C3 raw_reads.26.raw_reads.37.N3 && LAmerge -v L1.26.37 raw_reads.26.raw_reads.37.C0.S raw_reads.26.raw_reads.37.N0.S raw_reads.26.raw_reads.37.C1.S raw_reads.26.raw_reads.37.N1.S raw_reads.26.raw_reads.37.C2.S raw_reads.26.raw_reads.37.N2.S raw_reads.26.raw_reads.37.C3.S raw_reads.26.raw_reads.37.N3.S
LAsort -v raw_reads.27.raw_reads.37.C0 raw_reads.27.raw_reads.37.N0 raw_reads.27.raw_reads.37.C1 raw_reads.27.raw_reads.37.N1 raw_reads.27.raw_reads.37.C2 raw_reads.27.raw_reads.37.N2 raw_reads.27.raw_reads.37.C3 raw_reads.27.raw_reads.37.N3 && LAmerge -v L1.27.37 raw_reads.27.raw_reads.37.C0.S raw_reads.27.raw_reads.37.N0.S raw_reads.27.raw_reads.37.C1.S raw_reads.27.raw_reads.37.N1.S raw_reads.27.raw_reads.37.C2.S raw_reads.27.raw_reads.37.N2.S raw_reads.27.raw_reads.37.C3.S raw_reads.27.raw_reads.37.N3.S
LAsort -v raw_reads.28.raw_reads.37.C0 raw_reads.28.raw_reads.37.N0 raw_reads.28.raw_reads.37.C1 raw_reads.28.raw_reads.37.N1 raw_reads.28.raw_reads.37.C2 raw_reads.28.raw_reads.37.N2 raw_reads.28.raw_reads.37.C3 raw_reads.28.raw_reads.37.N3 && LAmerge -v L1.28.37 raw_reads.28.raw_reads.37.C0.S raw_reads.28.raw_reads.37.N0.S raw_reads.28.raw_reads.37.C1.S raw_reads.28.raw_reads.37.N1.S raw_reads.28.raw_reads.37.C2.S raw_reads.28.raw_reads.37.N2.S raw_reads.28.raw_reads.37.C3.S raw_reads.28.raw_reads.37.N3.S
LAsort -v raw_reads.29.raw_reads.37.C0 raw_reads.29.raw_reads.37.N0 raw_reads.29.raw_reads.37.C1 raw_reads.29.raw_reads.37.N1 raw_reads.29.raw_reads.37.C2 raw_reads.29.raw_reads.37.N2 raw_reads.29.raw_reads.37.C3 raw_reads.29.raw_reads.37.N3 && LAmerge -v L1.29.37 raw_reads.29.raw_reads.37.C0.S raw_reads.29.raw_reads.37.N0.S raw_reads.29.raw_reads.37.C1.S raw_reads.29.raw_reads.37.N1.S raw_reads.29.raw_reads.37.C2.S raw_reads.29.raw_reads.37.N2.S raw_reads.29.raw_reads.37.C3.S raw_reads.29.raw_reads.37.N3.S
LAsort -v raw_reads.30.raw_reads.37.C0 raw_reads.30.raw_reads.37.N0 raw_reads.30.raw_reads.37.C1 raw_reads.30.raw_reads.37.N1 raw_reads.30.raw_reads.37.C2 raw_reads.30.raw_reads.37.N2 raw_reads.30.raw_reads.37.C3 raw_reads.30.raw_reads.37.N3 && LAmerge -v L1.30.37 raw_reads.30.raw_reads.37.C0.S raw_reads.30.raw_reads.37.N0.S raw_reads.30.raw_reads.37.C1.S raw_reads.30.raw_reads.37.N1.S raw_reads.30.raw_reads.37.C2.S raw_reads.30.raw_reads.37.N2.S raw_reads.30.raw_reads.37.C3.S raw_reads.30.raw_reads.37.N3.S
LAsort -v raw_reads.31.raw_reads.37.C0 raw_reads.31.raw_reads.37.N0 raw_reads.31.raw_reads.37.C1 raw_reads.31.raw_reads.37.N1 raw_reads.31.raw_reads.37.C2 raw_reads.31.raw_reads.37.N2 raw_reads.31.raw_reads.37.C3 raw_reads.31.raw_reads.37.N3 && LAmerge -v L1.31.37 raw_reads.31.raw_reads.37.C0.S raw_reads.31.raw_reads.37.N0.S raw_reads.31.raw_reads.37.C1.S raw_reads.31.raw_reads.37.N1.S raw_reads.31.raw_reads.37.C2.S raw_reads.31.raw_reads.37.N2.S raw_reads.31.raw_reads.37.C3.S raw_reads.31.raw_reads.37.N3.S
LAsort -v raw_reads.32.raw_reads.37.C0 raw_reads.32.raw_reads.37.N0 raw_reads.32.raw_reads.37.C1 raw_reads.32.raw_reads.37.N1 raw_reads.32.raw_reads.37.C2 raw_reads.32.raw_reads.37.N2 raw_reads.32.raw_reads.37.C3 raw_reads.32.raw_reads.37.N3 && LAmerge -v L1.32.37 raw_reads.32.raw_reads.37.C0.S raw_reads.32.raw_reads.37.N0.S raw_reads.32.raw_reads.37.C1.S raw_reads.32.raw_reads.37.N1.S raw_reads.32.raw_reads.37.C2.S raw_reads.32.raw_reads.37.N2.S raw_reads.32.raw_reads.37.C3.S raw_reads.32.raw_reads.37.N3.S
LAsort -v raw_reads.33.raw_reads.37.C0 raw_reads.33.raw_reads.37.N0 raw_reads.33.raw_reads.37.C1 raw_reads.33.raw_reads.37.N1 raw_reads.33.raw_reads.37.C2 raw_reads.33.raw_reads.37.N2 raw_reads.33.raw_reads.37.C3 raw_reads.33.raw_reads.37.N3 && LAmerge -v L1.33.37 raw_reads.33.raw_reads.37.C0.S raw_reads.33.raw_reads.37.N0.S raw_reads.33.raw_reads.37.C1.S raw_reads.33.raw_reads.37.N1.S raw_reads.33.raw_reads.37.C2.S raw_reads.33.raw_reads.37.N2.S raw_reads.33.raw_reads.37.C3.S raw_reads.33.raw_reads.37.N3.S
LAsort -v raw_reads.34.raw_reads.37.C0 raw_reads.34.raw_reads.37.N0 raw_reads.34.raw_reads.37.C1 raw_reads.34.raw_reads.37.N1 raw_reads.34.raw_reads.37.C2 raw_reads.34.raw_reads.37.N2 raw_reads.34.raw_reads.37.C3 raw_reads.34.raw_reads.37.N3 && LAmerge -v L1.34.37 raw_reads.34.raw_reads.37.C0.S raw_reads.34.raw_reads.37.N0.S raw_reads.34.raw_reads.37.C1.S raw_reads.34.raw_reads.37.N1.S raw_reads.34.raw_reads.37.C2.S raw_reads.34.raw_reads.37.N2.S raw_reads.34.raw_reads.37.C3.S raw_reads.34.raw_reads.37.N3.S
LAsort -v raw_reads.35.raw_reads.37.C0 raw_reads.35.raw_reads.37.N0 raw_reads.35.raw_reads.37.C1 raw_reads.35.raw_reads.37.N1 raw_reads.35.raw_reads.37.C2 raw_reads.35.raw_reads.37.N2 raw_reads.35.raw_reads.37.C3 raw_reads.35.raw_reads.37.N3 && LAmerge -v L1.35.37 raw_reads.35.raw_reads.37.C0.S raw_reads.35.raw_reads.37.N0.S raw_reads.35.raw_reads.37.C1.S raw_reads.35.raw_reads.37.N1.S raw_reads.35.raw_reads.37.C2.S raw_reads.35.raw_reads.37.N2.S raw_reads.35.raw_reads.37.C3.S raw_reads.35.raw_reads.37.N3.S
LAsort -v raw_reads.36.raw_reads.37.C0 raw_reads.36.raw_reads.37.N0 raw_reads.36.raw_reads.37.C1 raw_reads.36.raw_reads.37.N1 raw_reads.36.raw_reads.37.C2 raw_reads.36.raw_reads.37.N2 raw_reads.36.raw_reads.37.C3 raw_reads.36.raw_reads.37.N3 && LAmerge -v L1.36.37 raw_reads.36.raw_reads.37.C0.S raw_reads.36.raw_reads.37.N0.S raw_reads.36.raw_reads.37.C1.S raw_reads.36.raw_reads.37.N1.S raw_reads.36.raw_reads.37.C2.S raw_reads.36.raw_reads.37.N2.S raw_reads.36.raw_reads.37.C3.S raw_reads.36.raw_reads.37.N3.S
LAsort -v raw_reads.37.raw_reads.19.C0 raw_reads.37.raw_reads.19.N0 raw_reads.37.raw_reads.19.C1 raw_reads.37.raw_reads.19.N1 raw_reads.37.raw_reads.19.C2 raw_reads.37.raw_reads.19.N2 raw_reads.37.raw_reads.19.C3 raw_reads.37.raw_reads.19.N3 && LAmerge -v L1.37.19 raw_reads.37.raw_reads.19.C0.S raw_reads.37.raw_reads.19.N0.S raw_reads.37.raw_reads.19.C1.S raw_reads.37.raw_reads.19.N1.S raw_reads.37.raw_reads.19.C2.S raw_reads.37.raw_reads.19.N2.S raw_reads.37.raw_reads.19.C3.S raw_reads.37.raw_reads.19.N3.S
LAsort -v raw_reads.37.raw_reads.20.C0 raw_reads.37.raw_reads.20.N0 raw_reads.37.raw_reads.20.C1 raw_reads.37.raw_reads.20.N1 raw_reads.37.raw_reads.20.C2 raw_reads.37.raw_reads.20.N2 raw_reads.37.raw_reads.20.C3 raw_reads.37.raw_reads.20.N3 && LAmerge -v L1.37.20 raw_reads.37.raw_reads.20.C0.S raw_reads.37.raw_reads.20.N0.S raw_reads.37.raw_reads.20.C1.S raw_reads.37.raw_reads.20.N1.S raw_reads.37.raw_reads.20.C2.S raw_reads.37.raw_reads.20.N2.S raw_reads.37.raw_reads.20.C3.S raw_reads.37.raw_reads.20.N3.S
LAsort -v raw_reads.37.raw_reads.21.C0 raw_reads.37.raw_reads.21.N0 raw_reads.37.raw_reads.21.C1 raw_reads.37.raw_reads.21.N1 raw_reads.37.raw_reads.21.C2 raw_reads.37.raw_reads.21.N2 raw_reads.37.raw_reads.21.C3 raw_reads.37.raw_reads.21.N3 && LAmerge -v L1.37.21 raw_reads.37.raw_reads.21.C0.S raw_reads.37.raw_reads.21.N0.S raw_reads.37.raw_reads.21.C1.S raw_reads.37.raw_reads.21.N1.S raw_reads.37.raw_reads.21.C2.S raw_reads.37.raw_reads.21.N2.S raw_reads.37.raw_reads.21.C3.S raw_reads.37.raw_reads.21.N3.S
LAsort -v raw_reads.37.raw_reads.22.C0 raw_reads.37.raw_reads.22.N0 raw_reads.37.raw_reads.22.C1 raw_reads.37.raw_reads.22.N1 raw_reads.37.raw_reads.22.C2 raw_reads.37.raw_reads.22.N2 raw_reads.37.raw_reads.22.C3 raw_reads.37.raw_reads.22.N3 && LAmerge -v L1.37.22 raw_reads.37.raw_reads.22.C0.S raw_reads.37.raw_reads.22.N0.S raw_reads.37.raw_reads.22.C1.S raw_reads.37.raw_reads.22.N1.S raw_reads.37.raw_reads.22.C2.S raw_reads.37.raw_reads.22.N2.S raw_reads.37.raw_reads.22.C3.S raw_reads.37.raw_reads.22.N3.S
LAsort -v raw_reads.37.raw_reads.23.C0 raw_reads.37.raw_reads.23.N0 raw_reads.37.raw_reads.23.C1 raw_reads.37.raw_reads.23.N1 raw_reads.37.raw_reads.23.C2 raw_reads.37.raw_reads.23.N2 raw_reads.37.raw_reads.23.C3 raw_reads.37.raw_reads.23.N3 && LAmerge -v L1.37.23 raw_reads.37.raw_reads.23.C0.S raw_reads.37.raw_reads.23.N0.S raw_reads.37.raw_reads.23.C1.S raw_reads.37.raw_reads.23.N1.S raw_reads.37.raw_reads.23.C2.S raw_reads.37.raw_reads.23.N2.S raw_reads.37.raw_reads.23.C3.S raw_reads.37.raw_reads.23.N3.S
LAsort -v raw_reads.37.raw_reads.24.C0 raw_reads.37.raw_reads.24.N0 raw_reads.37.raw_reads.24.C1 raw_reads.37.raw_reads.24.N1 raw_reads.37.raw_reads.24.C2 raw_reads.37.raw_reads.24.N2 raw_reads.37.raw_reads.24.C3 raw_reads.37.raw_reads.24.N3 && LAmerge -v L1.37.24 raw_reads.37.raw_reads.24.C0.S raw_reads.37.raw_reads.24.N0.S raw_reads.37.raw_reads.24.C1.S raw_reads.37.raw_reads.24.N1.S raw_reads.37.raw_reads.24.C2.S raw_reads.37.raw_reads.24.N2.S raw_reads.37.raw_reads.24.C3.S raw_reads.37.raw_reads.24.N3.S
LAsort -v raw_reads.37.raw_reads.25.C0 raw_reads.37.raw_reads.25.N0 raw_reads.37.raw_reads.25.C1 raw_reads.37.raw_reads.25.N1 raw_reads.37.raw_reads.25.C2 raw_reads.37.raw_reads.25.N2 raw_reads.37.raw_reads.25.C3 raw_reads.37.raw_reads.25.N3 && LAmerge -v L1.37.25 raw_reads.37.raw_reads.25.C0.S raw_reads.37.raw_reads.25.N0.S raw_reads.37.raw_reads.25.C1.S raw_reads.37.raw_reads.25.N1.S raw_reads.37.raw_reads.25.C2.S raw_reads.37.raw_reads.25.N2.S raw_reads.37.raw_reads.25.C3.S raw_reads.37.raw_reads.25.N3.S
LAsort -v raw_reads.37.raw_reads.26.C0 raw_reads.37.raw_reads.26.N0 raw_reads.37.raw_reads.26.C1 raw_reads.37.raw_reads.26.N1 raw_reads.37.raw_reads.26.C2 raw_reads.37.raw_reads.26.N2 raw_reads.37.raw_reads.26.C3 raw_reads.37.raw_reads.26.N3 && LAmerge -v L1.37.26 raw_reads.37.raw_reads.26.C0.S raw_reads.37.raw_reads.26.N0.S raw_reads.37.raw_reads.26.C1.S raw_reads.37.raw_reads.26.N1.S raw_reads.37.raw_reads.26.C2.S raw_reads.37.raw_reads.26.N2.S raw_reads.37.raw_reads.26.C3.S raw_reads.37.raw_reads.26.N3.S
LAsort -v raw_reads.37.raw_reads.27.C0 raw_reads.37.raw_reads.27.N0 raw_reads.37.raw_reads.27.C1 raw_reads.37.raw_reads.27.N1 raw_reads.37.raw_reads.27.C2 raw_reads.37.raw_reads.27.N2 raw_reads.37.raw_reads.27.C3 raw_reads.37.raw_reads.27.N3 && LAmerge -v L1.37.27 raw_reads.37.raw_reads.27.C0.S raw_reads.37.raw_reads.27.N0.S raw_reads.37.raw_reads.27.C1.S raw_reads.37.raw_reads.27.N1.S raw_reads.37.raw_reads.27.C2.S raw_reads.37.raw_reads.27.N2.S raw_reads.37.raw_reads.27.C3.S raw_reads.37.raw_reads.27.N3.S
LAsort -v raw_reads.37.raw_reads.28.C0 raw_reads.37.raw_reads.28.N0 raw_reads.37.raw_reads.28.C1 raw_reads.37.raw_reads.28.N1 raw_reads.37.raw_reads.28.C2 raw_reads.37.raw_reads.28.N2 raw_reads.37.raw_reads.28.C3 raw_reads.37.raw_reads.28.N3 && LAmerge -v L1.37.28 raw_reads.37.raw_reads.28.C0.S raw_reads.37.raw_reads.28.N0.S raw_reads.37.raw_reads.28.C1.S raw_reads.37.raw_reads.28.N1.S raw_reads.37.raw_reads.28.C2.S raw_reads.37.raw_reads.28.N2.S raw_reads.37.raw_reads.28.C3.S raw_reads.37.raw_reads.28.N3.S
LAsort -v raw_reads.37.raw_reads.29.C0 raw_reads.37.raw_reads.29.N0 raw_reads.37.raw_reads.29.C1 raw_reads.37.raw_reads.29.N1 raw_reads.37.raw_reads.29.C2 raw_reads.37.raw_reads.29.N2 raw_reads.37.raw_reads.29.C3 raw_reads.37.raw_reads.29.N3 && LAmerge -v L1.37.29 raw_reads.37.raw_reads.29.C0.S raw_reads.37.raw_reads.29.N0.S raw_reads.37.raw_reads.29.C1.S raw_reads.37.raw_reads.29.N1.S raw_reads.37.raw_reads.29.C2.S raw_reads.37.raw_reads.29.N2.S raw_reads.37.raw_reads.29.C3.S raw_reads.37.raw_reads.29.N3.S
LAsort -v raw_reads.37.raw_reads.30.C0 raw_reads.37.raw_reads.30.N0 raw_reads.37.raw_reads.30.C1 raw_reads.37.raw_reads.30.N1 raw_reads.37.raw_reads.30.C2 raw_reads.37.raw_reads.30.N2 raw_reads.37.raw_reads.30.C3 raw_reads.37.raw_reads.30.N3 && LAmerge -v L1.37.30 raw_reads.37.raw_reads.30.C0.S raw_reads.37.raw_reads.30.N0.S raw_reads.37.raw_reads.30.C1.S raw_reads.37.raw_reads.30.N1.S raw_reads.37.raw_reads.30.C2.S raw_reads.37.raw_reads.30.N2.S raw_reads.37.raw_reads.30.C3.S raw_reads.37.raw_reads.30.N3.S
LAsort -v raw_reads.37.raw_reads.31.C0 raw_reads.37.raw_reads.31.N0 raw_reads.37.raw_reads.31.C1 raw_reads.37.raw_reads.31.N1 raw_reads.37.raw_reads.31.C2 raw_reads.37.raw_reads.31.N2 raw_reads.37.raw_reads.31.C3 raw_reads.37.raw_reads.31.N3 && LAmerge -v L1.37.31 raw_reads.37.raw_reads.31.C0.S raw_reads.37.raw_reads.31.N0.S raw_reads.37.raw_reads.31.C1.S raw_reads.37.raw_reads.31.N1.S raw_reads.37.raw_reads.31.C2.S raw_reads.37.raw_reads.31.N2.S raw_reads.37.raw_reads.31.C3.S raw_reads.37.raw_reads.31.N3.S
LAsort -v raw_reads.37.raw_reads.32.C0 raw_reads.37.raw_reads.32.N0 raw_reads.37.raw_reads.32.C1 raw_reads.37.raw_reads.32.N1 raw_reads.37.raw_reads.32.C2 raw_reads.37.raw_reads.32.N2 raw_reads.37.raw_reads.32.C3 raw_reads.37.raw_reads.32.N3 && LAmerge -v L1.37.32 raw_reads.37.raw_reads.32.C0.S raw_reads.37.raw_reads.32.N0.S raw_reads.37.raw_reads.32.C1.S raw_reads.37.raw_reads.32.N1.S raw_reads.37.raw_reads.32.C2.S raw_reads.37.raw_reads.32.N2.S raw_reads.37.raw_reads.32.C3.S raw_reads.37.raw_reads.32.N3.S
LAsort -v raw_reads.37.raw_reads.33.C0 raw_reads.37.raw_reads.33.N0 raw_reads.37.raw_reads.33.C1 raw_reads.37.raw_reads.33.N1 raw_reads.37.raw_reads.33.C2 raw_reads.37.raw_reads.33.N2 raw_reads.37.raw_reads.33.C3 raw_reads.37.raw_reads.33.N3 && LAmerge -v L1.37.33 raw_reads.37.raw_reads.33.C0.S raw_reads.37.raw_reads.33.N0.S raw_reads.37.raw_reads.33.C1.S raw_reads.37.raw_reads.33.N1.S raw_reads.37.raw_reads.33.C2.S raw_reads.37.raw_reads.33.N2.S raw_reads.37.raw_reads.33.C3.S raw_reads.37.raw_reads.33.N3.S
LAsort -v raw_reads.37.raw_reads.34.C0 raw_reads.37.raw_reads.34.N0 raw_reads.37.raw_reads.34.C1 raw_reads.37.raw_reads.34.N1 raw_reads.37.raw_reads.34.C2 raw_reads.37.raw_reads.34.N2 raw_reads.37.raw_reads.34.C3 raw_reads.37.raw_reads.34.N3 && LAmerge -v L1.37.34 raw_reads.37.raw_reads.34.C0.S raw_reads.37.raw_reads.34.N0.S raw_reads.37.raw_reads.34.C1.S raw_reads.37.raw_reads.34.N1.S raw_reads.37.raw_reads.34.C2.S raw_reads.37.raw_reads.34.N2.S raw_reads.37.raw_reads.34.C3.S raw_reads.37.raw_reads.34.N3.S
LAsort -v raw_reads.37.raw_reads.35.C0 raw_reads.37.raw_reads.35.N0 raw_reads.37.raw_reads.35.C1 raw_reads.37.raw_reads.35.N1 raw_reads.37.raw_reads.35.C2 raw_reads.37.raw_reads.35.N2 raw_reads.37.raw_reads.35.C3 raw_reads.37.raw_reads.35.N3 && LAmerge -v L1.37.35 raw_reads.37.raw_reads.35.C0.S raw_reads.37.raw_reads.35.N0.S raw_reads.37.raw_reads.35.C1.S raw_reads.37.raw_reads.35.N1.S raw_reads.37.raw_reads.35.C2.S raw_reads.37.raw_reads.35.N2.S raw_reads.37.raw_reads.35.C3.S raw_reads.37.raw_reads.35.N3.S
LAsort -v raw_reads.37.raw_reads.36.C0 raw_reads.37.raw_reads.36.N0 raw_reads.37.raw_reads.36.C1 raw_reads.37.raw_reads.36.N1 raw_reads.37.raw_reads.36.C2 raw_reads.37.raw_reads.36.N2 raw_reads.37.raw_reads.36.C3 raw_reads.37.raw_reads.36.N3 && LAmerge -v L1.37.36 raw_reads.37.raw_reads.36.C0.S raw_reads.37.raw_reads.36.N0.S raw_reads.37.raw_reads.36.C1.S raw_reads.37.raw_reads.36.N1.S raw_reads.37.raw_reads.36.C2.S raw_reads.37.raw_reads.36.N2.S raw_reads.37.raw_reads.36.C3.S raw_reads.37.raw_reads.36.N3.S
LAsort -v raw_reads.37.raw_reads.37.C0 raw_reads.37.raw_reads.37.N0 raw_reads.37.raw_reads.37.C1 raw_reads.37.raw_reads.37.N1 raw_reads.37.raw_reads.37.C2 raw_reads.37.raw_reads.37.N2 raw_reads.37.raw_reads.37.C3 raw_reads.37.raw_reads.37.N3 && LAmerge -v L1.37.37 raw_reads.37.raw_reads.37.C0.S raw_reads.37.raw_reads.37.N0.S raw_reads.37.raw_reads.37.C1.S raw_reads.37.raw_reads.37.N1.S raw_reads.37.raw_reads.37.C2.S raw_reads.37.raw_reads.37.N2.S raw_reads.37.raw_reads.37.C3.S raw_reads.37.raw_reads.37.N3.S
LAcheck -vS raw_reads L1.19.37
LAcheck -vS raw_reads L1.20.37
LAcheck -vS raw_reads L1.21.37
LAcheck -vS raw_reads L1.22.37
LAcheck -vS raw_reads L1.23.37
LAcheck -vS raw_reads L1.24.37
LAcheck -vS raw_reads L1.25.37
LAcheck -vS raw_reads L1.26.37
LAcheck -vS raw_reads L1.27.37
LAcheck -vS raw_reads L1.28.37
LAcheck -vS raw_reads L1.29.37
LAcheck -vS raw_reads L1.30.37
LAcheck -vS raw_reads L1.31.37
LAcheck -vS raw_reads L1.32.37
LAcheck -vS raw_reads L1.33.37
LAcheck -vS raw_reads L1.34.37
LAcheck -vS raw_reads L1.35.37
LAcheck -vS raw_reads L1.36.37
LAcheck -vS raw_reads L1.37.19
LAcheck -vS raw_reads L1.37.20
LAcheck -vS raw_reads L1.37.21
LAcheck -vS raw_reads L1.37.22
LAcheck -vS raw_reads L1.37.23
LAcheck -vS raw_reads L1.37.24
LAcheck -vS raw_reads L1.37.25
LAcheck -vS raw_reads L1.37.26
LAcheck -vS raw_reads L1.37.27
LAcheck -vS raw_reads L1.37.28
LAcheck -vS raw_reads L1.37.29
LAcheck -vS raw_reads L1.37.30
LAcheck -vS raw_reads L1.37.31
LAcheck -vS raw_reads L1.37.32
LAcheck -vS raw_reads L1.37.33
LAcheck -vS raw_reads L1.37.34
LAcheck -vS raw_reads L1.37.35
LAcheck -vS raw_reads L1.37.36
LAcheck -vS raw_reads L1.37.37
daligner -v -k18 -h70 -t14 -H7000 -e0.75 raw_reads.25 raw_reads.1 raw_reads.2 raw_reads.3 raw_reads.4 raw_reads.5 raw_reads.6 raw_reads.7 raw_reads.8 raw_reads.9 raw_reads.10 raw_reads.11 raw_reads.12
LAcheck -v raw_reads *.las
LAsort -v raw_reads.1.raw_reads.25.C0 raw_reads.1.raw_reads.25.N0 raw_reads.1.raw_reads.25.C1 raw_reads.1.raw_reads.25.N1 raw_reads.1.raw_reads.25.C2 raw_reads.1.raw_reads.25.N2 raw_reads.1.raw_reads.25.C3 raw_reads.1.raw_reads.25.N3 && LAmerge -v L1.1.25 raw_reads.1.raw_reads.25.C0.S raw_reads.1.raw_reads.25.N0.S raw_reads.1.raw_reads.25.C1.S raw_reads.1.raw_reads.25.N1.S raw_reads.1.raw_reads.25.C2.S raw_reads.1.raw_reads.25.N2.S raw_reads.1.raw_reads.25.C3.S raw_reads.1.raw_reads.25.N3.S
LAsort -v raw_reads.2.raw_reads.25.C0 raw_reads.2.raw_reads.25.N0 raw_reads.2.raw_reads.25.C1 raw_reads.2.raw_reads.25.N1 raw_reads.2.raw_reads.25.C2 raw_reads.2.raw_reads.25.N2 raw_reads.2.raw_reads.25.C3 raw_reads.2.raw_reads.25.N3 && LAmerge -v L1.2.25 raw_reads.2.raw_reads.25.C0.S raw_reads.2.raw_reads.25.N0.S raw_reads.2.raw_reads.25.C1.S raw_reads.2.raw_reads.25.N1.S raw_reads.2.raw_reads.25.C2.S raw_reads.2.raw_reads.25.N2.S raw_reads.2.raw_reads.25.C3.S raw_reads.2.raw_reads.25.N3.S
LAsort -v raw_reads.3.raw_reads.25.C0 raw_reads.3.raw_reads.25.N0 raw_reads.3.raw_reads.25.C1 raw_reads.3.raw_reads.25.N1 raw_reads.3.raw_reads.25.C2 raw_reads.3.raw_reads.25.N2 raw_reads.3.raw_reads.25.C3 raw_reads.3.raw_reads.25.N3 && LAmerge -v L1.3.25 raw_reads.3.raw_reads.25.C0.S raw_reads.3.raw_reads.25.N0.S raw_reads.3.raw_reads.25.C1.S raw_reads.3.raw_reads.25.N1.S raw_reads.3.raw_reads.25.C2.S raw_reads.3.raw_reads.25.N2.S raw_reads.3.raw_reads.25.C3.S raw_reads.3.raw_reads.25.N3.S
LAsort -v raw_reads.4.raw_reads.25.C0 raw_reads.4.raw_reads.25.N0 raw_reads.4.raw_reads.25.C1 raw_reads.4.raw_reads.25.N1 raw_reads.4.raw_reads.25.C2 raw_reads.4.raw_reads.25.N2 raw_reads.4.raw_reads.25.C3 raw_reads.4.raw_reads.25.N3 && LAmerge -v L1.4.25 raw_reads.4.raw_reads.25.C0.S raw_reads.4.raw_reads.25.N0.S raw_reads.4.raw_reads.25.C1.S raw_reads.4.raw_reads.25.N1.S raw_reads.4.raw_reads.25.C2.S raw_reads.4.raw_reads.25.N2.S raw_reads.4.raw_reads.25.C3.S raw_reads.4.raw_reads.25.N3.S
LAsort -v raw_reads.5.raw_reads.25.C0 raw_reads.5.raw_reads.25.N0 raw_reads.5.raw_reads.25.C1 raw_reads.5.raw_reads.25.N1 raw_reads.5.raw_reads.25.C2 raw_reads.5.raw_reads.25.N2 raw_reads.5.raw_reads.25.C3 raw_reads.5.raw_reads.25.N3 && LAmerge -v L1.5.25 raw_reads.5.raw_reads.25.C0.S raw_reads.5.raw_reads.25.N0.S raw_reads.5.raw_reads.25.C1.S raw_reads.5.raw_reads.25.N1.S raw_reads.5.raw_reads.25.C2.S raw_reads.5.raw_reads.25.N2.S raw_reads.5.raw_reads.25.C3.S raw_reads.5.raw_reads.25.N3.S
LAsort -v raw_reads.6.raw_reads.25.C0 raw_reads.6.raw_reads.25.N0 raw_reads.6.raw_reads.25.C1 raw_reads.6.raw_reads.25.N1 raw_reads.6.raw_reads.25.C2 raw_reads.6.raw_reads.25.N2 raw_reads.6.raw_reads.25.C3 raw_reads.6.raw_reads.25.N3 && LAmerge -v L1.6.25 raw_reads.6.raw_reads.25.C0.S raw_reads.6.raw_reads.25.N0.S raw_reads.6.raw_reads.25.C1.S raw_reads.6.raw_reads.25.N1.S raw_reads.6.raw_reads.25.C2.S raw_reads.6.raw_reads.25.N2.S raw_reads.6.raw_reads.25.C3.S raw_reads.6.raw_reads.25.N3.S
LAsort -v raw_reads.7.raw_reads.25.C0 raw_reads.7.raw_reads.25.N0 raw_reads.7.raw_reads.25.C1 raw_reads.7.raw_reads.25.N1 raw_reads.7.raw_reads.25.C2 raw_reads.7.raw_reads.25.N2 raw_reads.7.raw_reads.25.C3 raw_reads.7.raw_reads.25.N3 && LAmerge -v L1.7.25 raw_reads.7.raw_reads.25.C0.S raw_reads.7.raw_reads.25.N0.S raw_reads.7.raw_reads.25.C1.S raw_reads.7.raw_reads.25.N1.S raw_reads.7.raw_reads.25.C2.S raw_reads.7.raw_reads.25.N2.S raw_reads.7.raw_reads.25.C3.S raw_reads.7.raw_reads.25.N3.S
LAsort -v raw_reads.8.raw_reads.25.C0 raw_reads.8.raw_reads.25.N0 raw_reads.8.raw_reads.25.C1 raw_reads.8.raw_reads.25.N1 raw_reads.8.raw_reads.25.C2 raw_reads.8.raw_reads.25.N2 raw_reads.8.raw_reads.25.C3 raw_reads.8.raw_reads.25.N3 && LAmerge -v L1.8.25 raw_reads.8.raw_reads.25.C0.S raw_reads.8.raw_reads.25.N0.S raw_reads.8.raw_reads.25.C1.S raw_reads.8.raw_reads.25.N1.S raw_reads.8.raw_reads.25.C2.S raw_reads.8.raw_reads.25.N2.S raw_reads.8.raw_reads.25.C3.S raw_reads.8.raw_reads.25.N3.S
LAsort -v raw_reads.9.raw_reads.25.C0 raw_reads.9.raw_reads.25.N0 raw_reads.9.raw_reads.25.C1 raw_reads.9.raw_reads.25.N1 raw_reads.9.raw_reads.25.C2 raw_reads.9.raw_reads.25.N2 raw_reads.9.raw_reads.25.C3 raw_reads.9.raw_reads.25.N3 && LAmerge -v L1.9.25 raw_reads.9.raw_reads.25.C0.S raw_reads.9.raw_reads.25.N0.S raw_reads.9.raw_reads.25.C1.S raw_reads.9.raw_reads.25.N1.S raw_reads.9.raw_reads.25.C2.S raw_reads.9.raw_reads.25.N2.S raw_reads.9.raw_reads.25.C3.S raw_reads.9.raw_reads.25.N3.S
LAsort -v raw_reads.10.raw_reads.25.C0 raw_reads.10.raw_reads.25.N0 raw_reads.10.raw_reads.25.C1 raw_reads.10.raw_reads.25.N1 raw_reads.10.raw_reads.25.C2 raw_reads.10.raw_reads.25.N2 raw_reads.10.raw_reads.25.C3 raw_reads.10.raw_reads.25.N3 && LAmerge -v L1.10.25 raw_reads.10.raw_reads.25.C0.S raw_reads.10.raw_reads.25.N0.S raw_reads.10.raw_reads.25.C1.S raw_reads.10.raw_reads.25.N1.S raw_reads.10.raw_reads.25.C2.S raw_reads.10.raw_reads.25.N2.S raw_reads.10.raw_reads.25.C3.S raw_reads.10.raw_reads.25.N3.S
LAsort -v raw_reads.11.raw_reads.25.C0 raw_reads.11.raw_reads.25.N0 raw_reads.11.raw_reads.25.C1 raw_reads.11.raw_reads.25.N1 raw_reads.11.raw_reads.25.C2 raw_reads.11.raw_reads.25.N2 raw_reads.11.raw_reads.25.C3 raw_reads.11.raw_reads.25.N3 && LAmerge -v L1.11.25 raw_reads.11.raw_reads.25.C0.S raw_reads.11.raw_reads.25.N0.S raw_reads.11.raw_reads.25.C1.S raw_reads.11.raw_reads.25.N1.S raw_reads.11.raw_reads.25.C2.S raw_reads.11.raw_reads.25.N2.S raw_reads.11.raw_reads.25.C3.S raw_reads.11.raw_reads.25.N3.S
LAsort -v raw_reads.12.raw_reads.25.C0 raw_reads.12.raw_reads.25.N0 raw_reads.12.raw_reads.25.C1 raw_reads.12.raw_reads.25.N1 raw_reads.12.raw_reads.25.C2 raw_reads.12.raw_reads.25.N2 raw_reads.12.raw_reads.25.C3 raw_reads.12.raw_reads.25.N3 && LAmerge -v L1.12.25 raw_reads.12.raw_reads.25.C0.S raw_reads.12.raw_reads.25.N0.S raw_reads.12.raw_reads.25.C1.S raw_reads.12.raw_reads.25.N1.S raw_reads.12.raw_reads.25.C2.S raw_reads.12.raw_reads.25.N2.S raw_reads.12.raw_reads.25.C3.S raw_reads.12.raw_reads.25.N3.S
LAsort -v raw_reads.25.raw_reads.1.C0 raw_reads.25.raw_reads.1.N0 raw_reads.25.raw_reads.1.C1 raw_reads.25.raw_reads.1.N1 raw_reads.25.raw_reads.1.C2 raw_reads.25.raw_reads.1.N2 raw_reads.25.raw_reads.1.C3 raw_reads.25.raw_reads.1.N3 && LAmerge -v L1.25.1 raw_reads.25.raw_reads.1.C0.S raw_reads.25.raw_reads.1.N0.S raw_reads.25.raw_reads.1.C1.S raw_reads.25.raw_reads.1.N1.S raw_reads.25.raw_reads.1.C2.S raw_reads.25.raw_reads.1.N2.S raw_reads.25.raw_reads.1.C3.S raw_reads.25.raw_reads.1.N3.S
LAsort -v raw_reads.25.raw_reads.2.C0 raw_reads.25.raw_reads.2.N0 raw_reads.25.raw_reads.2.C1 raw_reads.25.raw_reads.2.N1 raw_reads.25.raw_reads.2.C2 raw_reads.25.raw_reads.2.N2 raw_reads.25.raw_reads.2.C3 raw_reads.25.raw_reads.2.N3 && LAmerge -v L1.25.2 raw_reads.25.raw_reads.2.C0.S raw_reads.25.raw_reads.2.N0.S raw_reads.25.raw_reads.2.C1.S raw_reads.25.raw_reads.2.N1.S raw_reads.25.raw_reads.2.C2.S raw_reads.25.raw_reads.2.N2.S raw_reads.25.raw_reads.2.C3.S raw_reads.25.raw_reads.2.N3.S
LAsort -v raw_reads.25.raw_reads.3.C0 raw_reads.25.raw_reads.3.N0 raw_reads.25.raw_reads.3.C1 raw_reads.25.raw_reads.3.N1 raw_reads.25.raw_reads.3.C2 raw_reads.25.raw_reads.3.N2 raw_reads.25.raw_reads.3.C3 raw_reads.25.raw_reads.3.N3 && LAmerge -v L1.25.3 raw_reads.25.raw_reads.3.C0.S raw_reads.25.raw_reads.3.N0.S raw_reads.25.raw_reads.3.C1.S raw_reads.25.raw_reads.3.N1.S raw_reads.25.raw_reads.3.C2.S raw_reads.25.raw_reads.3.N2.S raw_reads.25.raw_reads.3.C3.S raw_reads.25.raw_reads.3.N3.S
LAsort -v raw_reads.25.raw_reads.4.C0 raw_reads.25.raw_reads.4.N0 raw_reads.25.raw_reads.4.C1 raw_reads.25.raw_reads.4.N1 raw_reads.25.raw_reads.4.C2 raw_reads.25.raw_reads.4.N2 raw_reads.25.raw_reads.4.C3 raw_reads.25.raw_reads.4.N3 && LAmerge -v L1.25.4 raw_reads.25.raw_reads.4.C0.S raw_reads.25.raw_reads.4.N0.S raw_reads.25.raw_reads.4.C1.S raw_reads.25.raw_reads.4.N1.S raw_reads.25.raw_reads.4.C2.S raw_reads.25.raw_reads.4.N2.S raw_reads.25.raw_reads.4.C3.S raw_reads.25.raw_reads.4.N3.S
LAsort -v raw_reads.25.raw_reads.5.C0 raw_reads.25.raw_reads.5.N0 raw_reads.25.raw_reads.5.C1 raw_reads.25.raw_reads.5.N1 raw_reads.25.raw_reads.5.C2 raw_reads.25.raw_reads.5.N2 raw_reads.25.raw_reads.5.C3 raw_reads.25.raw_reads.5.N3 && LAmerge -v L1.25.5 raw_reads.25.raw_reads.5.C0.S raw_reads.25.raw_reads.5.N0.S raw_reads.25.raw_reads.5.C1.S raw_reads.25.raw_reads.5.N1.S raw_reads.25.raw_reads.5.C2.S raw_reads.25.raw_reads.5.N2.S raw_reads.25.raw_reads.5.C3.S raw_reads.25.raw_reads.5.N3.S
LAsort -v raw_reads.25.raw_reads.6.C0 raw_reads.25.raw_reads.6.N0 raw_reads.25.raw_reads.6.C1 raw_reads.25.raw_reads.6.N1 raw_reads.25.raw_reads.6.C2 raw_reads.25.raw_reads.6.N2 raw_reads.25.raw_reads.6.C3 raw_reads.25.raw_reads.6.N3 && LAmerge -v L1.25.6 raw_reads.25.raw_reads.6.C0.S raw_reads.25.raw_reads.6.N0.S raw_reads.25.raw_reads.6.C1.S raw_reads.25.raw_reads.6.N1.S raw_reads.25.raw_reads.6.C2.S raw_reads.25.raw_reads.6.N2.S raw_reads.25.raw_reads.6.C3.S raw_reads.25.raw_reads.6.N3.S
LAsort -v raw_reads.25.raw_reads.7.C0 raw_reads.25.raw_reads.7.N0 raw_reads.25.raw_reads.7.C1 raw_reads.25.raw_reads.7.N1 raw_reads.25.raw_reads.7.C2 raw_reads.25.raw_reads.7.N2 raw_reads.25.raw_reads.7.C3 raw_reads.25.raw_reads.7.N3 && LAmerge -v L1.25.7 raw_reads.25.raw_reads.7.C0.S raw_reads.25.raw_reads.7.N0.S raw_reads.25.raw_reads.7.C1.S raw_reads.25.raw_reads.7.N1.S raw_reads.25.raw_reads.7.C2.S raw_reads.25.raw_reads.7.N2.S raw_reads.25.raw_reads.7.C3.S raw_reads.25.raw_reads.7.N3.S
LAsort -v raw_reads.25.raw_reads.8.C0 raw_reads.25.raw_reads.8.N0 raw_reads.25.raw_reads.8.C1 raw_reads.25.raw_reads.8.N1 raw_reads.25.raw_reads.8.C2 raw_reads.25.raw_reads.8.N2 raw_reads.25.raw_reads.8.C3 raw_reads.25.raw_reads.8.N3 && LAmerge -v L1.25.8 raw_reads.25.raw_reads.8.C0.S raw_reads.25.raw_reads.8.N0.S raw_reads.25.raw_reads.8.C1.S raw_reads.25.raw_reads.8.N1.S raw_reads.25.raw_reads.8.C2.S raw_reads.25.raw_reads.8.N2.S raw_reads.25.raw_reads.8.C3.S raw_reads.25.raw_reads.8.N3.S
LAsort -v raw_reads.25.raw_reads.9.C0 raw_reads.25.raw_reads.9.N0 raw_reads.25.raw_reads.9.C1 raw_reads.25.raw_reads.9.N1 raw_reads.25.raw_reads.9.C2 raw_reads.25.raw_reads.9.N2 raw_reads.25.raw_reads.9.C3 raw_reads.25.raw_reads.9.N3 && LAmerge -v L1.25.9 raw_reads.25.raw_reads.9.C0.S raw_reads.25.raw_reads.9.N0.S raw_reads.25.raw_reads.9.C1.S raw_reads.25.raw_reads.9.N1.S raw_reads.25.raw_reads.9.C2.S raw_reads.25.raw_reads.9.N2.S raw_reads.25.raw_reads.9.C3.S raw_reads.25.raw_reads.9.N3.S
LAsort -v raw_reads.25.raw_reads.10.C0 raw_reads.25.raw_reads.10.N0 raw_reads.25.raw_reads.10.C1 raw_reads.25.raw_reads.10.N1 raw_reads.25.raw_reads.10.C2 raw_reads.25.raw_reads.10.N2 raw_reads.25.raw_reads.10.C3 raw_reads.25.raw_reads.10.N3 && LAmerge -v L1.25.10 raw_reads.25.raw_reads.10.C0.S raw_reads.25.raw_reads.10.N0.S raw_reads.25.raw_reads.10.C1.S raw_reads.25.raw_reads.10.N1.S raw_reads.25.raw_reads.10.C2.S raw_reads.25.raw_reads.10.N2.S raw_reads.25.raw_reads.10.C3.S raw_reads.25.raw_reads.10.N3.S
LAsort -v raw_reads.25.raw_reads.11.C0 raw_reads.25.raw_reads.11.N0 raw_reads.25.raw_reads.11.C1 raw_reads.25.raw_reads.11.N1 raw_reads.25.raw_reads.11.C2 raw_reads.25.raw_reads.11.N2 raw_reads.25.raw_reads.11.C3 raw_reads.25.raw_reads.11.N3 && LAmerge -v L1.25.11 raw_reads.25.raw_reads.11.C0.S raw_reads.25.raw_reads.11.N0.S raw_reads.25.raw_reads.11.C1.S raw_reads.25.raw_reads.11.N1.S raw_reads.25.raw_reads.11.C2.S raw_reads.25.raw_reads.11.N2.S raw_reads.25.raw_reads.11.C3.S raw_reads.25.raw_reads.11.N3.S
LAsort -v raw_reads.25.raw_reads.12.C0 raw_reads.25.raw_reads.12.N0 raw_reads.25.raw_reads.12.C1 raw_reads.25.raw_reads.12.N1 raw_reads.25.raw_reads.12.C2 raw_reads.25.raw_reads.12.N2 raw_reads.25.raw_reads.12.C3 raw_reads.25.raw_reads.12.N3 && LAmerge -v L1.25.12 raw_reads.25.raw_reads.12.C0.S raw_reads.25.raw_reads.12.N0.S raw_reads.25.raw_reads.12.C1.S raw_reads.25.raw_reads.12.N1.S raw_reads.25.raw_reads.12.C2.S raw_reads.25.raw_reads.12.N2.S raw_reads.25.raw_reads.12.C3.S raw_reads.25.raw_reads.12.N3.S
LAcheck -vS raw_reads L1.1.25
LAcheck -vS raw_reads L1.2.25
LAcheck -vS raw_reads L1.3.25
LAcheck -vS raw_reads L1.4.25
LAcheck -vS raw_reads L1.5.25
LAcheck -vS raw_reads L1.6.25
LAcheck -vS raw_reads L1.7.25
LAcheck -vS raw_reads L1.8.25
LAcheck -vS raw_reads L1.9.25
LAcheck -vS raw_reads L1.10.25
LAcheck -vS raw_reads L1.11.25
LAcheck -vS raw_reads L1.12.25
LAcheck -vS raw_reads L1.25.1
LAcheck -vS raw_reads L1.25.2
LAcheck -vS raw_reads L1.25.3
LAcheck -vS raw_reads L1.25.4
LAcheck -vS raw_reads L1.25.5
LAcheck -vS raw_reads L1.25.6
LAcheck -vS raw_reads L1.25.7
LAcheck -vS raw_reads L1.25.8
LAcheck -vS raw_reads L1.25.9
LAcheck -vS raw_reads L1.25.10
LAcheck -vS raw_reads L1.25.11
LAcheck -vS raw_reads L1.25.12
daligner -v -k18 -h70 -t14 -H7000 -e0.75 raw_reads.99 raw_reads.60 raw_reads.61 raw_reads.62 raw_reads.63 raw_reads.64 raw_reads.65 raw_reads.66 raw_reads.67 raw_reads.68 raw_reads.69 raw_reads.70 raw_reads.71 raw_reads.72 raw_reads.73 raw_reads.74 raw_reads.75 raw_reads.76 raw_reads.77 raw_reads.78 raw_reads.79
LAcheck -v raw_reads *.las
LAsort -v raw_reads.60.raw_reads.99.C0 raw_reads.60.raw_reads.99.N0 raw_reads.60.raw_reads.99.C1 raw_reads.60.raw_reads.99.N1 raw_reads.60.raw_reads.99.C2 raw_reads.60.raw_reads.99.N2 raw_reads.60.raw_reads.99.C3 raw_reads.60.raw_reads.99.N3 && LAmerge -v L1.60.99 raw_reads.60.raw_reads.99.C0.S raw_reads.60.raw_reads.99.N0.S raw_reads.60.raw_reads.99.C1.S raw_reads.60.raw_reads.99.N1.S raw_reads.60.raw_reads.99.C2.S raw_reads.60.raw_reads.99.N2.S raw_reads.60.raw_reads.99.C3.S raw_reads.60.raw_reads.99.N3.S
LAsort -v raw_reads.61.raw_reads.99.C0 raw_reads.61.raw_reads.99.N0 raw_reads.61.raw_reads.99.C1 raw_reads.61.raw_reads.99.N1 raw_reads.61.raw_reads.99.C2 raw_reads.61.raw_reads.99.N2 raw_reads.61.raw_reads.99.C3 raw_reads.61.raw_reads.99.N3 && LAmerge -v L1.61.99 raw_reads.61.raw_reads.99.C0.S raw_reads.61.raw_reads.99.N0.S raw_reads.61.raw_reads.99.C1.S raw_reads.61.raw_reads.99.N1.S raw_reads.61.raw_reads.99.C2.S raw_reads.61.raw_reads.99.N2.S raw_reads.61.raw_reads.99.C3.S raw_reads.61.raw_reads.99.N3.S
LAsort -v raw_reads.62.raw_reads.99.C0 raw_reads.62.raw_reads.99.N0 raw_reads.62.raw_reads.99.C1 raw_reads.62.raw_reads.99.N1 raw_reads.62.raw_reads.99.C2 raw_reads.62.raw_reads.99.N2 raw_reads.62.raw_reads.99.C3 raw_reads.62.raw_reads.99.N3 && LAmerge -v L1.62.99 raw_reads.62.raw_reads.99.C0.S raw_reads.62.raw_reads.99.N0.S raw_reads.62.raw_reads.99.C1.S raw_reads.62.raw_reads.99.N1.S raw_reads.62.raw_reads.99.C2.S raw_reads.62.raw_reads.99.N2.S raw_reads.62.raw_reads.99.C3.S raw_reads.62.raw_reads.99.N3.S
LAsort -v raw_reads.63.raw_reads.99.C0 raw_reads.63.raw_reads.99.N0 raw_reads.63.raw_reads.99.C1 raw_reads.63.raw_reads.99.N1 raw_reads.63.raw_reads.99.C2 raw_reads.63.raw_reads.99.N2 raw_reads.63.raw_reads.99.C3 raw_reads.63.raw_reads.99.N3 && LAmerge -v L1.63.99 raw_reads.63.raw_reads.99.C0.S raw_reads.63.raw_reads.99.N0.S raw_reads.63.raw_reads.99.C1.S raw_reads.63.raw_reads.99.N1.S raw_reads.63.raw_reads.99.C2.S raw_reads.63.raw_reads.99.N2.S raw_reads.63.raw_reads.99.C3.S raw_reads.63.raw_reads.99.N3.S
LAsort -v raw_reads.64.raw_reads.99.C0 raw_reads.64.raw_reads.99.N0 raw_reads.64.raw_reads.99.C1 raw_reads.64.raw_reads.99.N1 raw_reads.64.raw_reads.99.C2 raw_reads.64.raw_reads.99.N2 raw_reads.64.raw_reads.99.C3 raw_reads.64.raw_reads.99.N3 && LAmerge -v L1.64.99 raw_reads.64.raw_reads.99.C0.S raw_reads.64.raw_reads.99.N0.S raw_reads.64.raw_reads.99.C1.S raw_reads.64.raw_reads.99.N1.S raw_reads.64.raw_reads.99.C2.S raw_reads.64.raw_reads.99.N2.S raw_reads.64.raw_reads.99.C3.S raw_reads.64.raw_reads.99.N3.S
LAsort -v raw_reads.65.raw_reads.99.C0 raw_reads.65.raw_reads.99.N0 raw_reads.65.raw_reads.99.C1 raw_reads.65.raw_reads.99.N1 raw_reads.65.raw_reads.99.C2 raw_reads.65.raw_reads.99.N2 raw_reads.65.raw_reads.99.C3 raw_reads.65.raw_reads.99.N3 && LAmerge -v L1.65.99 raw_reads.65.raw_reads.99.C0.S raw_reads.65.raw_reads.99.N0.S raw_reads.65.raw_reads.99.C1.S raw_reads.65.raw_reads.99.N1.S raw_reads.65.raw_reads.99.C2.S raw_reads.65.raw_reads.99.N2.S raw_reads.65.raw_reads.99.C3.S raw_reads.65.raw_reads.99.N3.S
LAsort -v raw_reads.66.raw_reads.99.C0 raw_reads.66.raw_reads.99.N0 raw_reads.66.raw_reads.99.C1 raw_reads.66.raw_reads.99.N1 raw_reads.66.raw_reads.99.C2 raw_reads.66.raw_reads.99.N2 raw_reads.66.raw_reads.99.C3 raw_reads.66.raw_reads.99.N3 && LAmerge -v L1.66.99 raw_reads.66.raw_reads.99.C0.S raw_reads.66.raw_reads.99.N0.S raw_reads.66.raw_reads.99.C1.S raw_reads.66.raw_reads.99.N1.S raw_reads.66.raw_reads.99.C2.S raw_reads.66.raw_reads.99.N2.S raw_reads.66.raw_reads.99.C3.S raw_reads.66.raw_reads.99.N3.S
LAsort -v raw_reads.67.raw_reads.99.C0 raw_reads.67.raw_reads.99.N0 raw_reads.67.raw_reads.99.C1 raw_reads.67.raw_reads.99.N1 raw_reads.67.raw_reads.99.C2 raw_reads.67.raw_reads.99.N2 raw_reads.67.raw_reads.99.C3 raw_reads.67.raw_reads.99.N3 && LAmerge -v L1.67.99 raw_reads.67.raw_reads.99.C0.S raw_reads.67.raw_reads.99.N0.S raw_reads.67.raw_reads.99.C1.S raw_reads.67.raw_reads.99.N1.S raw_reads.67.raw_reads.99.C2.S raw_reads.67.raw_reads.99.N2.S raw_reads.67.raw_reads.99.C3.S raw_reads.67.raw_reads.99.N3.S
LAsort -v raw_reads.68.raw_reads.99.C0 raw_reads.68.raw_reads.99.N0 raw_reads.68.raw_reads.99.C1 raw_reads.68.raw_reads.99.N1 raw_reads.68.raw_reads.99.C2 raw_reads.68.raw_reads.99.N2 raw_reads.68.raw_reads.99.C3 raw_reads.68.raw_reads.99.N3 && LAmerge -v L1.68.99 raw_reads.68.raw_reads.99.C0.S raw_reads.68.raw_reads.99.N0.S raw_reads.68.raw_reads.99.C1.S raw_reads.68.raw_reads.99.N1.S raw_reads.68.raw_reads.99.C2.S raw_reads.68.raw_reads.99.N2.S raw_reads.68.raw_reads.99.C3.S raw_reads.68.raw_reads.99.N3.S
LAsort -v raw_reads.69.raw_reads.99.C0 raw_reads.69.raw_reads.99.N0 raw_reads.69.raw_reads.99.C1 raw_reads.69.raw_reads.99.N1 raw_reads.69.raw_reads.99.C2 raw_reads.69.raw_reads.99.N2 raw_reads.69.raw_reads.99.C3 raw_reads.69.raw_reads.99.N3 && LAmerge -v L1.69.99 raw_reads.69.raw_reads.99.C0.S raw_reads.69.raw_reads.99.N0.S raw_reads.69.raw_reads.99.C1.S raw_reads.69.raw_reads.99.N1.S raw_reads.69.raw_reads.99.C2.S raw_reads.69.raw_reads.99.N2.S raw_reads.69.raw_reads.99.C3.S raw_reads.69.raw_reads.99.N3.S
LAsort -v raw_reads.70.raw_reads.99.C0 raw_reads.70.raw_reads.99.N0 raw_reads.70.raw_reads.99.C1 raw_reads.70.raw_reads.99.N1 raw_reads.70.raw_reads.99.C2 raw_reads.70.raw_reads.99.N2 raw_reads.70.raw_reads.99.C3 raw_reads.70.raw_reads.99.N3 && LAmerge -v L1.70.99 raw_reads.70.raw_reads.99.C0.S raw_reads.70.raw_reads.99.N0.S raw_reads.70.raw_reads.99.C1.S raw_reads.70.raw_reads.99.N1.S raw_reads.70.raw_reads.99.C2.S raw_reads.70.raw_reads.99.N2.S raw_reads.70.raw_reads.99.C3.S raw_reads.70.raw_reads.99.N3.S
LAsort -v raw_reads.71.raw_reads.99.C0 raw_reads.71.raw_reads.99.N0 raw_reads.71.raw_reads.99.C1 raw_reads.71.raw_reads.99.N1 raw_reads.71.raw_reads.99.C2 raw_reads.71.raw_reads.99.N2 raw_reads.71.raw_reads.99.C3 raw_reads.71.raw_reads.99.N3 && LAmerge -v L1.71.99 raw_reads.71.raw_reads.99.C0.S raw_reads.71.raw_reads.99.N0.S raw_reads.71.raw_reads.99.C1.S raw_reads.71.raw_reads.99.N1.S raw_reads.71.raw_reads.99.C2.S raw_reads.71.raw_reads.99.N2.S raw_reads.71.raw_reads.99.C3.S raw_reads.71.raw_reads.99.N3.S
LAsort -v raw_reads.72.raw_reads.99.C0 raw_reads.72.raw_reads.99.N0 raw_reads.72.raw_reads.99.C1 raw_reads.72.raw_reads.99.N1 raw_reads.72.raw_reads.99.C2 raw_reads.72.raw_reads.99.N2 raw_reads.72.raw_reads.99.C3 raw_reads.72.raw_reads.99.N3 && LAmerge -v L1.72.99 raw_reads.72.raw_reads.99.C0.S raw_reads.72.raw_reads.99.N0.S raw_reads.72.raw_reads.99.C1.S raw_reads.72.raw_reads.99.N1.S raw_reads.72.raw_reads.99.C2.S raw_reads.72.raw_reads.99.N2.S raw_reads.72.raw_reads.99.C3.S raw_reads.72.raw_reads.99.N3.S
LAsort -v raw_reads.73.raw_reads.99.C0 raw_reads.73.raw_reads.99.N0 raw_reads.73.raw_reads.99.C1 raw_reads.73.raw_reads.99.N1 raw_reads.73.raw_reads.99.C2 raw_reads.73.raw_reads.99.N2 raw_reads.73.raw_reads.99.C3 raw_reads.73.raw_reads.99.N3 && LAmerge -v L1.73.99 raw_reads.73.raw_reads.99.C0.S raw_reads.73.raw_reads.99.N0.S raw_reads.73.raw_reads.99.C1.S raw_reads.73.raw_reads.99.N1.S raw_reads.73.raw_reads.99.C2.S raw_reads.73.raw_reads.99.N2.S raw_reads.73.raw_reads.99.C3.S raw_reads.73.raw_reads.99.N3.S
LAsort -v raw_reads.74.raw_reads.99.C0 raw_reads.74.raw_reads.99.N0 raw_reads.74.raw_reads.99.C1 raw_reads.74.raw_reads.99.N1 raw_reads.74.raw_reads.99.C2 raw_reads.74.raw_reads.99.N2 raw_reads.74.raw_reads.99.C3 raw_reads.74.raw_reads.99.N3 && LAmerge -v L1.74.99 raw_reads.74.raw_reads.99.C0.S raw_reads.74.raw_reads.99.N0.S raw_reads.74.raw_reads.99.C1.S raw_reads.74.raw_reads.99.N1.S raw_reads.74.raw_reads.99.C2.S raw_reads.74.raw_reads.99.N2.S raw_reads.74.raw_reads.99.C3.S raw_reads.74.raw_reads.99.N3.S
LAsort -v raw_reads.75.raw_reads.99.C0 raw_reads.75.raw_reads.99.N0 raw_reads.75.raw_reads.99.C1 raw_reads.75.raw_reads.99.N1 raw_reads.75.raw_reads.99.C2 raw_reads.75.raw_reads.99.N2 raw_reads.75.raw_reads.99.C3 raw_reads.75.raw_reads.99.N3 && LAmerge -v L1.75.99 raw_reads.75.raw_reads.99.C0.S raw_reads.75.raw_reads.99.N0.S raw_reads.75.raw_reads.99.C1.S raw_reads.75.raw_reads.99.N1.S raw_reads.75.raw_reads.99.C2.S raw_reads.75.raw_reads.99.N2.S raw_reads.75.raw_reads.99.C3.S raw_reads.75.raw_reads.99.N3.S
LAsort -v raw_reads.76.raw_reads.99.C0 raw_reads.76.raw_reads.99.N0 raw_reads.76.raw_reads.99.C1 raw_reads.76.raw_reads.99.N1 raw_reads.76.raw_reads.99.C2 raw_reads.76.raw_reads.99.N2 raw_reads.76.raw_reads.99.C3 raw_reads.76.raw_reads.99.N3 && LAmerge -v L1.76.99 raw_reads.76.raw_reads.99.C0.S raw_reads.76.raw_reads.99.N0.S raw_reads.76.raw_reads.99.C1.S raw_reads.76.raw_reads.99.N1.S raw_reads.76.raw_reads.99.C2.S raw_reads.76.raw_reads.99.N2.S raw_reads.76.raw_reads.99.C3.S raw_reads.76.raw_reads.99.N3.S
LAsort -v raw_reads.77.raw_reads.99.C0 raw_reads.77.raw_reads.99.N0 raw_reads.77.raw_reads.99.C1 raw_reads.77.raw_reads.99.N1 raw_reads.77.raw_reads.99.C2 raw_reads.77.raw_reads.99.N2 raw_reads.77.raw_reads.99.C3 raw_reads.77.raw_reads.99.N3 && LAmerge -v L1.77.99 raw_reads.77.raw_reads.99.C0.S raw_reads.77.raw_reads.99.N0.S raw_reads.77.raw_reads.99.C1.S raw_reads.77.raw_reads.99.N1.S raw_reads.77.raw_reads.99.C2.S raw_reads.77.raw_reads.99.N2.S raw_reads.77.raw_reads.99.C3.S raw_reads.77.raw_reads.99.N3.S
LAsort -v raw_reads.78.raw_reads.99.C0 raw_reads.78.raw_reads.99.N0 raw_reads.78.raw_reads.99.C1 raw_reads.78.raw_reads.99.N1 raw_reads.78.raw_reads.99.C2 raw_reads.78.raw_reads.99.N2 raw_reads.78.raw_reads.99.C3 raw_reads.78.raw_reads.99.N3 && LAmerge -v L1.78.99 raw_reads.78.raw_reads.99.C0.S raw_reads.78.raw_reads.99.N0.S raw_reads.78.raw_reads.99.C1.S raw_reads.78.raw_reads.99.N1.S raw_reads.78.raw_reads.99.C2.S raw_reads.78.raw_reads.99.N2.S raw_reads.78.raw_reads.99.C3.S raw_reads.78.raw_reads.99.N3.S
LAsort -v raw_reads.79.raw_reads.99.C0 raw_reads.79.raw_reads.99.N0 raw_reads.79.raw_reads.99.C1 raw_reads.79.raw_reads.99.N1 raw_reads.79.raw_reads.99.C2 raw_reads.79.raw_reads.99.N2 raw_reads.79.raw_reads.99.C3 raw_reads.79.raw_reads.99.N3 && LAmerge -v L1.79.99 raw_reads.79.raw_reads.99.C0.S raw_reads.79.raw_reads.99.N0.S raw_reads.79.raw_reads.99.C1.S raw_reads.79.raw_reads.99.N1.S raw_reads.79.raw_reads.99.C2.S raw_reads.79.raw_reads.99.N2.S raw_reads.79.raw_reads.99.C3.S raw_reads.79.raw_reads.99.N3.S
LAsort -v raw_reads.99.raw_reads.60.C0 raw_reads.99.raw_reads.60.N0 raw_reads.99.raw_reads.60.C1 raw_reads.99.raw_reads.60.N1 raw_reads.99.raw_reads.60.C2 raw_reads.99.raw_reads.60.N2 raw_reads.99.raw_reads.60.C3 raw_reads.99.raw_reads.60.N3 && LAmerge -v L1.99.60 raw_reads.99.raw_reads.60.C0.S raw_reads.99.raw_reads.60.N0.S raw_reads.99.raw_reads.60.C1.S raw_reads.99.raw_reads.60.N1.S raw_reads.99.raw_reads.60.C2.S raw_reads.99.raw_reads.60.N2.S raw_reads.99.raw_reads.60.C3.S raw_reads.99.raw_reads.60.N3.S
LAsort -v raw_reads.99.raw_reads.61.C0 raw_reads.99.raw_reads.61.N0 raw_reads.99.raw_reads.61.C1 raw_reads.99.raw_reads.61.N1 raw_reads.99.raw_reads.61.C2 raw_reads.99.raw_reads.61.N2 raw_reads.99.raw_reads.61.C3 raw_reads.99.raw_reads.61.N3 && LAmerge -v L1.99.61 raw_reads.99.raw_reads.61.C0.S raw_reads.99.raw_reads.61.N0.S raw_reads.99.raw_reads.61.C1.S raw_reads.99.raw_reads.61.N1.S raw_reads.99.raw_reads.61.C2.S raw_reads.99.raw_reads.61.N2.S raw_reads.99.raw_reads.61.C3.S raw_reads.99.raw_reads.61.N3.S
LAsort -v raw_reads.99.raw_reads.62.C0 raw_reads.99.raw_reads.62.N0 raw_reads.99.raw_reads.62.C1 raw_reads.99.raw_reads.62.N1 raw_reads.99.raw_reads.62.C2 raw_reads.99.raw_reads.62.N2 raw_reads.99.raw_reads.62.C3 raw_reads.99.raw_reads.62.N3 && LAmerge -v L1.99.62 raw_reads.99.raw_reads.62.C0.S raw_reads.99.raw_reads.62.N0.S raw_reads.99.raw_reads.62.C1.S raw_reads.99.raw_reads.62.N1.S raw_reads.99.raw_reads.62.C2.S raw_reads.99.raw_reads.62.N2.S raw_reads.99.raw_reads.62.C3.S raw_reads.99.raw_reads.62.N3.S
LAsort -v raw_reads.99.raw_reads.63.C0 raw_reads.99.raw_reads.63.N0 raw_reads.99.raw_reads.63.C1 raw_reads.99.raw_reads.63.N1 raw_reads.99.raw_reads.63.C2 raw_reads.99.raw_reads.63.N2 raw_reads.99.raw_reads.63.C3 raw_reads.99.raw_reads.63.N3 && LAmerge -v L1.99.63 raw_reads.99.raw_reads.63.C0.S raw_reads.99.raw_reads.63.N0.S raw_reads.99.raw_reads.63.C1.S raw_reads.99.raw_reads.63.N1.S raw_reads.99.raw_reads.63.C2.S raw_reads.99.raw_reads.63.N2.S raw_reads.99.raw_reads.63.C3.S raw_reads.99.raw_reads.63.N3.S
LAsort -v raw_reads.99.raw_reads.64.C0 raw_reads.99.raw_reads.64.N0 raw_reads.99.raw_reads.64.C1 raw_reads.99.raw_reads.64.N1 raw_reads.99.raw_reads.64.C2 raw_reads.99.raw_reads.64.N2 raw_reads.99.raw_reads.64.C3 raw_reads.99.raw_reads.64.N3 && LAmerge -v L1.99.64 raw_reads.99.raw_reads.64.C0.S raw_reads.99.raw_reads.64.N0.S raw_reads.99.raw_reads.64.C1.S raw_reads.99.raw_reads.64.N1.S raw_reads.99.raw_reads.64.C2.S raw_reads.99.raw_reads.64.N2.S raw_reads.99.raw_reads.64.C3.S raw_reads.99.raw_reads.64.N3.S
LAsort -v raw_reads.99.raw_reads.65.C0 raw_reads.99.raw_reads.65.N0 raw_reads.99.raw_reads.65.C1 raw_reads.99.raw_reads.65.N1 raw_reads.99.raw_reads.65.C2 raw_reads.99.raw_reads.65.N2 raw_reads.99.raw_reads.65.C3 raw_reads.99.raw_reads.65.N3 && LAmerge -v L1.99.65 raw_reads.99.raw_reads.65.C0.S raw_reads.99.raw_reads.65.N0.S raw_reads.99.raw_reads.65.C1.S raw_reads.99.raw_reads.65.N1.S raw_reads.99.raw_reads.65.C2.S raw_reads.99.raw_reads.65.N2.S raw_reads.99.raw_reads.65.C3.S raw_reads.99.raw_reads.65.N3.S
LAsort -v raw_reads.99.raw_reads.66.C0 raw_reads.99.raw_reads.66.N0 raw_reads.99.raw_reads.66.C1 raw_reads.99.raw_reads.66.N1 raw_reads.99.raw_reads.66.C2 raw_reads.99.raw_reads.66.N2 raw_reads.99.raw_reads.66.C3 raw_reads.99.raw_reads.66.N3 && LAmerge -v L1.99.66 raw_reads.99.raw_reads.66.C0.S raw_reads.99.raw_reads.66.N0.S raw_reads.99.raw_reads.66.C1.S raw_reads.99.raw_reads.66.N1.S raw_reads.99.raw_reads.66.C2.S raw_reads.99.raw_reads.66.N2.S raw_reads.99.raw_reads.66.C3.S raw_reads.99.raw_reads.66.N3.S
LAsort -v raw_reads.99.raw_reads.67.C0 raw_reads.99.raw_reads.67.N0 raw_reads.99.raw_reads.67.C1 raw_reads.99.raw_reads.67.N1 raw_reads.99.raw_reads.67.C2 raw_reads.99.raw_reads.67.N2 raw_reads.99.raw_reads.67.C3 raw_reads.99.raw_reads.67.N3 && LAmerge -v L1.99.67 raw_reads.99.raw_reads.67.C0.S raw_reads.99.raw_reads.67.N0.S raw_reads.99.raw_reads.67.C1.S raw_reads.99.raw_reads.67.N1.S raw_reads.99.raw_reads.67.C2.S raw_reads.99.raw_reads.67.N2.S raw_reads.99.raw_reads.67.C3.S raw_reads.99.raw_reads.67.N3.S
LAsort -v raw_reads.99.raw_reads.68.C0 raw_reads.99.raw_reads.68.N0 raw_reads.99.raw_reads.68.C1 raw_reads.99.raw_reads.68.N1 raw_reads.99.raw_reads.68.C2 raw_reads.99.raw_reads.68.N2 raw_reads.99.raw_reads.68.C3 raw_reads.99.raw_reads.68.N3 && LAmerge -v L1.99.68 raw_reads.99.raw_reads.68.C0.S raw_reads.99.raw_reads.68.N0.S raw_reads.99.raw_reads.68.C1.S raw_reads.99.raw_reads.68.N1.S raw_reads.99.raw_reads.68.C2.S raw_reads.99.raw_reads.68.N2.S raw_reads.99.raw_reads.68.C3.S raw_reads.99.raw_reads.68.N3.S
LAsort -v raw_reads.99.raw_reads.69.C0 raw_reads.99.raw_reads.69.N0 raw_reads.99.raw_reads.69.C1 raw_reads.99.raw_reads.69.N1 raw_reads.99.raw_reads.69.C2 raw_reads.99.raw_reads.69.N2 raw_reads.99.raw_reads.69.C3 raw_reads.99.raw_reads.69.N3 && LAmerge -v L1.99.69 raw_reads.99.raw_reads.69.C0.S raw_reads.99.raw_reads.69.N0.S raw_reads.99.raw_reads.69.C1.S raw_reads.99.raw_reads.69.N1.S raw_reads.99.raw_reads.69.C2.S raw_reads.99.raw_reads.69.N2.S raw_reads.99.raw_reads.69.C3.S raw_reads.99.raw_reads.69.N3.S
LAsort -v raw_reads.99.raw_reads.70.C0 raw_reads.99.raw_reads.70.N0 raw_reads.99.raw_reads.70.C1 raw_reads.99.raw_reads.70.N1 raw_reads.99.raw_reads.70.C2 raw_reads.99.raw_reads.70.N2 raw_reads.99.raw_reads.70.C3 raw_reads.99.raw_reads.70.N3 && LAmerge -v L1.99.70 raw_reads.99.raw_reads.70.C0.S raw_reads.99.raw_reads.70.N0.S raw_reads.99.raw_reads.70.C1.S raw_reads.99.raw_reads.70.N1.S raw_reads.99.raw_reads.70.C2.S raw_reads.99.raw_reads.70.N2.S raw_reads.99.raw_reads.70.C3.S raw_reads.99.raw_reads.70.N3.S
LAsort -v raw_reads.99.raw_reads.71.C0 raw_reads.99.raw_reads.71.N0 raw_reads.99.raw_reads.71.C1 raw_reads.99.raw_reads.71.N1 raw_reads.99.raw_reads.71.C2 raw_reads.99.raw_reads.71.N2 raw_reads.99.raw_reads.71.C3 raw_reads.99.raw_reads.71.N3 && LAmerge -v L1.99.71 raw_reads.99.raw_reads.71.C0.S raw_reads.99.raw_reads.71.N0.S raw_reads.99.raw_reads.71.C1.S raw_reads.99.raw_reads.71.N1.S raw_reads.99.raw_reads.71.C2.S raw_reads.99.raw_reads.71.N2.S raw_reads.99.raw_reads.71.C3.S raw_reads.99.raw_reads.71.N3.S
LAsort -v raw_reads.99.raw_reads.72.C0 raw_reads.99.raw_reads.72.N0 raw_reads.99.raw_reads.72.C1 raw_reads.99.raw_reads.72.N1 raw_reads.99.raw_reads.72.C2 raw_reads.99.raw_reads.72.N2 raw_reads.99.raw_reads.72.C3 raw_reads.99.raw_reads.72.N3 && LAmerge -v L1.99.72 raw_reads.99.raw_reads.72.C0.S raw_reads.99.raw_reads.72.N0.S raw_reads.99.raw_reads.72.C1.S raw_reads.99.raw_reads.72.N1.S raw_reads.99.raw_reads.72.C2.S raw_reads.99.raw_reads.72.N2.S raw_reads.99.raw_reads.72.C3.S raw_reads.99.raw_reads.72.N3.S
LAsort -v raw_reads.99.raw_reads.73.C0 raw_reads.99.raw_reads.73.N0 raw_reads.99.raw_reads.73.C1 raw_reads.99.raw_reads.73.N1 raw_reads.99.raw_reads.73.C2 raw_reads.99.raw_reads.73.N2 raw_reads.99.raw_reads.73.C3 raw_reads.99.raw_reads.73.N3 && LAmerge -v L1.99.73 raw_reads.99.raw_reads.73.C0.S raw_reads.99.raw_reads.73.N0.S raw_reads.99.raw_reads.73.C1.S raw_reads.99.raw_reads.73.N1.S raw_reads.99.raw_reads.73.C2.S raw_reads.99.raw_reads.73.N2.S raw_reads.99.raw_reads.73.C3.S raw_reads.99.raw_reads.73.N3.S
LAsort -v raw_reads.99.raw_reads.74.C0 raw_reads.99.raw_reads.74.N0 raw_reads.99.raw_reads.74.C1 raw_reads.99.raw_reads.74.N1 raw_reads.99.raw_reads.74.C2 raw_reads.99.raw_reads.74.N2 raw_reads.99.raw_reads.74.C3 raw_reads.99.raw_reads.74.N3 && LAmerge -v L1.99.74 raw_reads.99.raw_reads.74.C0.S raw_reads.99.raw_reads.74.N0.S raw_reads.99.raw_reads.74.C1.S raw_reads.99.raw_reads.74.N1.S raw_reads.99.raw_reads.74.C2.S raw_reads.99.raw_reads.74.N2.S raw_reads.99.raw_reads.74.C3.S raw_reads.99.raw_reads.74.N3.S
LAsort -v raw_reads.99.raw_reads.75.C0 raw_reads.99.raw_reads.75.N0 raw_reads.99.raw_reads.75.C1 raw_reads.99.raw_reads.75.N1 raw_reads.99.raw_reads.75.C2 raw_reads.99.raw_reads.75.N2 raw_reads.99.raw_reads.75.C3 raw_reads.99.raw_reads.75.N3 && LAmerge -v L1.99.75 raw_reads.99.raw_reads.75.C0.S raw_reads.99.raw_reads.75.N0.S raw_reads.99.raw_reads.75.C1.S raw_reads.99.raw_reads.75.N1.S raw_reads.99.raw_reads.75.C2.S raw_reads.99.raw_reads.75.N2.S raw_reads.99.raw_reads.75.C3.S raw_reads.99.raw_reads.75.N3.S
LAsort -v raw_reads.99.raw_reads.76.C0 raw_reads.99.raw_reads.76.N0 raw_reads.99.raw_reads.76.C1 raw_reads.99.raw_reads.76.N1 raw_reads.99.raw_reads.76.C2 raw_reads.99.raw_reads.76.N2 raw_reads.99.raw_reads.76.C3 raw_reads.99.raw_reads.76.N3 && LAmerge -v L1.99.76 raw_reads.99.raw_reads.76.C0.S raw_reads.99.raw_reads.76.N0.S raw_reads.99.raw_reads.76.C1.S raw_reads.99.raw_reads.76.N1.S raw_reads.99.raw_reads.76.C2.S raw_reads.99.raw_reads.76.N2.S raw_reads.99.raw_reads.76.C3.S raw_reads.99.raw_reads.76.N3.S
LAsort -v raw_reads.99.raw_reads.77.C0 raw_reads.99.raw_reads.77.N0 raw_reads.99.raw_reads.77.C1 raw_reads.99.raw_reads.77.N1 raw_reads.99.raw_reads.77.C2 raw_reads.99.raw_reads.77.N2 raw_reads.99.raw_reads.77.C3 raw_reads.99.raw_reads.77.N3 && LAmerge -v L1.99.77 raw_reads.99.raw_reads.77.C0.S raw_reads.99.raw_reads.77.N0.S raw_reads.99.raw_reads.77.C1.S raw_reads.99.raw_reads.77.N1.S raw_reads.99.raw_reads.77.C2.S raw_reads.99.raw_reads.77.N2.S raw_reads.99.raw_reads.77.C3.S raw_reads.99.raw_reads.77.N3.S
LAsort -v raw_reads.99.raw_reads.78.C0 raw_reads.99.raw_reads.78.N0 raw_reads.99.raw_reads.78.C1 raw_reads.99.raw_reads.78.N1 raw_reads.99.raw_reads.78.C2 raw_reads.99.raw_reads.78.N2 raw_reads.99.raw_reads.78.C3 raw_reads.99.raw_reads.78.N3 && LAmerge -v L1.99.78 raw_reads.99.raw_reads.78.C0.S raw_reads.99.raw_reads.78.N0.S raw_reads.99.raw_reads.78.C1.S raw_reads.99.raw_reads.78.N1.S raw_reads.99.raw_reads.78.C2.S raw_reads.99.raw_reads.78.N2.S raw_reads.99.raw_reads.78.C3.S raw_reads.99.raw_reads.78.N3.S
LAsort -v raw_reads.99.raw_reads.79.C0 raw_reads.99.raw_reads.79.N0 raw_reads.99.raw_reads.79.C1 raw_reads.99.raw_reads.79.N1 raw_reads.99.raw_reads.79.C2 raw_reads.99.raw_reads.79.N2 raw_reads.99.raw_reads.79.C3 raw_reads.99.raw_reads.79.N3 && LAmerge -v L1.99.79 raw_reads.99.raw_reads.79.C0.S raw_reads.99.raw_reads.79.N0.S raw_reads.99.raw_reads.79.C1.S raw_reads.99.raw_reads.79.N1.S raw_reads.99.raw_reads.79.C2.S raw_reads.99.raw_reads.79.N2.S raw_reads.99.raw_reads.79.C3.S raw_reads.99.raw_reads.79.N3.S
LAcheck -vS raw_reads L1.60.99
LAcheck -vS raw_reads L1.61.99
LAcheck -vS raw_reads L1.62.99
LAcheck -vS raw_reads L1.63.99
LAcheck -vS raw_reads L1.64.99
LAcheck -vS raw_reads L1.65.99
LAcheck -vS raw_reads L1.66.99
LAcheck -vS raw_reads L1.67.99
LAcheck -vS raw_reads L1.68.99
LAcheck -vS raw_reads L1.69.99
LAcheck -vS raw_reads L1.70.99
LAcheck -vS raw_reads L1.71.99
LAcheck -vS raw_reads L1.72.99
LAcheck -vS raw_reads L1.73.99
LAcheck -vS raw_reads L1.74.99
LAcheck -vS raw_reads L1.75.99
LAcheck -vS raw_reads L1.76.99
LAcheck -vS raw_reads L1.77.99
LAcheck -vS raw_reads L1.78.99
LAcheck -vS raw_reads L1.79.99
LAcheck -vS raw_reads L1.99.60
LAcheck -vS raw_reads L1.99.61
LAcheck -vS raw_reads L1.99.62
LAcheck -vS raw_reads L1.99.63
LAcheck -vS raw_reads L1.99.64
LAcheck -vS raw_reads L1.99.65
LAcheck -vS raw_reads L1.99.66
LAcheck -vS raw_reads L1.99.67
LAcheck -vS raw_reads L1.99.68
LAcheck -vS raw_reads L1.99.69
LAcheck -vS raw_reads L1.99.70
LAcheck -vS raw_reads L1.99.71
LAcheck -vS raw_reads L1.99.72
LAcheck -vS raw_reads L1.99.73
LAcheck -vS raw_reads L1.99.74
LAcheck -vS raw_reads L1.99.75
LAcheck -vS raw_reads L1.99.76
LAcheck -vS raw_reads L1.99.77
LAcheck -vS raw_reads L1.99.78
LAcheck -vS raw_reads L1.99.79
daligner -v -k18 -h70 -t14 -H7000 -e0.75 raw_reads.40 raw_reads.21 raw_reads.22 raw_reads.23 raw_reads.24 raw_reads.25 raw_reads.26 raw_reads.27 raw_reads.28 raw_reads.29 raw_reads.30 raw_reads.31 raw_reads.32 raw_reads.33 raw_reads.34 raw_reads.35 raw_reads.36 raw_reads.37 raw_reads.38 raw_reads.39 raw_reads.40
LAcheck -v raw_reads *.las
LAsort -v raw_reads.21.raw_reads.40.C0 raw_reads.21.raw_reads.40.N0 raw_reads.21.raw_reads.40.C1 raw_reads.21.raw_reads.40.N1 raw_reads.21.raw_reads.40.C2 raw_reads.21.raw_reads.40.N2 raw_reads.21.raw_reads.40.C3 raw_reads.21.raw_reads.40.N3 && LAmerge -v L1.21.40 raw_reads.21.raw_reads.40.C0.S raw_reads.21.raw_reads.40.N0.S raw_reads.21.raw_reads.40.C1.S raw_reads.21.raw_reads.40.N1.S raw_reads.21.raw_reads.40.C2.S raw_reads.21.raw_reads.40.N2.S raw_reads.21.raw_reads.40.C3.S raw_reads.21.raw_reads.40.N3.S
LAsort -v raw_reads.22.raw_reads.40.C0 raw_reads.22.raw_reads.40.N0 raw_reads.22.raw_reads.40.C1 raw_reads.22.raw_reads.40.N1 raw_reads.22.raw_reads.40.C2 raw_reads.22.raw_reads.40.N2 raw_reads.22.raw_reads.40.C3 raw_reads.22.raw_reads.40.N3 && LAmerge -v L1.22.40 raw_reads.22.raw_reads.40.C0.S raw_reads.22.raw_reads.40.N0.S raw_reads.22.raw_reads.40.C1.S raw_reads.22.raw_reads.40.N1.S raw_reads.22.raw_reads.40.C2.S raw_reads.22.raw_reads.40.N2.S raw_reads.22.raw_reads.40.C3.S raw_reads.22.raw_reads.40.N3.S
LAsort -v raw_reads.23.raw_reads.40.C0 raw_reads.23.raw_reads.40.N0 raw_reads.23.raw_reads.40.C1 raw_reads.23.raw_reads.40.N1 raw_reads.23.raw_reads.40.C2 raw_reads.23.raw_reads.40.N2 raw_reads.23.raw_reads.40.C3 raw_reads.23.raw_reads.40.N3 && LAmerge -v L1.23.40 raw_reads.23.raw_reads.40.C0.S raw_reads.23.raw_reads.40.N0.S raw_reads.23.raw_reads.40.C1.S raw_reads.23.raw_reads.40.N1.S raw_reads.23.raw_reads.40.C2.S raw_reads.23.raw_reads.40.N2.S raw_reads.23.raw_reads.40.C3.S raw_reads.23.raw_reads.40.N3.S
LAsort -v raw_reads.24.raw_reads.40.C0 raw_reads.24.raw_reads.40.N0 raw_reads.24.raw_reads.40.C1 raw_reads.24.raw_reads.40.N1 raw_reads.24.raw_reads.40.C2 raw_reads.24.raw_reads.40.N2 raw_reads.24.raw_reads.40.C3 raw_reads.24.raw_reads.40.N3 && LAmerge -v L1.24.40 raw_reads.24.raw_reads.40.C0.S raw_reads.24.raw_reads.40.N0.S raw_reads.24.raw_reads.40.C1.S raw_reads.24.raw_reads.40.N1.S raw_reads.24.raw_reads.40.C2.S raw_reads.24.raw_reads.40.N2.S raw_reads.24.raw_reads.40.C3.S raw_reads.24.raw_reads.40.N3.S
LAsort -v raw_reads.25.raw_reads.40.C0 raw_reads.25.raw_reads.40.N0 raw_reads.25.raw_reads.40.C1 raw_reads.25.raw_reads.40.N1 raw_reads.25.raw_reads.40.C2 raw_reads.25.raw_reads.40.N2 raw_reads.25.raw_reads.40.C3 raw_reads.25.raw_reads.40.N3 && LAmerge -v L1.25.40 raw_reads.25.raw_reads.40.C0.S raw_reads.25.raw_reads.40.N0.S raw_reads.25.raw_reads.40.C1.S raw_reads.25.raw_reads.40.N1.S raw_reads.25.raw_reads.40.C2.S raw_reads.25.raw_reads.40.N2.S raw_reads.25.raw_reads.40.C3.S raw_reads.25.raw_reads.40.N3.S
LAsort -v raw_reads.26.raw_reads.40.C0 raw_reads.26.raw_reads.40.N0 raw_reads.26.raw_reads.40.C1 raw_reads.26.raw_reads.40.N1 raw_reads.26.raw_reads.40.C2 raw_reads.26.raw_reads.40.N2 raw_reads.26.raw_reads.40.C3 raw_reads.26.raw_reads.40.N3 && LAmerge -v L1.26.40 raw_reads.26.raw_reads.40.C0.S raw_reads.26.raw_reads.40.N0.S raw_reads.26.raw_reads.40.C1.S raw_reads.26.raw_reads.40.N1.S raw_reads.26.raw_reads.40.C2.S raw_reads.26.raw_reads.40.N2.S raw_reads.26.raw_reads.40.C3.S raw_reads.26.raw_reads.40.N3.S
LAsort -v raw_reads.27.raw_reads.40.C0 raw_reads.27.raw_reads.40.N0 raw_reads.27.raw_reads.40.C1 raw_reads.27.raw_reads.40.N1 raw_reads.27.raw_reads.40.C2 raw_reads.27.raw_reads.40.N2 raw_reads.27.raw_reads.40.C3 raw_reads.27.raw_reads.40.N3 && LAmerge -v L1.27.40 raw_reads.27.raw_reads.40.C0.S raw_reads.27.raw_reads.40.N0.S raw_reads.27.raw_reads.40.C1.S raw_reads.27.raw_reads.40.N1.S raw_reads.27.raw_reads.40.C2.S raw_reads.27.raw_reads.40.N2.S raw_reads.27.raw_reads.40.C3.S raw_reads.27.raw_reads.40.N3.S
LAsort -v raw_reads.28.raw_reads.40.C0 raw_reads.28.raw_reads.40.N0 raw_reads.28.raw_reads.40.C1 raw_reads.28.raw_reads.40.N1 raw_reads.28.raw_reads.40.C2 raw_reads.28.raw_reads.40.N2 raw_reads.28.raw_reads.40.C3 raw_reads.28.raw_reads.40.N3 && LAmerge -v L1.28.40 raw_reads.28.raw_reads.40.C0.S raw_reads.28.raw_reads.40.N0.S raw_reads.28.raw_reads.40.C1.S raw_reads.28.raw_reads.40.N1.S raw_reads.28.raw_reads.40.C2.S raw_reads.28.raw_reads.40.N2.S raw_reads.28.raw_reads.40.C3.S raw_reads.28.raw_reads.40.N3.S
LAsort -v raw_reads.29.raw_reads.40.C0 raw_reads.29.raw_reads.40.N0 raw_reads.29.raw_reads.40.C1 raw_reads.29.raw_reads.40.N1 raw_reads.29.raw_reads.40.C2 raw_reads.29.raw_reads.40.N2 raw_reads.29.raw_reads.40.C3 raw_reads.29.raw_reads.40.N3 && LAmerge -v L1.29.40 raw_reads.29.raw_reads.40.C0.S raw_reads.29.raw_reads.40.N0.S raw_reads.29.raw_reads.40.C1.S raw_reads.29.raw_reads.40.N1.S raw_reads.29.raw_reads.40.C2.S raw_reads.29.raw_reads.40.N2.S raw_reads.29.raw_reads.40.C3.S raw_reads.29.raw_reads.40.N3.S
LAsort -v raw_reads.30.raw_reads.40.C0 raw_reads.30.raw_reads.40.N0 raw_reads.30.raw_reads.40.C1 raw_reads.30.raw_reads.40.N1 raw_reads.30.raw_reads.40.C2 raw_reads.30.raw_reads.40.N2 raw_reads.30.raw_reads.40.C3 raw_reads.30.raw_reads.40.N3 && LAmerge -v L1.30.40 raw_reads.30.raw_reads.40.C0.S raw_reads.30.raw_reads.40.N0.S raw_reads.30.raw_reads.40.C1.S raw_reads.30.raw_reads.40.N1.S raw_reads.30.raw_reads.40.C2.S raw_reads.30.raw_reads.40.N2.S raw_reads.30.raw_reads.40.C3.S raw_reads.30.raw_reads.40.N3.S
LAsort -v raw_reads.31.raw_reads.40.C0 raw_reads.31.raw_reads.40.N0 raw_reads.31.raw_reads.40.C1 raw_reads.31.raw_reads.40.N1 raw_reads.31.raw_reads.40.C2 raw_reads.31.raw_reads.40.N2 raw_reads.31.raw_reads.40.C3 raw_reads.31.raw_reads.40.N3 && LAmerge -v L1.31.40 raw_reads.31.raw_reads.40.C0.S raw_reads.31.raw_reads.40.N0.S raw_reads.31.raw_reads.40.C1.S raw_reads.31.raw_reads.40.N1.S raw_reads.31.raw_reads.40.C2.S raw_reads.31.raw_reads.40.N2.S raw_reads.31.raw_reads.40.C3.S raw_reads.31.raw_reads.40.N3.S
LAsort -v raw_reads.32.raw_reads.40.C0 raw_reads.32.raw_reads.40.N0 raw_reads.32.raw_reads.40.C1 raw_reads.32.raw_reads.40.N1 raw_reads.32.raw_reads.40.C2 raw_reads.32.raw_reads.40.N2 raw_reads.32.raw_reads.40.C3 raw_reads.32.raw_reads.40.N3 && LAmerge -v L1.32.40 raw_reads.32.raw_reads.40.C0.S raw_reads.32.raw_reads.40.N0.S raw_reads.32.raw_reads.40.C1.S raw_reads.32.raw_reads.40.N1.S raw_reads.32.raw_reads.40.C2.S raw_reads.32.raw_reads.40.N2.S raw_reads.32.raw_reads.40.C3.S raw_reads.32.raw_reads.40.N3.S
LAsort -v raw_reads.33.raw_reads.40.C0 raw_reads.33.raw_reads.40.N0 raw_reads.33.raw_reads.40.C1 raw_reads.33.raw_reads.40.N1 raw_reads.33.raw_reads.40.C2 raw_reads.33.raw_reads.40.N2 raw_reads.33.raw_reads.40.C3 raw_reads.33.raw_reads.40.N3 && LAmerge -v L1.33.40 raw_reads.33.raw_reads.40.C0.S raw_reads.33.raw_reads.40.N0.S raw_reads.33.raw_reads.40.C1.S raw_reads.33.raw_reads.40.N1.S raw_reads.33.raw_reads.40.C2.S raw_reads.33.raw_reads.40.N2.S raw_reads.33.raw_reads.40.C3.S raw_reads.33.raw_reads.40.N3.S
LAsort -v raw_reads.34.raw_reads.40.C0 raw_reads.34.raw_reads.40.N0 raw_reads.34.raw_reads.40.C1 raw_reads.34.raw_reads.40.N1 raw_reads.34.raw_reads.40.C2 raw_reads.34.raw_reads.40.N2 raw_reads.34.raw_reads.40.C3 raw_reads.34.raw_reads.40.N3 && LAmerge -v L1.34.40 raw_reads.34.raw_reads.40.C0.S raw_reads.34.raw_reads.40.N0.S raw_reads.34.raw_reads.40.C1.S raw_reads.34.raw_reads.40.N1.S raw_reads.34.raw_reads.40.C2.S raw_reads.34.raw_reads.40.N2.S raw_reads.34.raw_reads.40.C3.S raw_reads.34.raw_reads.40.N3.S
LAsort -v raw_reads.35.raw_reads.40.C0 raw_reads.35.raw_reads.40.N0 raw_reads.35.raw_reads.40.C1 raw_reads.35.raw_reads.40.N1 raw_reads.35.raw_reads.40.C2 raw_reads.35.raw_reads.40.N2 raw_reads.35.raw_reads.40.C3 raw_reads.35.raw_reads.40.N3 && LAmerge -v L1.35.40 raw_reads.35.raw_reads.40.C0.S raw_reads.35.raw_reads.40.N0.S raw_reads.35.raw_reads.40.C1.S raw_reads.35.raw_reads.40.N1.S raw_reads.35.raw_reads.40.C2.S raw_reads.35.raw_reads.40.N2.S raw_reads.35.raw_reads.40.C3.S raw_reads.35.raw_reads.40.N3.S
LAsort -v raw_reads.36.raw_reads.40.C0 raw_reads.36.raw_reads.40.N0 raw_reads.36.raw_reads.40.C1 raw_reads.36.raw_reads.40.N1 raw_reads.36.raw_reads.40.C2 raw_reads.36.raw_reads.40.N2 raw_reads.36.raw_reads.40.C3 raw_reads.36.raw_reads.40.N3 && LAmerge -v L1.36.40 raw_reads.36.raw_reads.40.C0.S raw_reads.36.raw_reads.40.N0.S raw_reads.36.raw_reads.40.C1.S raw_reads.36.raw_reads.40.N1.S raw_reads.36.raw_reads.40.C2.S raw_reads.36.raw_reads.40.N2.S raw_reads.36.raw_reads.40.C3.S raw_reads.36.raw_reads.40.N3.S
LAsort -v raw_reads.37.raw_reads.40.C0 raw_reads.37.raw_reads.40.N0 raw_reads.37.raw_reads.40.C1 raw_reads.37.raw_reads.40.N1 raw_reads.37.raw_reads.40.C2 raw_reads.37.raw_reads.40.N2 raw_reads.37.raw_reads.40.C3 raw_reads.37.raw_reads.40.N3 && LAmerge -v L1.37.40 raw_reads.37.raw_reads.40.C0.S raw_reads.37.raw_reads.40.N0.S raw_reads.37.raw_reads.40.C1.S raw_reads.37.raw_reads.40.N1.S raw_reads.37.raw_reads.40.C2.S raw_reads.37.raw_reads.40.N2.S raw_reads.37.raw_reads.40.C3.S raw_reads.37.raw_reads.40.N3.S
LAsort -v raw_reads.38.raw_reads.40.C0 raw_reads.38.raw_reads.40.N0 raw_reads.38.raw_reads.40.C1 raw_reads.38.raw_reads.40.N1 raw_reads.38.raw_reads.40.C2 raw_reads.38.raw_reads.40.N2 raw_reads.38.raw_reads.40.C3 raw_reads.38.raw_reads.40.N3 && LAmerge -v L1.38.40 raw_reads.38.raw_reads.40.C0.S raw_reads.38.raw_reads.40.N0.S raw_reads.38.raw_reads.40.C1.S raw_reads.38.raw_reads.40.N1.S raw_reads.38.raw_reads.40.C2.S raw_reads.38.raw_reads.40.N2.S raw_reads.38.raw_reads.40.C3.S raw_reads.38.raw_reads.40.N3.S
LAsort -v raw_reads.39.raw_reads.40.C0 raw_reads.39.raw_reads.40.N0 raw_reads.39.raw_reads.40.C1 raw_reads.39.raw_reads.40.N1 raw_reads.39.raw_reads.40.C2 raw_reads.39.raw_reads.40.N2 raw_reads.39.raw_reads.40.C3 raw_reads.39.raw_reads.40.N3 && LAmerge -v L1.39.40 raw_reads.39.raw_reads.40.C0.S raw_reads.39.raw_reads.40.N0.S raw_reads.39.raw_reads.40.C1.S raw_reads.39.raw_reads.40.N1.S raw_reads.39.raw_reads.40.C2.S raw_reads.39.raw_reads.40.N2.S raw_reads.39.raw_reads.40.C3.S raw_reads.39.raw_reads.40.N3.S
LAsort -v raw_reads.40.raw_reads.21.C0 raw_reads.40.raw_reads.21.N0 raw_reads.40.raw_reads.21.C1 raw_reads.40.raw_reads.21.N1 raw_reads.40.raw_reads.21.C2 raw_reads.40.raw_reads.21.N2 raw_reads.40.raw_reads.21.C3 raw_reads.40.raw_reads.21.N3 && LAmerge -v L1.40.21 raw_reads.40.raw_reads.21.C0.S raw_reads.40.raw_reads.21.N0.S raw_reads.40.raw_reads.21.C1.S raw_reads.40.raw_reads.21.N1.S raw_reads.40.raw_reads.21.C2.S raw_reads.40.raw_reads.21.N2.S raw_reads.40.raw_reads.21.C3.S raw_reads.40.raw_reads.21.N3.S
LAsort -v raw_reads.40.raw_reads.22.C0 raw_reads.40.raw_reads.22.N0 raw_reads.40.raw_reads.22.C1 raw_reads.40.raw_reads.22.N1 raw_reads.40.raw_reads.22.C2 raw_reads.40.raw_reads.22.N2 raw_reads.40.raw_reads.22.C3 raw_reads.40.raw_reads.22.N3 && LAmerge -v L1.40.22 raw_reads.40.raw_reads.22.C0.S raw_reads.40.raw_reads.22.N0.S raw_reads.40.raw_reads.22.C1.S raw_reads.40.raw_reads.22.N1.S raw_reads.40.raw_reads.22.C2.S raw_reads.40.raw_reads.22.N2.S raw_reads.40.raw_reads.22.C3.S raw_reads.40.raw_reads.22.N3.S
LAsort -v raw_reads.40.raw_reads.23.C0 raw_reads.40.raw_reads.23.N0 raw_reads.40.raw_reads.23.C1 raw_reads.40.raw_reads.23.N1 raw_reads.40.raw_reads.23.C2 raw_reads.40.raw_reads.23.N2 raw_reads.40.raw_reads.23.C3 raw_reads.40.raw_reads.23.N3 && LAmerge -v L1.40.23 raw_reads.40.raw_reads.23.C0.S raw_reads.40.raw_reads.23.N0.S raw_reads.40.raw_reads.23.C1.S raw_reads.40.raw_reads.23.N1.S raw_reads.40.raw_reads.23.C2.S raw_reads.40.raw_reads.23.N2.S raw_reads.40.raw_reads.23.C3.S raw_reads.40.raw_reads.23.N3.S
LAsort -v raw_reads.40.raw_reads.24.C0 raw_reads.40.raw_reads.24.N0 raw_reads.40.raw_reads.24.C1 raw_reads.40.raw_reads.24.N1 raw_reads.40.raw_reads.24.C2 raw_reads.40.raw_reads.24.N2 raw_reads.40.raw_reads.24.C3 raw_reads.40.raw_reads.24.N3 && LAmerge -v L1.40.24 raw_reads.40.raw_reads.24.C0.S raw_reads.40.raw_reads.24.N0.S raw_reads.40.raw_reads.24.C1.S raw_reads.40.raw_reads.24.N1.S raw_reads.40.raw_reads.24.C2.S raw_reads.40.raw_reads.24.N2.S raw_reads.40.raw_reads.24.C3.S raw_reads.40.raw_reads.24.N3.S
LAsort -v raw_reads.40.raw_reads.25.C0 raw_reads.40.raw_reads.25.N0 raw_reads.40.raw_reads.25.C1 raw_reads.40.raw_reads.25.N1 raw_reads.40.raw_reads.25.C2 raw_reads.40.raw_reads.25.N2 raw_reads.40.raw_reads.25.C3 raw_reads.40.raw_reads.25.N3 && LAmerge -v L1.40.25 raw_reads.40.raw_reads.25.C0.S raw_reads.40.raw_reads.25.N0.S raw_reads.40.raw_reads.25.C1.S raw_reads.40.raw_reads.25.N1.S raw_reads.40.raw_reads.25.C2.S raw_reads.40.raw_reads.25.N2.S raw_reads.40.raw_reads.25.C3.S raw_reads.40.raw_reads.25.N3.S
LAsort -v raw_reads.40.raw_reads.26.C0 raw_reads.40.raw_reads.26.N0 raw_reads.40.raw_reads.26.C1 raw_reads.40.raw_reads.26.N1 raw_reads.40.raw_reads.26.C2 raw_reads.40.raw_reads.26.N2 raw_reads.40.raw_reads.26.C3 raw_reads.40.raw_reads.26.N3 && LAmerge -v L1.40.26 raw_reads.40.raw_reads.26.C0.S raw_reads.40.raw_reads.26.N0.S raw_reads.40.raw_reads.26.C1.S raw_reads.40.raw_reads.26.N1.S raw_reads.40.raw_reads.26.C2.S raw_reads.40.raw_reads.26.N2.S raw_reads.40.raw_reads.26.C3.S raw_reads.40.raw_reads.26.N3.S
LAsort -v raw_reads.40.raw_reads.27.C0 raw_reads.40.raw_reads.27.N0 raw_reads.40.raw_reads.27.C1 raw_reads.40.raw_reads.27.N1 raw_reads.40.raw_reads.27.C2 raw_reads.40.raw_reads.27.N2 raw_reads.40.raw_reads.27.C3 raw_reads.40.raw_reads.27.N3 && LAmerge -v L1.40.27 raw_reads.40.raw_reads.27.C0.S raw_reads.40.raw_reads.27.N0.S raw_reads.40.raw_reads.27.C1.S raw_reads.40.raw_reads.27.N1.S raw_reads.40.raw_reads.27.C2.S raw_reads.40.raw_reads.27.N2.S raw_reads.40.raw_reads.27.C3.S raw_reads.40.raw_reads.27.N3.S
LAsort -v raw_reads.40.raw_reads.28.C0 raw_reads.40.raw_reads.28.N0 raw_reads.40.raw_reads.28.C1 raw_reads.40.raw_reads.28.N1 raw_reads.40.raw_reads.28.C2 raw_reads.40.raw_reads.28.N2 raw_reads.40.raw_reads.28.C3 raw_reads.40.raw_reads.28.N3 && LAmerge -v L1.40.28 raw_reads.40.raw_reads.28.C0.S raw_reads.40.raw_reads.28.N0.S raw_reads.40.raw_reads.28.C1.S raw_reads.40.raw_reads.28.N1.S raw_reads.40.raw_reads.28.C2.S raw_reads.40.raw_reads.28.N2.S raw_reads.40.raw_reads.28.C3.S raw_reads.40.raw_reads.28.N3.S
LAsort -v raw_reads.40.raw_reads.29.C0 raw_reads.40.raw_reads.29.N0 raw_reads.40.raw_reads.29.C1 raw_reads.40.raw_reads.29.N1 raw_reads.40.raw_reads.29.C2 raw_reads.40.raw_reads.29.N2 raw_reads.40.raw_reads.29.C3 raw_reads.40.raw_reads.29.N3 && LAmerge -v L1.40.29 raw_reads.40.raw_reads.29.C0.S raw_reads.40.raw_reads.29.N0.S raw_reads.40.raw_reads.29.C1.S raw_reads.40.raw_reads.29.N1.S raw_reads.40.raw_reads.29.C2.S raw_reads.40.raw_reads.29.N2.S raw_reads.40.raw_reads.29.C3.S raw_reads.40.raw_reads.29.N3.S
LAsort -v raw_reads.40.raw_reads.30.C0 raw_reads.40.raw_reads.30.N0 raw_reads.40.raw_reads.30.C1 raw_reads.40.raw_reads.30.N1 raw_reads.40.raw_reads.30.C2 raw_reads.40.raw_reads.30.N2 raw_reads.40.raw_reads.30.C3 raw_reads.40.raw_reads.30.N3 && LAmerge -v L1.40.30 raw_reads.40.raw_reads.30.C0.S raw_reads.40.raw_reads.30.N0.S raw_reads.40.raw_reads.30.C1.S raw_reads.40.raw_reads.30.N1.S raw_reads.40.raw_reads.30.C2.S raw_reads.40.raw_reads.30.N2.S raw_reads.40.raw_reads.30.C3.S raw_reads.40.raw_reads.30.N3.S
LAsort -v raw_reads.40.raw_reads.31.C0 raw_reads.40.raw_reads.31.N0 raw_reads.40.raw_reads.31.C1 raw_reads.40.raw_reads.31.N1 raw_reads.40.raw_reads.31.C2 raw_reads.40.raw_reads.31.N2 raw_reads.40.raw_reads.31.C3 raw_reads.40.raw_reads.31.N3 && LAmerge -v L1.40.31 raw_reads.40.raw_reads.31.C0.S raw_reads.40.raw_reads.31.N0.S raw_reads.40.raw_reads.31.C1.S raw_reads.40.raw_reads.31.N1.S raw_reads.40.raw_reads.31.C2.S raw_reads.40.raw_reads.31.N2.S raw_reads.40.raw_reads.31.C3.S raw_reads.40.raw_reads.31.N3.S
LAsort -v raw_reads.40.raw_reads.32.C0 raw_reads.40.raw_reads.32.N0 raw_reads.40.raw_reads.32.C1 raw_reads.40.raw_reads.32.N1 raw_reads.40.raw_reads.32.C2 raw_reads.40.raw_reads.32.N2 raw_reads.40.raw_reads.32.C3 raw_reads.40.raw_reads.32.N3 && LAmerge -v L1.40.32 raw_reads.40.raw_reads.32.C0.S raw_reads.40.raw_reads.32.N0.S raw_reads.40.raw_reads.32.C1.S raw_reads.40.raw_reads.32.N1.S raw_reads.40.raw_reads.32.C2.S raw_reads.40.raw_reads.32.N2.S raw_reads.40.raw_reads.32.C3.S raw_reads.40.raw_reads.32.N3.S
LAsort -v raw_reads.40.raw_reads.33.C0 raw_reads.40.raw_reads.33.N0 raw_reads.40.raw_reads.33.C1 raw_reads.40.raw_reads.33.N1 raw_reads.40.raw_reads.33.C2 raw_reads.40.raw_reads.33.N2 raw_reads.40.raw_reads.33.C3 raw_reads.40.raw_reads.33.N3 && LAmerge -v L1.40.33 raw_reads.40.raw_reads.33.C0.S raw_reads.40.raw_reads.33.N0.S raw_reads.40.raw_reads.33.C1.S raw_reads.40.raw_reads.33.N1.S raw_reads.40.raw_reads.33.C2.S raw_reads.40.raw_reads.33.N2.S raw_reads.40.raw_reads.33.C3.S raw_reads.40.raw_reads.33.N3.S
LAsort -v raw_reads.40.raw_reads.34.C0 raw_reads.40.raw_reads.34.N0 raw_reads.40.raw_reads.34.C1 raw_reads.40.raw_reads.34.N1 raw_reads.40.raw_reads.34.C2 raw_reads.40.raw_reads.34.N2 raw_reads.40.raw_reads.34.C3 raw_reads.40.raw_reads.34.N3 && LAmerge -v L1.40.34 raw_reads.40.raw_reads.34.C0.S raw_reads.40.raw_reads.34.N0.S raw_reads.40.raw_reads.34.C1.S raw_reads.40.raw_reads.34.N1.S raw_reads.40.raw_reads.34.C2.S raw_reads.40.raw_reads.34.N2.S raw_reads.40.raw_reads.34.C3.S raw_reads.40.raw_reads.34.N3.S
LAsort -v raw_reads.40.raw_reads.35.C0 raw_reads.40.raw_reads.35.N0 raw_reads.40.raw_reads.35.C1 raw_reads.40.raw_reads.35.N1 raw_reads.40.raw_reads.35.C2 raw_reads.40.raw_reads.35.N2 raw_reads.40.raw_reads.35.C3 raw_reads.40.raw_reads.35.N3 && LAmerge -v L1.40.35 raw_reads.40.raw_reads.35.C0.S raw_reads.40.raw_reads.35.N0.S raw_reads.40.raw_reads.35.C1.S raw_reads.40.raw_reads.35.N1.S raw_reads.40.raw_reads.35.C2.S raw_reads.40.raw_reads.35.N2.S raw_reads.40.raw_reads.35.C3.S raw_reads.40.raw_reads.35.N3.S
LAsort -v raw_reads.40.raw_reads.36.C0 raw_reads.40.raw_reads.36.N0 raw_reads.40.raw_reads.36.C1 raw_reads.40.raw_reads.36.N1 raw_reads.40.raw_reads.36.C2 raw_reads.40.raw_reads.36.N2 raw_reads.40.raw_reads.36.C3 raw_reads.40.raw_reads.36.N3 && LAmerge -v L1.40.36 raw_reads.40.raw_reads.36.C0.S raw_reads.40.raw_reads.36.N0.S raw_reads.40.raw_reads.36.C1.S raw_reads.40.raw_reads.36.N1.S raw_reads.40.raw_reads.36.C2.S raw_reads.40.raw_reads.36.N2.S raw_reads.40.raw_reads.36.C3.S raw_reads.40.raw_reads.36.N3.S
LAsort -v raw_reads.40.raw_reads.37.C0 raw_reads.40.raw_reads.37.N0 raw_reads.40.raw_reads.37.C1 raw_reads.40.raw_reads.37.N1 raw_reads.40.raw_reads.37.C2 raw_reads.40.raw_reads.37.N2 raw_reads.40.raw_reads.37.C3 raw_reads.40.raw_reads.37.N3 && LAmerge -v L1.40.37 raw_reads.40.raw_reads.37.C0.S raw_reads.40.raw_reads.37.N0.S raw_reads.40.raw_reads.37.C1.S raw_reads.40.raw_reads.37.N1.S raw_reads.40.raw_reads.37.C2.S raw_reads.40.raw_reads.37.N2.S raw_reads.40.raw_reads.37.C3.S raw_reads.40.raw_reads.37.N3.S
LAsort -v raw_reads.40.raw_reads.38.C0 raw_reads.40.raw_reads.38.N0 raw_reads.40.raw_reads.38.C1 raw_reads.40.raw_reads.38.N1 raw_reads.40.raw_reads.38.C2 raw_reads.40.raw_reads.38.N2 raw_reads.40.raw_reads.38.C3 raw_reads.40.raw_reads.38.N3 && LAmerge -v L1.40.38 raw_reads.40.raw_reads.38.C0.S raw_reads.40.raw_reads.38.N0.S raw_reads.40.raw_reads.38.C1.S raw_reads.40.raw_reads.38.N1.S raw_reads.40.raw_reads.38.C2.S raw_reads.40.raw_reads.38.N2.S raw_reads.40.raw_reads.38.C3.S raw_reads.40.raw_reads.38.N3.S
LAsort -v raw_reads.40.raw_reads.39.C0 raw_reads.40.raw_reads.39.N0 raw_reads.40.raw_reads.39.C1 raw_reads.40.raw_reads.39.N1 raw_reads.40.raw_reads.39.C2 raw_reads.40.raw_reads.39.N2 raw_reads.40.raw_reads.39.C3 raw_reads.40.raw_reads.39.N3 && LAmerge -v L1.40.39 raw_reads.40.raw_reads.39.C0.S raw_reads.40.raw_reads.39.N0.S raw_reads.40.raw_reads.39.C1.S raw_reads.40.raw_reads.39.N1.S raw_reads.40.raw_reads.39.C2.S raw_reads.40.raw_reads.39.N2.S raw_reads.40.raw_reads.39.C3.S raw_reads.40.raw_reads.39.N3.S
LAsort -v raw_reads.40.raw_reads.40.C0 raw_reads.40.raw_reads.40.N0 raw_reads.40.raw_reads.40.C1 raw_reads.40.raw_reads.40.N1 raw_reads.40.raw_reads.40.C2 raw_reads.40.raw_reads.40.N2 raw_reads.40.raw_reads.40.C3 raw_reads.40.raw_reads.40.N3 && LAmerge -v L1.40.40 raw_reads.40.raw_reads.40.C0.S raw_reads.40.raw_reads.40.N0.S raw_reads.40.raw_reads.40.C1.S raw_reads.40.raw_reads.40.N1.S raw_reads.40.raw_reads.40.C2.S raw_reads.40.raw_reads.40.N2.S raw_reads.40.raw_reads.40.C3.S raw_reads.40.raw_reads.40.N3.S
LAcheck -vS raw_reads L1.21.40
LAcheck -vS raw_reads L1.22.40
LAcheck -vS raw_reads L1.23.40
LAcheck -vS raw_reads L1.24.40
LAcheck -vS raw_reads L1.25.40
LAcheck -vS raw_reads L1.26.40
LAcheck -vS raw_reads L1.27.40
LAcheck -vS raw_reads L1.28.40
LAcheck -vS raw_reads L1.29.40
LAcheck -vS raw_reads L1.30.40
LAcheck -vS raw_reads L1.31.40
LAcheck -vS raw_reads L1.32.40
LAcheck -vS raw_reads L1.33.40
LAcheck -vS raw_reads L1.34.40
LAcheck -vS raw_reads L1.35.40
LAcheck -vS raw_reads L1.36.40
LAcheck -vS raw_reads L1.37.40
LAcheck -vS raw_reads L1.38.40
LAcheck -vS raw_reads L1.39.40
LAcheck -vS raw_reads L1.40.21
LAcheck -vS raw_reads L1.40.22
LAcheck -vS raw_reads L1.40.23
LAcheck -vS raw_reads L1.40.24
LAcheck -vS raw_reads L1.40.25
LAcheck -vS raw_reads L1.40.26
LAcheck -vS raw_reads L1.40.27
LAcheck -vS raw_reads L1.40.28
LAcheck -vS raw_reads L1.40.29
LAcheck -vS raw_reads L1.40.30
LAcheck -vS raw_reads L1.40.31
LAcheck -vS raw_reads L1.40.32
LAcheck -vS raw_reads L1.40.33
LAcheck -vS raw_reads L1.40.34
LAcheck -vS raw_reads L1.40.35
LAcheck -vS raw_reads L1.40.36
LAcheck -vS raw_reads L1.40.37
LAcheck -vS raw_reads L1.40.38
LAcheck -vS raw_reads L1.40.39
LAcheck -vS raw_reads L1.40.40
daligner -v -k18 -h70 -t14 -H7000 -e0.75 raw_reads.63 raw_reads.1 raw_reads.2 raw_reads.3 raw_reads.4 raw_reads.5 raw_reads.6 raw_reads.7 raw_reads.8 raw_reads.9 raw_reads.10 raw_reads.11 raw_reads.12 raw_reads.13 raw_reads.14 raw_reads.15 raw_reads.16 raw_reads.17 raw_reads.18 raw_reads.19 raw_reads.20 raw_reads.21
LAcheck -v raw_reads *.las
LAsort -v raw_reads.1.raw_reads.63.C0 raw_reads.1.raw_reads.63.N0 raw_reads.1.raw_reads.63.C1 raw_reads.1.raw_reads.63.N1 raw_reads.1.raw_reads.63.C2 raw_reads.1.raw_reads.63.N2 raw_reads.1.raw_reads.63.C3 raw_reads.1.raw_reads.63.N3 && LAmerge -v L1.1.63 raw_reads.1.raw_reads.63.C0.S raw_reads.1.raw_reads.63.N0.S raw_reads.1.raw_reads.63.C1.S raw_reads.1.raw_reads.63.N1.S raw_reads.1.raw_reads.63.C2.S raw_reads.1.raw_reads.63.N2.S raw_reads.1.raw_reads.63.C3.S raw_reads.1.raw_reads.63.N3.S
LAsort -v raw_reads.2.raw_reads.63.C0 raw_reads.2.raw_reads.63.N0 raw_reads.2.raw_reads.63.C1 raw_reads.2.raw_reads.63.N1 raw_reads.2.raw_reads.63.C2 raw_reads.2.raw_reads.63.N2 raw_reads.2.raw_reads.63.C3 raw_reads.2.raw_reads.63.N3 && LAmerge -v L1.2.63 raw_reads.2.raw_reads.63.C0.S raw_reads.2.raw_reads.63.N0.S raw_reads.2.raw_reads.63.C1.S raw_reads.2.raw_reads.63.N1.S raw_reads.2.raw_reads.63.C2.S raw_reads.2.raw_reads.63.N2.S raw_reads.2.raw_reads.63.C3.S raw_reads.2.raw_reads.63.N3.S
LAsort -v raw_reads.3.raw_reads.63.C0 raw_reads.3.raw_reads.63.N0 raw_reads.3.raw_reads.63.C1 raw_reads.3.raw_reads.63.N1 raw_reads.3.raw_reads.63.C2 raw_reads.3.raw_reads.63.N2 raw_reads.3.raw_reads.63.C3 raw_reads.3.raw_reads.63.N3 && LAmerge -v L1.3.63 raw_reads.3.raw_reads.63.C0.S raw_reads.3.raw_reads.63.N0.S raw_reads.3.raw_reads.63.C1.S raw_reads.3.raw_reads.63.N1.S raw_reads.3.raw_reads.63.C2.S raw_reads.3.raw_reads.63.N2.S raw_reads.3.raw_reads.63.C3.S raw_reads.3.raw_reads.63.N3.S
LAsort -v raw_reads.4.raw_reads.63.C0 raw_reads.4.raw_reads.63.N0 raw_reads.4.raw_reads.63.C1 raw_reads.4.raw_reads.63.N1 raw_reads.4.raw_reads.63.C2 raw_reads.4.raw_reads.63.N2 raw_reads.4.raw_reads.63.C3 raw_reads.4.raw_reads.63.N3 && LAmerge -v L1.4.63 raw_reads.4.raw_reads.63.C0.S raw_reads.4.raw_reads.63.N0.S raw_reads.4.raw_reads.63.C1.S raw_reads.4.raw_reads.63.N1.S raw_reads.4.raw_reads.63.C2.S raw_reads.4.raw_reads.63.N2.S raw_reads.4.raw_reads.63.C3.S raw_reads.4.raw_reads.63.N3.S
LAsort -v raw_reads.5.raw_reads.63.C0 raw_reads.5.raw_reads.63.N0 raw_reads.5.raw_reads.63.C1 raw_reads.5.raw_reads.63.N1 raw_reads.5.raw_reads.63.C2 raw_reads.5.raw_reads.63.N2 raw_reads.5.raw_reads.63.C3 raw_reads.5.raw_reads.63.N3 && LAmerge -v L1.5.63 raw_reads.5.raw_reads.63.C0.S raw_reads.5.raw_reads.63.N0.S raw_reads.5.raw_reads.63.C1.S raw_reads.5.raw_reads.63.N1.S raw_reads.5.raw_reads.63.C2.S raw_reads.5.raw_reads.63.N2.S raw_reads.5.raw_reads.63.C3.S raw_reads.5.raw_reads.63.N3.S
LAsort -v raw_reads.6.raw_reads.63.C0 raw_reads.6.raw_reads.63.N0 raw_reads.6.raw_reads.63.C1 raw_reads.6.raw_reads.63.N1 raw_reads.6.raw_reads.63.C2 raw_reads.6.raw_reads.63.N2 raw_reads.6.raw_reads.63.C3 raw_reads.6.raw_reads.63.N3 && LAmerge -v L1.6.63 raw_reads.6.raw_reads.63.C0.S raw_reads.6.raw_reads.63.N0.S raw_reads.6.raw_reads.63.C1.S raw_reads.6.raw_reads.63.N1.S raw_reads.6.raw_reads.63.C2.S raw_reads.6.raw_reads.63.N2.S raw_reads.6.raw_reads.63.C3.S raw_reads.6.raw_reads.63.N3.S
LAsort -v raw_reads.7.raw_reads.63.C0 raw_reads.7.raw_reads.63.N0 raw_reads.7.raw_reads.63.C1 raw_reads.7.raw_reads.63.N1 raw_reads.7.raw_reads.63.C2 raw_reads.7.raw_reads.63.N2 raw_reads.7.raw_reads.63.C3 raw_reads.7.raw_reads.63.N3 && LAmerge -v L1.7.63 raw_reads.7.raw_reads.63.C0.S raw_reads.7.raw_reads.63.N0.S raw_reads.7.raw_reads.63.C1.S raw_reads.7.raw_reads.63.N1.S raw_reads.7.raw_reads.63.C2.S raw_reads.7.raw_reads.63.N2.S raw_reads.7.raw_reads.63.C3.S raw_reads.7.raw_reads.63.N3.S
LAsort -v raw_reads.8.raw_reads.63.C0 raw_reads.8.raw_reads.63.N0 raw_reads.8.raw_reads.63.C1 raw_reads.8.raw_reads.63.N1 raw_reads.8.raw_reads.63.C2 raw_reads.8.raw_reads.63.N2 raw_reads.8.raw_reads.63.C3 raw_reads.8.raw_reads.63.N3 && LAmerge -v L1.8.63 raw_reads.8.raw_reads.63.C0.S raw_reads.8.raw_reads.63.N0.S raw_reads.8.raw_reads.63.C1.S raw_reads.8.raw_reads.63.N1.S raw_reads.8.raw_reads.63.C2.S raw_reads.8.raw_reads.63.N2.S raw_reads.8.raw_reads.63.C3.S raw_reads.8.raw_reads.63.N3.S
LAsort -v raw_reads.9.raw_reads.63.C0 raw_reads.9.raw_reads.63.N0 raw_reads.9.raw_reads.63.C1 raw_reads.9.raw_reads.63.N1 raw_reads.9.raw_reads.63.C2 raw_reads.9.raw_reads.63.N2 raw_reads.9.raw_reads.63.C3 raw_reads.9.raw_reads.63.N3 && LAmerge -v L1.9.63 raw_reads.9.raw_reads.63.C0.S raw_reads.9.raw_reads.63.N0.S raw_reads.9.raw_reads.63.C1.S raw_reads.9.raw_reads.63.N1.S raw_reads.9.raw_reads.63.C2.S raw_reads.9.raw_reads.63.N2.S raw_reads.9.raw_reads.63.C3.S raw_reads.9.raw_reads.63.N3.S
LAsort -v raw_reads.10.raw_reads.63.C0 raw_reads.10.raw_reads.63.N0 raw_reads.10.raw_reads.63.C1 raw_reads.10.raw_reads.63.N1 raw_reads.10.raw_reads.63.C2 raw_reads.10.raw_reads.63.N2 raw_reads.10.raw_reads.63.C3 raw_reads.10.raw_reads.63.N3 && LAmerge -v L1.10.63 raw_reads.10.raw_reads.63.C0.S raw_reads.10.raw_reads.63.N0.S raw_reads.10.raw_reads.63.C1.S raw_reads.10.raw_reads.63.N1.S raw_reads.10.raw_reads.63.C2.S raw_reads.10.raw_reads.63.N2.S raw_reads.10.raw_reads.63.C3.S raw_reads.10.raw_reads.63.N3.S
LAsort -v raw_reads.11.raw_reads.63.C0 raw_reads.11.raw_reads.63.N0 raw_reads.11.raw_reads.63.C1 raw_reads.11.raw_reads.63.N1 raw_reads.11.raw_reads.63.C2 raw_reads.11.raw_reads.63.N2 raw_reads.11.raw_reads.63.C3 raw_reads.11.raw_reads.63.N3 && LAmerge -v L1.11.63 raw_reads.11.raw_reads.63.C0.S raw_reads.11.raw_reads.63.N0.S raw_reads.11.raw_reads.63.C1.S raw_reads.11.raw_reads.63.N1.S raw_reads.11.raw_reads.63.C2.S raw_reads.11.raw_reads.63.N2.S raw_reads.11.raw_reads.63.C3.S raw_reads.11.raw_reads.63.N3.S
LAsort -v raw_reads.12.raw_reads.63.C0 raw_reads.12.raw_reads.63.N0 raw_reads.12.raw_reads.63.C1 raw_reads.12.raw_reads.63.N1 raw_reads.12.raw_reads.63.C2 raw_reads.12.raw_reads.63.N2 raw_reads.12.raw_reads.63.C3 raw_reads.12.raw_reads.63.N3 && LAmerge -v L1.12.63 raw_reads.12.raw_reads.63.C0.S raw_reads.12.raw_reads.63.N0.S raw_reads.12.raw_reads.63.C1.S raw_reads.12.raw_reads.63.N1.S raw_reads.12.raw_reads.63.C2.S raw_reads.12.raw_reads.63.N2.S raw_reads.12.raw_reads.63.C3.S raw_reads.12.raw_reads.63.N3.S
LAsort -v raw_reads.13.raw_reads.63.C0 raw_reads.13.raw_reads.63.N0 raw_reads.13.raw_reads.63.C1 raw_reads.13.raw_reads.63.N1 raw_reads.13.raw_reads.63.C2 raw_reads.13.raw_reads.63.N2 raw_reads.13.raw_reads.63.C3 raw_reads.13.raw_reads.63.N3 && LAmerge -v L1.13.63 raw_reads.13.raw_reads.63.C0.S raw_reads.13.raw_reads.63.N0.S raw_reads.13.raw_reads.63.C1.S raw_reads.13.raw_reads.63.N1.S raw_reads.13.raw_reads.63.C2.S raw_reads.13.raw_reads.63.N2.S raw_reads.13.raw_reads.63.C3.S raw_reads.13.raw_reads.63.N3.S
LAsort -v raw_reads.14.raw_reads.63.C0 raw_reads.14.raw_reads.63.N0 raw_reads.14.raw_reads.63.C1 raw_reads.14.raw_reads.63.N1 raw_reads.14.raw_reads.63.C2 raw_reads.14.raw_reads.63.N2 raw_reads.14.raw_reads.63.C3 raw_reads.14.raw_reads.63.N3 && LAmerge -v L1.14.63 raw_reads.14.raw_reads.63.C0.S raw_reads.14.raw_reads.63.N0.S raw_reads.14.raw_reads.63.C1.S raw_reads.14.raw_reads.63.N1.S raw_reads.14.raw_reads.63.C2.S raw_reads.14.raw_reads.63.N2.S raw_reads.14.raw_reads.63.C3.S raw_reads.14.raw_reads.63.N3.S
LAsort -v raw_reads.15.raw_reads.63.C0 raw_reads.15.raw_reads.63.N0 raw_reads.15.raw_reads.63.C1 raw_reads.15.raw_reads.63.N1 raw_reads.15.raw_reads.63.C2 raw_reads.15.raw_reads.63.N2 raw_reads.15.raw_reads.63.C3 raw_reads.15.raw_reads.63.N3 && LAmerge -v L1.15.63 raw_reads.15.raw_reads.63.C0.S raw_reads.15.raw_reads.63.N0.S raw_reads.15.raw_reads.63.C1.S raw_reads.15.raw_reads.63.N1.S raw_reads.15.raw_reads.63.C2.S raw_reads.15.raw_reads.63.N2.S raw_reads.15.raw_reads.63.C3.S raw_reads.15.raw_reads.63.N3.S
LAsort -v raw_reads.16.raw_reads.63.C0 raw_reads.16.raw_reads.63.N0 raw_reads.16.raw_reads.63.C1 raw_reads.16.raw_reads.63.N1 raw_reads.16.raw_reads.63.C2 raw_reads.16.raw_reads.63.N2 raw_reads.16.raw_reads.63.C3 raw_reads.16.raw_reads.63.N3 && LAmerge -v L1.16.63 raw_reads.16.raw_reads.63.C0.S raw_reads.16.raw_reads.63.N0.S raw_reads.16.raw_reads.63.C1.S raw_reads.16.raw_reads.63.N1.S raw_reads.16.raw_reads.63.C2.S raw_reads.16.raw_reads.63.N2.S raw_reads.16.raw_reads.63.C3.S raw_reads.16.raw_reads.63.N3.S
LAsort -v raw_reads.17.raw_reads.63.C0 raw_reads.17.raw_reads.63.N0 raw_reads.17.raw_reads.63.C1 raw_reads.17.raw_reads.63.N1 raw_reads.17.raw_reads.63.C2 raw_reads.17.raw_reads.63.N2 raw_reads.17.raw_reads.63.C3 raw_reads.17.raw_reads.63.N3 && LAmerge -v L1.17.63 raw_reads.17.raw_reads.63.C0.S raw_reads.17.raw_reads.63.N0.S raw_reads.17.raw_reads.63.C1.S raw_reads.17.raw_reads.63.N1.S raw_reads.17.raw_reads.63.C2.S raw_reads.17.raw_reads.63.N2.S raw_reads.17.raw_reads.63.C3.S raw_reads.17.raw_reads.63.N3.S
LAsort -v raw_reads.18.raw_reads.63.C0 raw_reads.18.raw_reads.63.N0 raw_reads.18.raw_reads.63.C1 raw_reads.18.raw_reads.63.N1 raw_reads.18.raw_reads.63.C2 raw_reads.18.raw_reads.63.N2 raw_reads.18.raw_reads.63.C3 raw_reads.18.raw_reads.63.N3 && LAmerge -v L1.18.63 raw_reads.18.raw_reads.63.C0.S raw_reads.18.raw_reads.63.N0.S raw_reads.18.raw_reads.63.C1.S raw_reads.18.raw_reads.63.N1.S raw_reads.18.raw_reads.63.C2.S raw_reads.18.raw_reads.63.N2.S raw_reads.18.raw_reads.63.C3.S raw_reads.18.raw_reads.63.N3.S
LAsort -v raw_reads.19.raw_reads.63.C0 raw_reads.19.raw_reads.63.N0 raw_reads.19.raw_reads.63.C1 raw_reads.19.raw_reads.63.N1 raw_reads.19.raw_reads.63.C2 raw_reads.19.raw_reads.63.N2 raw_reads.19.raw_reads.63.C3 raw_reads.19.raw_reads.63.N3 && LAmerge -v L1.19.63 raw_reads.19.raw_reads.63.C0.S raw_reads.19.raw_reads.63.N0.S raw_reads.19.raw_reads.63.C1.S raw_reads.19.raw_reads.63.N1.S raw_reads.19.raw_reads.63.C2.S raw_reads.19.raw_reads.63.N2.S raw_reads.19.raw_reads.63.C3.S raw_reads.19.raw_reads.63.N3.S
LAsort -v raw_reads.20.raw_reads.63.C0 raw_reads.20.raw_reads.63.N0 raw_reads.20.raw_reads.63.C1 raw_reads.20.raw_reads.63.N1 raw_reads.20.raw_reads.63.C2 raw_reads.20.raw_reads.63.N2 raw_reads.20.raw_reads.63.C3 raw_reads.20.raw_reads.63.N3 && LAmerge -v L1.20.63 raw_reads.20.raw_reads.63.C0.S raw_reads.20.raw_reads.63.N0.S raw_reads.20.raw_reads.63.C1.S raw_reads.20.raw_reads.63.N1.S raw_reads.20.raw_reads.63.C2.S raw_reads.20.raw_reads.63.N2.S raw_reads.20.raw_reads.63.C3.S raw_reads.20.raw_reads.63.N3.S
LAsort -v raw_reads.21.raw_reads.63.C0 raw_reads.21.raw_reads.63.N0 raw_reads.21.raw_reads.63.C1 raw_reads.21.raw_reads.63.N1 raw_reads.21.raw_reads.63.C2 raw_reads.21.raw_reads.63.N2 raw_reads.21.raw_reads.63.C3 raw_reads.21.raw_reads.63.N3 && LAmerge -v L1.21.63 raw_reads.21.raw_reads.63.C0.S raw_reads.21.raw_reads.63.N0.S raw_reads.21.raw_reads.63.C1.S raw_reads.21.raw_reads.63.N1.S raw_reads.21.raw_reads.63.C2.S raw_reads.21.raw_reads.63.N2.S raw_reads.21.raw_reads.63.C3.S raw_reads.21.raw_reads.63.N3.S
LAsort -v raw_reads.63.raw_reads.1.C0 raw_reads.63.raw_reads.1.N0 raw_reads.63.raw_reads.1.C1 raw_reads.63.raw_reads.1.N1 raw_reads.63.raw_reads.1.C2 raw_reads.63.raw_reads.1.N2 raw_reads.63.raw_reads.1.C3 raw_reads.63.raw_reads.1.N3 && LAmerge -v L1.63.1 raw_reads.63.raw_reads.1.C0.S raw_reads.63.raw_reads.1.N0.S raw_reads.63.raw_reads.1.C1.S raw_reads.63.raw_reads.1.N1.S raw_reads.63.raw_reads.1.C2.S raw_reads.63.raw_reads.1.N2.S raw_reads.63.raw_reads.1.C3.S raw_reads.63.raw_reads.1.N3.S
LAsort -v raw_reads.63.raw_reads.2.C0 raw_reads.63.raw_reads.2.N0 raw_reads.63.raw_reads.2.C1 raw_reads.63.raw_reads.2.N1 raw_reads.63.raw_reads.2.C2 raw_reads.63.raw_reads.2.N2 raw_reads.63.raw_reads.2.C3 raw_reads.63.raw_reads.2.N3 && LAmerge -v L1.63.2 raw_reads.63.raw_reads.2.C0.S raw_reads.63.raw_reads.2.N0.S raw_reads.63.raw_reads.2.C1.S raw_reads.63.raw_reads.2.N1.S raw_reads.63.raw_reads.2.C2.S raw_reads.63.raw_reads.2.N2.S raw_reads.63.raw_reads.2.C3.S raw_reads.63.raw_reads.2.N3.S
LAsort -v raw_reads.63.raw_reads.3.C0 raw_reads.63.raw_reads.3.N0 raw_reads.63.raw_reads.3.C1 raw_reads.63.raw_reads.3.N1 raw_reads.63.raw_reads.3.C2 raw_reads.63.raw_reads.3.N2 raw_reads.63.raw_reads.3.C3 raw_reads.63.raw_reads.3.N3 && LAmerge -v L1.63.3 raw_reads.63.raw_reads.3.C0.S raw_reads.63.raw_reads.3.N0.S raw_reads.63.raw_reads.3.C1.S raw_reads.63.raw_reads.3.N1.S raw_reads.63.raw_reads.3.C2.S raw_reads.63.raw_reads.3.N2.S raw_reads.63.raw_reads.3.C3.S raw_reads.63.raw_reads.3.N3.S
LAsort -v raw_reads.63.raw_reads.4.C0 raw_reads.63.raw_reads.4.N0 raw_reads.63.raw_reads.4.C1 raw_reads.63.raw_reads.4.N1 raw_reads.63.raw_reads.4.C2 raw_reads.63.raw_reads.4.N2 raw_reads.63.raw_reads.4.C3 raw_reads.63.raw_reads.4.N3 && LAmerge -v L1.63.4 raw_reads.63.raw_reads.4.C0.S raw_reads.63.raw_reads.4.N0.S raw_reads.63.raw_reads.4.C1.S raw_reads.63.raw_reads.4.N1.S raw_reads.63.raw_reads.4.C2.S raw_reads.63.raw_reads.4.N2.S raw_reads.63.raw_reads.4.C3.S raw_reads.63.raw_reads.4.N3.S
LAsort -v raw_reads.63.raw_reads.5.C0 raw_reads.63.raw_reads.5.N0 raw_reads.63.raw_reads.5.C1 raw_reads.63.raw_reads.5.N1 raw_reads.63.raw_reads.5.C2 raw_reads.63.raw_reads.5.N2 raw_reads.63.raw_reads.5.C3 raw_reads.63.raw_reads.5.N3 && LAmerge -v L1.63.5 raw_reads.63.raw_reads.5.C0.S raw_reads.63.raw_reads.5.N0.S raw_reads.63.raw_reads.5.C1.S raw_reads.63.raw_reads.5.N1.S raw_reads.63.raw_reads.5.C2.S raw_reads.63.raw_reads.5.N2.S raw_reads.63.raw_reads.5.C3.S raw_reads.63.raw_reads.5.N3.S
LAsort -v raw_reads.63.raw_reads.6.C0 raw_reads.63.raw_reads.6.N0 raw_reads.63.raw_reads.6.C1 raw_reads.63.raw_reads.6.N1 raw_reads.63.raw_reads.6.C2 raw_reads.63.raw_reads.6.N2 raw_reads.63.raw_reads.6.C3 raw_reads.63.raw_reads.6.N3 && LAmerge -v L1.63.6 raw_reads.63.raw_reads.6.C0.S raw_reads.63.raw_reads.6.N0.S raw_reads.63.raw_reads.6.C1.S raw_reads.63.raw_reads.6.N1.S raw_reads.63.raw_reads.6.C2.S raw_reads.63.raw_reads.6.N2.S raw_reads.63.raw_reads.6.C3.S raw_reads.63.raw_reads.6.N3.S
LAsort -v raw_reads.63.raw_reads.7.C0 raw_reads.63.raw_reads.7.N0 raw_reads.63.raw_reads.7.C1 raw_reads.63.raw_reads.7.N1 raw_reads.63.raw_reads.7.C2 raw_reads.63.raw_reads.7.N2 raw_reads.63.raw_reads.7.C3 raw_reads.63.raw_reads.7.N3 && LAmerge -v L1.63.7 raw_reads.63.raw_reads.7.C0.S raw_reads.63.raw_reads.7.N0.S raw_reads.63.raw_reads.7.C1.S raw_reads.63.raw_reads.7.N1.S raw_reads.63.raw_reads.7.C2.S raw_reads.63.raw_reads.7.N2.S raw_reads.63.raw_reads.7.C3.S raw_reads.63.raw_reads.7.N3.S
LAsort -v raw_reads.63.raw_reads.8.C0 raw_reads.63.raw_reads.8.N0 raw_reads.63.raw_reads.8.C1 raw_reads.63.raw_reads.8.N1 raw_reads.63.raw_reads.8.C2 raw_reads.63.raw_reads.8.N2 raw_reads.63.raw_reads.8.C3 raw_reads.63.raw_reads.8.N3 && LAmerge -v L1.63.8 raw_reads.63.raw_reads.8.C0.S raw_reads.63.raw_reads.8.N0.S raw_reads.63.raw_reads.8.C1.S raw_reads.63.raw_reads.8.N1.S raw_reads.63.raw_reads.8.C2.S raw_reads.63.raw_reads.8.N2.S raw_reads.63.raw_reads.8.C3.S raw_reads.63.raw_reads.8.N3.S
LAsort -v raw_reads.63.raw_reads.9.C0 raw_reads.63.raw_reads.9.N0 raw_reads.63.raw_reads.9.C1 raw_reads.63.raw_reads.9.N1 raw_reads.63.raw_reads.9.C2 raw_reads.63.raw_reads.9.N2 raw_reads.63.raw_reads.9.C3 raw_reads.63.raw_reads.9.N3 && LAmerge -v L1.63.9 raw_reads.63.raw_reads.9.C0.S raw_reads.63.raw_reads.9.N0.S raw_reads.63.raw_reads.9.C1.S raw_reads.63.raw_reads.9.N1.S raw_reads.63.raw_reads.9.C2.S raw_reads.63.raw_reads.9.N2.S raw_reads.63.raw_reads.9.C3.S raw_reads.63.raw_reads.9.N3.S
LAsort -v raw_reads.63.raw_reads.10.C0 raw_reads.63.raw_reads.10.N0 raw_reads.63.raw_reads.10.C1 raw_reads.63.raw_reads.10.N1 raw_reads.63.raw_reads.10.C2 raw_reads.63.raw_reads.10.N2 raw_reads.63.raw_reads.10.C3 raw_reads.63.raw_reads.10.N3 && LAmerge -v L1.63.10 raw_reads.63.raw_reads.10.C0.S raw_reads.63.raw_reads.10.N0.S raw_reads.63.raw_reads.10.C1.S raw_reads.63.raw_reads.10.N1.S raw_reads.63.raw_reads.10.C2.S raw_reads.63.raw_reads.10.N2.S raw_reads.63.raw_reads.10.C3.S raw_reads.63.raw_reads.10.N3.S
LAsort -v raw_reads.63.raw_reads.11.C0 raw_reads.63.raw_reads.11.N0 raw_reads.63.raw_reads.11.C1 raw_reads.63.raw_reads.11.N1 raw_reads.63.raw_reads.11.C2 raw_reads.63.raw_reads.11.N2 raw_reads.63.raw_reads.11.C3 raw_reads.63.raw_reads.11.N3 && LAmerge -v L1.63.11 raw_reads.63.raw_reads.11.C0.S raw_reads.63.raw_reads.11.N0.S raw_reads.63.raw_reads.11.C1.S raw_reads.63.raw_reads.11.N1.S raw_reads.63.raw_reads.11.C2.S raw_reads.63.raw_reads.11.N2.S raw_reads.63.raw_reads.11.C3.S raw_reads.63.raw_reads.11.N3.S
LAsort -v raw_reads.63.raw_reads.12.C0 raw_reads.63.raw_reads.12.N0 raw_reads.63.raw_reads.12.C1 raw_reads.63.raw_reads.12.N1 raw_reads.63.raw_reads.12.C2 raw_reads.63.raw_reads.12.N2 raw_reads.63.raw_reads.12.C3 raw_reads.63.raw_reads.12.N3 && LAmerge -v L1.63.12 raw_reads.63.raw_reads.12.C0.S raw_reads.63.raw_reads.12.N0.S raw_reads.63.raw_reads.12.C1.S raw_reads.63.raw_reads.12.N1.S raw_reads.63.raw_reads.12.C2.S raw_reads.63.raw_reads.12.N2.S raw_reads.63.raw_reads.12.C3.S raw_reads.63.raw_reads.12.N3.S
LAsort -v raw_reads.63.raw_reads.13.C0 raw_reads.63.raw_reads.13.N0 raw_reads.63.raw_reads.13.C1 raw_reads.63.raw_reads.13.N1 raw_reads.63.raw_reads.13.C2 raw_reads.63.raw_reads.13.N2 raw_reads.63.raw_reads.13.C3 raw_reads.63.raw_reads.13.N3 && LAmerge -v L1.63.13 raw_reads.63.raw_reads.13.C0.S raw_reads.63.raw_reads.13.N0.S raw_reads.63.raw_reads.13.C1.S raw_reads.63.raw_reads.13.N1.S raw_reads.63.raw_reads.13.C2.S raw_reads.63.raw_reads.13.N2.S raw_reads.63.raw_reads.13.C3.S raw_reads.63.raw_reads.13.N3.S
LAsort -v raw_reads.63.raw_reads.14.C0 raw_reads.63.raw_reads.14.N0 raw_reads.63.raw_reads.14.C1 raw_reads.63.raw_reads.14.N1 raw_reads.63.raw_reads.14.C2 raw_reads.63.raw_reads.14.N2 raw_reads.63.raw_reads.14.C3 raw_reads.63.raw_reads.14.N3 && LAmerge -v L1.63.14 raw_reads.63.raw_reads.14.C0.S raw_reads.63.raw_reads.14.N0.S raw_reads.63.raw_reads.14.C1.S raw_reads.63.raw_reads.14.N1.S raw_reads.63.raw_reads.14.C2.S raw_reads.63.raw_reads.14.N2.S raw_reads.63.raw_reads.14.C3.S raw_reads.63.raw_reads.14.N3.S
LAsort -v raw_reads.63.raw_reads.15.C0 raw_reads.63.raw_reads.15.N0 raw_reads.63.raw_reads.15.C1 raw_reads.63.raw_reads.15.N1 raw_reads.63.raw_reads.15.C2 raw_reads.63.raw_reads.15.N2 raw_reads.63.raw_reads.15.C3 raw_reads.63.raw_reads.15.N3 && LAmerge -v L1.63.15 raw_reads.63.raw_reads.15.C0.S raw_reads.63.raw_reads.15.N0.S raw_reads.63.raw_reads.15.C1.S raw_reads.63.raw_reads.15.N1.S raw_reads.63.raw_reads.15.C2.S raw_reads.63.raw_reads.15.N2.S raw_reads.63.raw_reads.15.C3.S raw_reads.63.raw_reads.15.N3.S
LAsort -v raw_reads.63.raw_reads.16.C0 raw_reads.63.raw_reads.16.N0 raw_reads.63.raw_reads.16.C1 raw_reads.63.raw_reads.16.N1 raw_reads.63.raw_reads.16.C2 raw_reads.63.raw_reads.16.N2 raw_reads.63.raw_reads.16.C3 raw_reads.63.raw_reads.16.N3 && LAmerge -v L1.63.16 raw_reads.63.raw_reads.16.C0.S raw_reads.63.raw_reads.16.N0.S raw_reads.63.raw_reads.16.C1.S raw_reads.63.raw_reads.16.N1.S raw_reads.63.raw_reads.16.C2.S raw_reads.63.raw_reads.16.N2.S raw_reads.63.raw_reads.16.C3.S raw_reads.63.raw_reads.16.N3.S
LAsort -v raw_reads.63.raw_reads.17.C0 raw_reads.63.raw_reads.17.N0 raw_reads.63.raw_reads.17.C1 raw_reads.63.raw_reads.17.N1 raw_reads.63.raw_reads.17.C2 raw_reads.63.raw_reads.17.N2 raw_reads.63.raw_reads.17.C3 raw_reads.63.raw_reads.17.N3 && LAmerge -v L1.63.17 raw_reads.63.raw_reads.17.C0.S raw_reads.63.raw_reads.17.N0.S raw_reads.63.raw_reads.17.C1.S raw_reads.63.raw_reads.17.N1.S raw_reads.63.raw_reads.17.C2.S raw_reads.63.raw_reads.17.N2.S raw_reads.63.raw_reads.17.C3.S raw_reads.63.raw_reads.17.N3.S
LAsort -v raw_reads.63.raw_reads.18.C0 raw_reads.63.raw_reads.18.N0 raw_reads.63.raw_reads.18.C1 raw_reads.63.raw_reads.18.N1 raw_reads.63.raw_reads.18.C2 raw_reads.63.raw_reads.18.N2 raw_reads.63.raw_reads.18.C3 raw_reads.63.raw_reads.18.N3 && LAmerge -v L1.63.18 raw_reads.63.raw_reads.18.C0.S raw_reads.63.raw_reads.18.N0.S raw_reads.63.raw_reads.18.C1.S raw_reads.63.raw_reads.18.N1.S raw_reads.63.raw_reads.18.C2.S raw_reads.63.raw_reads.18.N2.S raw_reads.63.raw_reads.18.C3.S raw_reads.63.raw_reads.18.N3.S
LAsort -v raw_reads.63.raw_reads.19.C0 raw_reads.63.raw_reads.19.N0 raw_reads.63.raw_reads.19.C1 raw_reads.63.raw_reads.19.N1 raw_reads.63.raw_reads.19.C2 raw_reads.63.raw_reads.19.N2 raw_reads.63.raw_reads.19.C3 raw_reads.63.raw_reads.19.N3 && LAmerge -v L1.63.19 raw_reads.63.raw_reads.19.C0.S raw_reads.63.raw_reads.19.N0.S raw_reads.63.raw_reads.19.C1.S raw_reads.63.raw_reads.19.N1.S raw_reads.63.raw_reads.19.C2.S raw_reads.63.raw_reads.19.N2.S raw_reads.63.raw_reads.19.C3.S raw_reads.63.raw_reads.19.N3.S
LAsort -v raw_reads.63.raw_reads.20.C0 raw_reads.63.raw_reads.20.N0 raw_reads.63.raw_reads.20.C1 raw_reads.63.raw_reads.20.N1 raw_reads.63.raw_reads.20.C2 raw_reads.63.raw_reads.20.N2 raw_reads.63.raw_reads.20.C3 raw_reads.63.raw_reads.20.N3 && LAmerge -v L1.63.20 raw_reads.63.raw_reads.20.C0.S raw_reads.63.raw_reads.20.N0.S raw_reads.63.raw_reads.20.C1.S raw_reads.63.raw_reads.20.N1.S raw_reads.63.raw_reads.20.C2.S raw_reads.63.raw_reads.20.N2.S raw_reads.63.raw_reads.20.C3.S raw_reads.63.raw_reads.20.N3.S
LAsort -v raw_reads.63.raw_reads.21.C0 raw_reads.63.raw_reads.21.N0 raw_reads.63.raw_reads.21.C1 raw_reads.63.raw_reads.21.N1 raw_reads.63.raw_reads.21.C2 raw_reads.63.raw_reads.21.N2 raw_reads.63.raw_reads.21.C3 raw_reads.63.raw_reads.21.N3 && LAmerge -v L1.63.21 raw_reads.63.raw_reads.21.C0.S raw_reads.63.raw_reads.21.N0.S raw_reads.63.raw_reads.21.C1.S raw_reads.63.raw_reads.21.N1.S raw_reads.63.raw_reads.21.C2.S raw_reads.63.raw_reads.21.N2.S raw_reads.63.raw_reads.21.C3.S raw_reads.63.raw_reads.21.N3.S
LAcheck -vS raw_reads L1.1.63
LAcheck -vS raw_reads L1.2.63
LAcheck -vS raw_reads L1.3.63
LAcheck -vS raw_reads L1.4.63
LAcheck -vS raw_reads L1.5.63
LAcheck -vS raw_reads L1.6.63
LAcheck -vS raw_reads L1.7.63
LAcheck -vS raw_reads L1.8.63
LAcheck -vS raw_reads L1.9.63
LAcheck -vS raw_reads L1.10.63
LAcheck -vS raw_reads L1.11.63
LAcheck -vS raw_reads L1.12.63
LAcheck -vS raw_reads L1.13.63
LAcheck -vS raw_reads L1.14.63
LAcheck -vS raw_reads L1.15.63
LAcheck -vS raw_reads L1.16.63
LAcheck -vS raw_reads L1.17.63
LAcheck -vS raw_reads L1.18.63
LAcheck -vS raw_reads L1.19.63
LAcheck -vS raw_reads L1.20.63
LAcheck -vS raw_reads L1.21.63
LAcheck -vS raw_reads L1.63.1
LAcheck -vS raw_reads L1.63.2
LAcheck -vS raw_reads L1.63.3
LAcheck -vS raw_reads L1.63.4
LAcheck -vS raw_reads L1.63.5
LAcheck -vS raw_reads L1.63.6
LAcheck -vS raw_reads L1.63.7
LAcheck -vS raw_reads L1.63.8
LAcheck -vS raw_reads L1.63.9
LAcheck -vS raw_reads L1.63.10
LAcheck -vS raw_reads L1.63.11
LAcheck -vS raw_reads L1.63.12
LAcheck -vS raw_reads L1.63.13
LAcheck -vS raw_reads L1.63.14
LAcheck -vS raw_reads L1.63.15
LAcheck -vS raw_reads L1.63.16
LAcheck -vS raw_reads L1.63.17
LAcheck -vS raw_reads L1.63.18
LAcheck -vS raw_reads L1.63.19
LAcheck -vS raw_reads L1.63.20
LAcheck -vS raw_reads L1.63.21
daligner -v -k18 -h70 -t14 -H7000 -e0.75 raw_reads.50 raw_reads.34 raw_reads.35 raw_reads.36 raw_reads.37 raw_reads.38 raw_reads.39 raw_reads.40 raw_reads.41 raw_reads.42 raw_reads.43 raw_reads.44 raw_reads.45 raw_reads.46 raw_reads.47 raw_reads.48 raw_reads.49 raw_reads.50
LAcheck -v raw_reads *.las
LAsort -v raw_reads.34.raw_reads.50.C0 raw_reads.34.raw_reads.50.N0 raw_reads.34.raw_reads.50.C1 raw_reads.34.raw_reads.50.N1 raw_reads.34.raw_reads.50.C2 raw_reads.34.raw_reads.50.N2 raw_reads.34.raw_reads.50.C3 raw_reads.34.raw_reads.50.N3 && LAmerge -v L1.34.50 raw_reads.34.raw_reads.50.C0.S raw_reads.34.raw_reads.50.N0.S raw_reads.34.raw_reads.50.C1.S raw_reads.34.raw_reads.50.N1.S raw_reads.34.raw_reads.50.C2.S raw_reads.34.raw_reads.50.N2.S raw_reads.34.raw_reads.50.C3.S raw_reads.34.raw_reads.50.N3.S
LAsort -v raw_reads.35.raw_reads.50.C0 raw_reads.35.raw_reads.50.N0 raw_reads.35.raw_reads.50.C1 raw_reads.35.raw_reads.50.N1 raw_reads.35.raw_reads.50.C2 raw_reads.35.raw_reads.50.N2 raw_reads.35.raw_reads.50.C3 raw_reads.35.raw_reads.50.N3 && LAmerge -v L1.35.50 raw_reads.35.raw_reads.50.C0.S raw_reads.35.raw_reads.50.N0.S raw_reads.35.raw_reads.50.C1.S raw_reads.35.raw_reads.50.N1.S raw_reads.35.raw_reads.50.C2.S raw_reads.35.raw_reads.50.N2.S raw_reads.35.raw_reads.50.C3.S raw_reads.35.raw_reads.50.N3.S
LAsort -v raw_reads.36.raw_reads.50.C0 raw_reads.36.raw_reads.50.N0 raw_reads.36.raw_reads.50.C1 raw_reads.36.raw_reads.50.N1 raw_reads.36.raw_reads.50.C2 raw_reads.36.raw_reads.50.N2 raw_reads.36.raw_reads.50.C3 raw_reads.36.raw_reads.50.N3 && LAmerge -v L1.36.50 raw_reads.36.raw_reads.50.C0.S raw_reads.36.raw_reads.50.N0.S raw_reads.36.raw_reads.50.C1.S raw_reads.36.raw_reads.50.N1.S raw_reads.36.raw_reads.50.C2.S raw_reads.36.raw_reads.50.N2.S raw_reads.36.raw_reads.50.C3.S raw_reads.36.raw_reads.50.N3.S
LAsort -v raw_reads.37.raw_reads.50.C0 raw_reads.37.raw_reads.50.N0 raw_reads.37.raw_reads.50.C1 raw_reads.37.raw_reads.50.N1 raw_reads.37.raw_reads.50.C2 raw_reads.37.raw_reads.50.N2 raw_reads.37.raw_reads.50.C3 raw_reads.37.raw_reads.50.N3 && LAmerge -v L1.37.50 raw_reads.37.raw_reads.50.C0.S raw_reads.37.raw_reads.50.N0.S raw_reads.37.raw_reads.50.C1.S raw_reads.37.raw_reads.50.N1.S raw_reads.37.raw_reads.50.C2.S raw_reads.37.raw_reads.50.N2.S raw_reads.37.raw_reads.50.C3.S raw_reads.37.raw_reads.50.N3.S
LAsort -v raw_reads.38.raw_reads.50.C0 raw_reads.38.raw_reads.50.N0 raw_reads.38.raw_reads.50.C1 raw_reads.38.raw_reads.50.N1 raw_reads.38.raw_reads.50.C2 raw_reads.38.raw_reads.50.N2 raw_reads.38.raw_reads.50.C3 raw_reads.38.raw_reads.50.N3 && LAmerge -v L1.38.50 raw_reads.38.raw_reads.50.C0.S raw_reads.38.raw_reads.50.N0.S raw_reads.38.raw_reads.50.C1.S raw_reads.38.raw_reads.50.N1.S raw_reads.38.raw_reads.50.C2.S raw_reads.38.raw_reads.50.N2.S raw_reads.38.raw_reads.50.C3.S raw_reads.38.raw_reads.50.N3.S
LAsort -v raw_reads.39.raw_reads.50.C0 raw_reads.39.raw_reads.50.N0 raw_reads.39.raw_reads.50.C1 raw_reads.39.raw_reads.50.N1 raw_reads.39.raw_reads.50.C2 raw_reads.39.raw_reads.50.N2 raw_reads.39.raw_reads.50.C3 raw_reads.39.raw_reads.50.N3 && LAmerge -v L1.39.50 raw_reads.39.raw_reads.50.C0.S raw_reads.39.raw_reads.50.N0.S raw_reads.39.raw_reads.50.C1.S raw_reads.39.raw_reads.50.N1.S raw_reads.39.raw_reads.50.C2.S raw_reads.39.raw_reads.50.N2.S raw_reads.39.raw_reads.50.C3.S raw_reads.39.raw_reads.50.N3.S
LAsort -v raw_reads.40.raw_reads.50.C0 raw_reads.40.raw_reads.50.N0 raw_reads.40.raw_reads.50.C1 raw_reads.40.raw_reads.50.N1 raw_reads.40.raw_reads.50.C2 raw_reads.40.raw_reads.50.N2 raw_reads.40.raw_reads.50.C3 raw_reads.40.raw_reads.50.N3 && LAmerge -v L1.40.50 raw_reads.40.raw_reads.50.C0.S raw_reads.40.raw_reads.50.N0.S raw_reads.40.raw_reads.50.C1.S raw_reads.40.raw_reads.50.N1.S raw_reads.40.raw_reads.50.C2.S raw_reads.40.raw_reads.50.N2.S raw_reads.40.raw_reads.50.C3.S raw_reads.40.raw_reads.50.N3.S
LAsort -v raw_reads.41.raw_reads.50.C0 raw_reads.41.raw_reads.50.N0 raw_reads.41.raw_reads.50.C1 raw_reads.41.raw_reads.50.N1 raw_reads.41.raw_reads.50.C2 raw_reads.41.raw_reads.50.N2 raw_reads.41.raw_reads.50.C3 raw_reads.41.raw_reads.50.N3 && LAmerge -v L1.41.50 raw_reads.41.raw_reads.50.C0.S raw_reads.41.raw_reads.50.N0.S raw_reads.41.raw_reads.50.C1.S raw_reads.41.raw_reads.50.N1.S raw_reads.41.raw_reads.50.C2.S raw_reads.41.raw_reads.50.N2.S raw_reads.41.raw_reads.50.C3.S raw_reads.41.raw_reads.50.N3.S
LAsort -v raw_reads.42.raw_reads.50.C0 raw_reads.42.raw_reads.50.N0 raw_reads.42.raw_reads.50.C1 raw_reads.42.raw_reads.50.N1 raw_reads.42.raw_reads.50.C2 raw_reads.42.raw_reads.50.N2 raw_reads.42.raw_reads.50.C3 raw_reads.42.raw_reads.50.N3 && LAmerge -v L1.42.50 raw_reads.42.raw_reads.50.C0.S raw_reads.42.raw_reads.50.N0.S raw_reads.42.raw_reads.50.C1.S raw_reads.42.raw_reads.50.N1.S raw_reads.42.raw_reads.50.C2.S raw_reads.42.raw_reads.50.N2.S raw_reads.42.raw_reads.50.C3.S raw_reads.42.raw_reads.50.N3.S
LAsort -v raw_reads.43.raw_reads.50.C0 raw_reads.43.raw_reads.50.N0 raw_reads.43.raw_reads.50.C1 raw_reads.43.raw_reads.50.N1 raw_reads.43.raw_reads.50.C2 raw_reads.43.raw_reads.50.N2 raw_reads.43.raw_reads.50.C3 raw_reads.43.raw_reads.50.N3 && LAmerge -v L1.43.50 raw_reads.43.raw_reads.50.C0.S raw_reads.43.raw_reads.50.N0.S raw_reads.43.raw_reads.50.C1.S raw_reads.43.raw_reads.50.N1.S raw_reads.43.raw_reads.50.C2.S raw_reads.43.raw_reads.50.N2.S raw_reads.43.raw_reads.50.C3.S raw_reads.43.raw_reads.50.N3.S
LAsort -v raw_reads.44.raw_reads.50.C0 raw_reads.44.raw_reads.50.N0 raw_reads.44.raw_reads.50.C1 raw_reads.44.raw_reads.50.N1 raw_reads.44.raw_reads.50.C2 raw_reads.44.raw_reads.50.N2 raw_reads.44.raw_reads.50.C3 raw_reads.44.raw_reads.50.N3 && LAmerge -v L1.44.50 raw_reads.44.raw_reads.50.C0.S raw_reads.44.raw_reads.50.N0.S raw_reads.44.raw_reads.50.C1.S raw_reads.44.raw_reads.50.N1.S raw_reads.44.raw_reads.50.C2.S raw_reads.44.raw_reads.50.N2.S raw_reads.44.raw_reads.50.C3.S raw_reads.44.raw_reads.50.N3.S
LAsort -v raw_reads.45.raw_reads.50.C0 raw_reads.45.raw_reads.50.N0 raw_reads.45.raw_reads.50.C1 raw_reads.45.raw_reads.50.N1 raw_reads.45.raw_reads.50.C2 raw_reads.45.raw_reads.50.N2 raw_reads.45.raw_reads.50.C3 raw_reads.45.raw_reads.50.N3 && LAmerge -v L1.45.50 raw_reads.45.raw_reads.50.C0.S raw_reads.45.raw_reads.50.N0.S raw_reads.45.raw_reads.50.C1.S raw_reads.45.raw_reads.50.N1.S raw_reads.45.raw_reads.50.C2.S raw_reads.45.raw_reads.50.N2.S raw_reads.45.raw_reads.50.C3.S raw_reads.45.raw_reads.50.N3.S
LAsort -v raw_reads.46.raw_reads.50.C0 raw_reads.46.raw_reads.50.N0 raw_reads.46.raw_reads.50.C1 raw_reads.46.raw_reads.50.N1 raw_reads.46.raw_reads.50.C2 raw_reads.46.raw_reads.50.N2 raw_reads.46.raw_reads.50.C3 raw_reads.46.raw_reads.50.N3 && LAmerge -v L1.46.50 raw_reads.46.raw_reads.50.C0.S raw_reads.46.raw_reads.50.N0.S raw_reads.46.raw_reads.50.C1.S raw_reads.46.raw_reads.50.N1.S raw_reads.46.raw_reads.50.C2.S raw_reads.46.raw_reads.50.N2.S raw_reads.46.raw_reads.50.C3.S raw_reads.46.raw_reads.50.N3.S
LAsort -v raw_reads.47.raw_reads.50.C0 raw_reads.47.raw_reads.50.N0 raw_reads.47.raw_reads.50.C1 raw_reads.47.raw_reads.50.N1 raw_reads.47.raw_reads.50.C2 raw_reads.47.raw_reads.50.N2 raw_reads.47.raw_reads.50.C3 raw_reads.47.raw_reads.50.N3 && LAmerge -v L1.47.50 raw_reads.47.raw_reads.50.C0.S raw_reads.47.raw_reads.50.N0.S raw_reads.47.raw_reads.50.C1.S raw_reads.47.raw_reads.50.N1.S raw_reads.47.raw_reads.50.C2.S raw_reads.47.raw_reads.50.N2.S raw_reads.47.raw_reads.50.C3.S raw_reads.47.raw_reads.50.N3.S
LAsort -v raw_reads.48.raw_reads.50.C0 raw_reads.48.raw_reads.50.N0 raw_reads.48.raw_reads.50.C1 raw_reads.48.raw_reads.50.N1 raw_reads.48.raw_reads.50.C2 raw_reads.48.raw_reads.50.N2 raw_reads.48.raw_reads.50.C3 raw_reads.48.raw_reads.50.N3 && LAmerge -v L1.48.50 raw_reads.48.raw_reads.50.C0.S raw_reads.48.raw_reads.50.N0.S raw_reads.48.raw_reads.50.C1.S raw_reads.48.raw_reads.50.N1.S raw_reads.48.raw_reads.50.C2.S raw_reads.48.raw_reads.50.N2.S raw_reads.48.raw_reads.50.C3.S raw_reads.48.raw_reads.50.N3.S
LAsort -v raw_reads.49.raw_reads.50.C0 raw_reads.49.raw_reads.50.N0 raw_reads.49.raw_reads.50.C1 raw_reads.49.raw_reads.50.N1 raw_reads.49.raw_reads.50.C2 raw_reads.49.raw_reads.50.N2 raw_reads.49.raw_reads.50.C3 raw_reads.49.raw_reads.50.N3 && LAmerge -v L1.49.50 raw_reads.49.raw_reads.50.C0.S raw_reads.49.raw_reads.50.N0.S raw_reads.49.raw_reads.50.C1.S raw_reads.49.raw_reads.50.N1.S raw_reads.49.raw_reads.50.C2.S raw_reads.49.raw_reads.50.N2.S raw_reads.49.raw_reads.50.C3.S raw_reads.49.raw_reads.50.N3.S
LAsort -v raw_reads.50.raw_reads.34.C0 raw_reads.50.raw_reads.34.N0 raw_reads.50.raw_reads.34.C1 raw_reads.50.raw_reads.34.N1 raw_reads.50.raw_reads.34.C2 raw_reads.50.raw_reads.34.N2 raw_reads.50.raw_reads.34.C3 raw_reads.50.raw_reads.34.N3 && LAmerge -v L1.50.34 raw_reads.50.raw_reads.34.C0.S raw_reads.50.raw_reads.34.N0.S raw_reads.50.raw_reads.34.C1.S raw_reads.50.raw_reads.34.N1.S raw_reads.50.raw_reads.34.C2.S raw_reads.50.raw_reads.34.N2.S raw_reads.50.raw_reads.34.C3.S raw_reads.50.raw_reads.34.N3.S
LAsort -v raw_reads.50.raw_reads.35.C0 raw_reads.50.raw_reads.35.N0 raw_reads.50.raw_reads.35.C1 raw_reads.50.raw_reads.35.N1 raw_reads.50.raw_reads.35.C2 raw_reads.50.raw_reads.35.N2 raw_reads.50.raw_reads.35.C3 raw_reads.50.raw_reads.35.N3 && LAmerge -v L1.50.35 raw_reads.50.raw_reads.35.C0.S raw_reads.50.raw_reads.35.N0.S raw_reads.50.raw_reads.35.C1.S raw_reads.50.raw_reads.35.N1.S raw_reads.50.raw_reads.35.C2.S raw_reads.50.raw_reads.35.N2.S raw_reads.50.raw_reads.35.C3.S raw_reads.50.raw_reads.35.N3.S
LAsort -v raw_reads.50.raw_reads.36.C0 raw_reads.50.raw_reads.36.N0 raw_reads.50.raw_reads.36.C1 raw_reads.50.raw_reads.36.N1 raw_reads.50.raw_reads.36.C2 raw_reads.50.raw_reads.36.N2 raw_reads.50.raw_reads.36.C3 raw_reads.50.raw_reads.36.N3 && LAmerge -v L1.50.36 raw_reads.50.raw_reads.36.C0.S raw_reads.50.raw_reads.36.N0.S raw_reads.50.raw_reads.36.C1.S raw_reads.50.raw_reads.36.N1.S raw_reads.50.raw_reads.36.C2.S raw_reads.50.raw_reads.36.N2.S raw_reads.50.raw_reads.36.C3.S raw_reads.50.raw_reads.36.N3.S
LAsort -v raw_reads.50.raw_reads.37.C0 raw_reads.50.raw_reads.37.N0 raw_reads.50.raw_reads.37.C1 raw_reads.50.raw_reads.37.N1 raw_reads.50.raw_reads.37.C2 raw_reads.50.raw_reads.37.N2 raw_reads.50.raw_reads.37.C3 raw_reads.50.raw_reads.37.N3 && LAmerge -v L1.50.37 raw_reads.50.raw_reads.37.C0.S raw_reads.50.raw_reads.37.N0.S raw_reads.50.raw_reads.37.C1.S raw_reads.50.raw_reads.37.N1.S raw_reads.50.raw_reads.37.C2.S raw_reads.50.raw_reads.37.N2.S raw_reads.50.raw_reads.37.C3.S raw_reads.50.raw_reads.37.N3.S
LAsort -v raw_reads.50.raw_reads.38.C0 raw_reads.50.raw_reads.38.N0 raw_reads.50.raw_reads.38.C1 raw_reads.50.raw_reads.38.N1 raw_reads.50.raw_reads.38.C2 raw_reads.50.raw_reads.38.N2 raw_reads.50.raw_reads.38.C3 raw_reads.50.raw_reads.38.N3 && LAmerge -v L1.50.38 raw_reads.50.raw_reads.38.C0.S raw_reads.50.raw_reads.38.N0.S raw_reads.50.raw_reads.38.C1.S raw_reads.50.raw_reads.38.N1.S raw_reads.50.raw_reads.38.C2.S raw_reads.50.raw_reads.38.N2.S raw_reads.50.raw_reads.38.C3.S raw_reads.50.raw_reads.38.N3.S
LAsort -v raw_reads.50.raw_reads.39.C0 raw_reads.50.raw_reads.39.N0 raw_reads.50.raw_reads.39.C1 raw_reads.50.raw_reads.39.N1 raw_reads.50.raw_reads.39.C2 raw_reads.50.raw_reads.39.N2 raw_reads.50.raw_reads.39.C3 raw_reads.50.raw_reads.39.N3 && LAmerge -v L1.50.39 raw_reads.50.raw_reads.39.C0.S raw_reads.50.raw_reads.39.N0.S raw_reads.50.raw_reads.39.C1.S raw_reads.50.raw_reads.39.N1.S raw_reads.50.raw_reads.39.C2.S raw_reads.50.raw_reads.39.N2.S raw_reads.50.raw_reads.39.C3.S raw_reads.50.raw_reads.39.N3.S
LAsort -v raw_reads.50.raw_reads.40.C0 raw_reads.50.raw_reads.40.N0 raw_reads.50.raw_reads.40.C1 raw_reads.50.raw_reads.40.N1 raw_reads.50.raw_reads.40.C2 raw_reads.50.raw_reads.40.N2 raw_reads.50.raw_reads.40.C3 raw_reads.50.raw_reads.40.N3 && LAmerge -v L1.50.40 raw_reads.50.raw_reads.40.C0.S raw_reads.50.raw_reads.40.N0.S raw_reads.50.raw_reads.40.C1.S raw_reads.50.raw_reads.40.N1.S raw_reads.50.raw_reads.40.C2.S raw_reads.50.raw_reads.40.N2.S raw_reads.50.raw_reads.40.C3.S raw_reads.50.raw_reads.40.N3.S
LAsort -v raw_reads.50.raw_reads.41.C0 raw_reads.50.raw_reads.41.N0 raw_reads.50.raw_reads.41.C1 raw_reads.50.raw_reads.41.N1 raw_reads.50.raw_reads.41.C2 raw_reads.50.raw_reads.41.N2 raw_reads.50.raw_reads.41.C3 raw_reads.50.raw_reads.41.N3 && LAmerge -v L1.50.41 raw_reads.50.raw_reads.41.C0.S raw_reads.50.raw_reads.41.N0.S raw_reads.50.raw_reads.41.C1.S raw_reads.50.raw_reads.41.N1.S raw_reads.50.raw_reads.41.C2.S raw_reads.50.raw_reads.41.N2.S raw_reads.50.raw_reads.41.C3.S raw_reads.50.raw_reads.41.N3.S
LAsort -v raw_reads.50.raw_reads.42.C0 raw_reads.50.raw_reads.42.N0 raw_reads.50.raw_reads.42.C1 raw_reads.50.raw_reads.42.N1 raw_reads.50.raw_reads.42.C2 raw_reads.50.raw_reads.42.N2 raw_reads.50.raw_reads.42.C3 raw_reads.50.raw_reads.42.N3 && LAmerge -v L1.50.42 raw_reads.50.raw_reads.42.C0.S raw_reads.50.raw_reads.42.N0.S raw_reads.50.raw_reads.42.C1.S raw_reads.50.raw_reads.42.N1.S raw_reads.50.raw_reads.42.C2.S raw_reads.50.raw_reads.42.N2.S raw_reads.50.raw_reads.42.C3.S raw_reads.50.raw_reads.42.N3.S
LAsort -v raw_reads.50.raw_reads.43.C0 raw_reads.50.raw_reads.43.N0 raw_reads.50.raw_reads.43.C1 raw_reads.50.raw_reads.43.N1 raw_reads.50.raw_reads.43.C2 raw_reads.50.raw_reads.43.N2 raw_reads.50.raw_reads.43.C3 raw_reads.50.raw_reads.43.N3 && LAmerge -v L1.50.43 raw_reads.50.raw_reads.43.C0.S raw_reads.50.raw_reads.43.N0.S raw_reads.50.raw_reads.43.C1.S raw_reads.50.raw_reads.43.N1.S raw_reads.50.raw_reads.43.C2.S raw_reads.50.raw_reads.43.N2.S raw_reads.50.raw_reads.43.C3.S raw_reads.50.raw_reads.43.N3.S
LAsort -v raw_reads.50.raw_reads.44.C0 raw_reads.50.raw_reads.44.N0 raw_reads.50.raw_reads.44.C1 raw_reads.50.raw_reads.44.N1 raw_reads.50.raw_reads.44.C2 raw_reads.50.raw_reads.44.N2 raw_reads.50.raw_reads.44.C3 raw_reads.50.raw_reads.44.N3 && LAmerge -v L1.50.44 raw_reads.50.raw_reads.44.C0.S raw_reads.50.raw_reads.44.N0.S raw_reads.50.raw_reads.44.C1.S raw_reads.50.raw_reads.44.N1.S raw_reads.50.raw_reads.44.C2.S raw_reads.50.raw_reads.44.N2.S raw_reads.50.raw_reads.44.C3.S raw_reads.50.raw_reads.44.N3.S
LAsort -v raw_reads.50.raw_reads.45.C0 raw_reads.50.raw_reads.45.N0 raw_reads.50.raw_reads.45.C1 raw_reads.50.raw_reads.45.N1 raw_reads.50.raw_reads.45.C2 raw_reads.50.raw_reads.45.N2 raw_reads.50.raw_reads.45.C3 raw_reads.50.raw_reads.45.N3 && LAmerge -v L1.50.45 raw_reads.50.raw_reads.45.C0.S raw_reads.50.raw_reads.45.N0.S raw_reads.50.raw_reads.45.C1.S raw_reads.50.raw_reads.45.N1.S raw_reads.50.raw_reads.45.C2.S raw_reads.50.raw_reads.45.N2.S raw_reads.50.raw_reads.45.C3.S raw_reads.50.raw_reads.45.N3.S
LAsort -v raw_reads.50.raw_reads.46.C0 raw_reads.50.raw_reads.46.N0 raw_reads.50.raw_reads.46.C1 raw_reads.50.raw_reads.46.N1 raw_reads.50.raw_reads.46.C2 raw_reads.50.raw_reads.46.N2 raw_reads.50.raw_reads.46.C3 raw_reads.50.raw_reads.46.N3 && LAmerge -v L1.50.46 raw_reads.50.raw_reads.46.C0.S raw_reads.50.raw_reads.46.N0.S raw_reads.50.raw_reads.46.C1.S raw_reads.50.raw_reads.46.N1.S raw_reads.50.raw_reads.46.C2.S raw_reads.50.raw_reads.46.N2.S raw_reads.50.raw_reads.46.C3.S raw_reads.50.raw_reads.46.N3.S
LAsort -v raw_reads.50.raw_reads.47.C0 raw_reads.50.raw_reads.47.N0 raw_reads.50.raw_reads.47.C1 raw_reads.50.raw_reads.47.N1 raw_reads.50.raw_reads.47.C2 raw_reads.50.raw_reads.47.N2 raw_reads.50.raw_reads.47.C3 raw_reads.50.raw_reads.47.N3 && LAmerge -v L1.50.47 raw_reads.50.raw_reads.47.C0.S raw_reads.50.raw_reads.47.N0.S raw_reads.50.raw_reads.47.C1.S raw_reads.50.raw_reads.47.N1.S raw_reads.50.raw_reads.47.C2.S raw_reads.50.raw_reads.47.N2.S raw_reads.50.raw_reads.47.C3.S raw_reads.50.raw_reads.47.N3.S
LAsort -v raw_reads.50.raw_reads.48.C0 raw_reads.50.raw_reads.48.N0 raw_reads.50.raw_reads.48.C1 raw_reads.50.raw_reads.48.N1 raw_reads.50.raw_reads.48.C2 raw_reads.50.raw_reads.48.N2 raw_reads.50.raw_reads.48.C3 raw_reads.50.raw_reads.48.N3 && LAmerge -v L1.50.48 raw_reads.50.raw_reads.48.C0.S raw_reads.50.raw_reads.48.N0.S raw_reads.50.raw_reads.48.C1.S raw_reads.50.raw_reads.48.N1.S raw_reads.50.raw_reads.48.C2.S raw_reads.50.raw_reads.48.N2.S raw_reads.50.raw_reads.48.C3.S raw_reads.50.raw_reads.48.N3.S
LAsort -v raw_reads.50.raw_reads.49.C0 raw_reads.50.raw_reads.49.N0 raw_reads.50.raw_reads.49.C1 raw_reads.50.raw_reads.49.N1 raw_reads.50.raw_reads.49.C2 raw_reads.50.raw_reads.49.N2 raw_reads.50.raw_reads.49.C3 raw_reads.50.raw_reads.49.N3 && LAmerge -v L1.50.49 raw_reads.50.raw_reads.49.C0.S raw_reads.50.raw_reads.49.N0.S raw_reads.50.raw_reads.49.C1.S raw_reads.50.raw_reads.49.N1.S raw_reads.50.raw_reads.49.C2.S raw_reads.50.raw_reads.49.N2.S raw_reads.50.raw_reads.49.C3.S raw_reads.50.raw_reads.49.N3.S
LAsort -v raw_reads.50.raw_reads.50.C0 raw_reads.50.raw_reads.50.N0 raw_reads.50.raw_reads.50.C1 raw_reads.50.raw_reads.50.N1 raw_reads.50.raw_reads.50.C2 raw_reads.50.raw_reads.50.N2 raw_reads.50.raw_reads.50.C3 raw_reads.50.raw_reads.50.N3 && LAmerge -v L1.50.50 raw_reads.50.raw_reads.50.C0.S raw_reads.50.raw_reads.50.N0.S raw_reads.50.raw_reads.50.C1.S raw_reads.50.raw_reads.50.N1.S raw_reads.50.raw_reads.50.C2.S raw_reads.50.raw_reads.50.N2.S raw_reads.50.raw_reads.50.C3.S raw_reads.50.raw_reads.50.N3.S
LAcheck -vS raw_reads L1.34.50
LAcheck -vS raw_reads L1.35.50
LAcheck -vS raw_reads L1.36.50
LAcheck -vS raw_reads L1.37.50
LAcheck -vS raw_reads L1.38.50
LAcheck -vS raw_reads L1.39.50
LAcheck -vS raw_reads L1.40.50
LAcheck -vS raw_reads L1.41.50
LAcheck -vS raw_reads L1.42.50
LAcheck -vS raw_reads L1.43.50
LAcheck -vS raw_reads L1.44.50
LAcheck -vS raw_reads L1.45.50
LAcheck -vS raw_reads L1.46.50
LAcheck -vS raw_reads L1.47.50
LAcheck -vS raw_reads L1.48.50
LAcheck -vS raw_reads L1.49.50
LAcheck -vS raw_reads L1.50.34
LAcheck -vS raw_reads L1.50.35
LAcheck -vS raw_reads L1.50.36
LAcheck -vS raw_reads L1.50.37
LAcheck -vS raw_reads L1.50.38
LAcheck -vS raw_reads L1.50.39
LAcheck -vS raw_reads L1.50.40
LAcheck -vS raw_reads L1.50.41
LAcheck -vS raw_reads L1.50.42
LAcheck -vS raw_reads L1.50.43
LAcheck -vS raw_reads L1.50.44
LAcheck -vS raw_reads L1.50.45
LAcheck -vS raw_reads L1.50.46
LAcheck -vS raw_reads L1.50.47
LAcheck -vS raw_reads L1.50.48
LAcheck -vS raw_reads L1.50.49
LAcheck -vS raw_reads L1.50.50
daligner -v -k18 -h70 -t14 -H7000 -e0.75 raw_reads.73 raw_reads.55 raw_reads.56 raw_reads.57 raw_reads.58 raw_reads.59 raw_reads.60 raw_reads.61 raw_reads.62 raw_reads.63 raw_reads.64 raw_reads.65 raw_reads.66 raw_reads.67 raw_reads.68 raw_reads.69 raw_reads.70 raw_reads.71 raw_reads.72 raw_reads.73
LAcheck -v raw_reads *.las
LAsort -v raw_reads.55.raw_reads.73.C0 raw_reads.55.raw_reads.73.N0 raw_reads.55.raw_reads.73.C1 raw_reads.55.raw_reads.73.N1 raw_reads.55.raw_reads.73.C2 raw_reads.55.raw_reads.73.N2 raw_reads.55.raw_reads.73.C3 raw_reads.55.raw_reads.73.N3 && LAmerge -v L1.55.73 raw_reads.55.raw_reads.73.C0.S raw_reads.55.raw_reads.73.N0.S raw_reads.55.raw_reads.73.C1.S raw_reads.55.raw_reads.73.N1.S raw_reads.55.raw_reads.73.C2.S raw_reads.55.raw_reads.73.N2.S raw_reads.55.raw_reads.73.C3.S raw_reads.55.raw_reads.73.N3.S
LAsort -v raw_reads.56.raw_reads.73.C0 raw_reads.56.raw_reads.73.N0 raw_reads.56.raw_reads.73.C1 raw_reads.56.raw_reads.73.N1 raw_reads.56.raw_reads.73.C2 raw_reads.56.raw_reads.73.N2 raw_reads.56.raw_reads.73.C3 raw_reads.56.raw_reads.73.N3 && LAmerge -v L1.56.73 raw_reads.56.raw_reads.73.C0.S raw_reads.56.raw_reads.73.N0.S raw_reads.56.raw_reads.73.C1.S raw_reads.56.raw_reads.73.N1.S raw_reads.56.raw_reads.73.C2.S raw_reads.56.raw_reads.73.N2.S raw_reads.56.raw_reads.73.C3.S raw_reads.56.raw_reads.73.N3.S
LAsort -v raw_reads.57.raw_reads.73.C0 raw_reads.57.raw_reads.73.N0 raw_reads.57.raw_reads.73.C1 raw_reads.57.raw_reads.73.N1 raw_reads.57.raw_reads.73.C2 raw_reads.57.raw_reads.73.N2 raw_reads.57.raw_reads.73.C3 raw_reads.57.raw_reads.73.N3 && LAmerge -v L1.57.73 raw_reads.57.raw_reads.73.C0.S raw_reads.57.raw_reads.73.N0.S raw_reads.57.raw_reads.73.C1.S raw_reads.57.raw_reads.73.N1.S raw_reads.57.raw_reads.73.C2.S raw_reads.57.raw_reads.73.N2.S raw_reads.57.raw_reads.73.C3.S raw_reads.57.raw_reads.73.N3.S
LAsort -v raw_reads.58.raw_reads.73.C0 raw_reads.58.raw_reads.73.N0 raw_reads.58.raw_reads.73.C1 raw_reads.58.raw_reads.73.N1 raw_reads.58.raw_reads.73.C2 raw_reads.58.raw_reads.73.N2 raw_reads.58.raw_reads.73.C3 raw_reads.58.raw_reads.73.N3 && LAmerge -v L1.58.73 raw_reads.58.raw_reads.73.C0.S raw_reads.58.raw_reads.73.N0.S raw_reads.58.raw_reads.73.C1.S raw_reads.58.raw_reads.73.N1.S raw_reads.58.raw_reads.73.C2.S raw_reads.58.raw_reads.73.N2.S raw_reads.58.raw_reads.73.C3.S raw_reads.58.raw_reads.73.N3.S
LAsort -v raw_reads.59.raw_reads.73.C0 raw_reads.59.raw_reads.73.N0 raw_reads.59.raw_reads.73.C1 raw_reads.59.raw_reads.73.N1 raw_reads.59.raw_reads.73.C2 raw_reads.59.raw_reads.73.N2 raw_reads.59.raw_reads.73.C3 raw_reads.59.raw_reads.73.N3 && LAmerge -v L1.59.73 raw_reads.59.raw_reads.73.C0.S raw_reads.59.raw_reads.73.N0.S raw_reads.59.raw_reads.73.C1.S raw_reads.59.raw_reads.73.N1.S raw_reads.59.raw_reads.73.C2.S raw_reads.59.raw_reads.73.N2.S raw_reads.59.raw_reads.73.C3.S raw_reads.59.raw_reads.73.N3.S
LAsort -v raw_reads.60.raw_reads.73.C0 raw_reads.60.raw_reads.73.N0 raw_reads.60.raw_reads.73.C1 raw_reads.60.raw_reads.73.N1 raw_reads.60.raw_reads.73.C2 raw_reads.60.raw_reads.73.N2 raw_reads.60.raw_reads.73.C3 raw_reads.60.raw_reads.73.N3 && LAmerge -v L1.60.73 raw_reads.60.raw_reads.73.C0.S raw_reads.60.raw_reads.73.N0.S raw_reads.60.raw_reads.73.C1.S raw_reads.60.raw_reads.73.N1.S raw_reads.60.raw_reads.73.C2.S raw_reads.60.raw_reads.73.N2.S raw_reads.60.raw_reads.73.C3.S raw_reads.60.raw_reads.73.N3.S
LAsort -v raw_reads.61.raw_reads.73.C0 raw_reads.61.raw_reads.73.N0 raw_reads.61.raw_reads.73.C1 raw_reads.61.raw_reads.73.N1 raw_reads.61.raw_reads.73.C2 raw_reads.61.raw_reads.73.N2 raw_reads.61.raw_reads.73.C3 raw_reads.61.raw_reads.73.N3 && LAmerge -v L1.61.73 raw_reads.61.raw_reads.73.C0.S raw_reads.61.raw_reads.73.N0.S raw_reads.61.raw_reads.73.C1.S raw_reads.61.raw_reads.73.N1.S raw_reads.61.raw_reads.73.C2.S raw_reads.61.raw_reads.73.N2.S raw_reads.61.raw_reads.73.C3.S raw_reads.61.raw_reads.73.N3.S
LAsort -v raw_reads.62.raw_reads.73.C0 raw_reads.62.raw_reads.73.N0 raw_reads.62.raw_reads.73.C1 raw_reads.62.raw_reads.73.N1 raw_reads.62.raw_reads.73.C2 raw_reads.62.raw_reads.73.N2 raw_reads.62.raw_reads.73.C3 raw_reads.62.raw_reads.73.N3 && LAmerge -v L1.62.73 raw_reads.62.raw_reads.73.C0.S raw_reads.62.raw_reads.73.N0.S raw_reads.62.raw_reads.73.C1.S raw_reads.62.raw_reads.73.N1.S raw_reads.62.raw_reads.73.C2.S raw_reads.62.raw_reads.73.N2.S raw_reads.62.raw_reads.73.C3.S raw_reads.62.raw_reads.73.N3.S
LAsort -v raw_reads.63.raw_reads.73.C0 raw_reads.63.raw_reads.73.N0 raw_reads.63.raw_reads.73.C1 raw_reads.63.raw_reads.73.N1 raw_reads.63.raw_reads.73.C2 raw_reads.63.raw_reads.73.N2 raw_reads.63.raw_reads.73.C3 raw_reads.63.raw_reads.73.N3 && LAmerge -v L1.63.73 raw_reads.63.raw_reads.73.C0.S raw_reads.63.raw_reads.73.N0.S raw_reads.63.raw_reads.73.C1.S raw_reads.63.raw_reads.73.N1.S raw_reads.63.raw_reads.73.C2.S raw_reads.63.raw_reads.73.N2.S raw_reads.63.raw_reads.73.C3.S raw_reads.63.raw_reads.73.N3.S
LAsort -v raw_reads.64.raw_reads.73.C0 raw_reads.64.raw_reads.73.N0 raw_reads.64.raw_reads.73.C1 raw_reads.64.raw_reads.73.N1 raw_reads.64.raw_reads.73.C2 raw_reads.64.raw_reads.73.N2 raw_reads.64.raw_reads.73.C3 raw_reads.64.raw_reads.73.N3 && LAmerge -v L1.64.73 raw_reads.64.raw_reads.73.C0.S raw_reads.64.raw_reads.73.N0.S raw_reads.64.raw_reads.73.C1.S raw_reads.64.raw_reads.73.N1.S raw_reads.64.raw_reads.73.C2.S raw_reads.64.raw_reads.73.N2.S raw_reads.64.raw_reads.73.C3.S raw_reads.64.raw_reads.73.N3.S
LAsort -v raw_reads.65.raw_reads.73.C0 raw_reads.65.raw_reads.73.N0 raw_reads.65.raw_reads.73.C1 raw_reads.65.raw_reads.73.N1 raw_reads.65.raw_reads.73.C2 raw_reads.65.raw_reads.73.N2 raw_reads.65.raw_reads.73.C3 raw_reads.65.raw_reads.73.N3 && LAmerge -v L1.65.73 raw_reads.65.raw_reads.73.C0.S raw_reads.65.raw_reads.73.N0.S raw_reads.65.raw_reads.73.C1.S raw_reads.65.raw_reads.73.N1.S raw_reads.65.raw_reads.73.C2.S raw_reads.65.raw_reads.73.N2.S raw_reads.65.raw_reads.73.C3.S raw_reads.65.raw_reads.73.N3.S
LAsort -v raw_reads.66.raw_reads.73.C0 raw_reads.66.raw_reads.73.N0 raw_reads.66.raw_reads.73.C1 raw_reads.66.raw_reads.73.N1 raw_reads.66.raw_reads.73.C2 raw_reads.66.raw_reads.73.N2 raw_reads.66.raw_reads.73.C3 raw_reads.66.raw_reads.73.N3 && LAmerge -v L1.66.73 raw_reads.66.raw_reads.73.C0.S raw_reads.66.raw_reads.73.N0.S raw_reads.66.raw_reads.73.C1.S raw_reads.66.raw_reads.73.N1.S raw_reads.66.raw_reads.73.C2.S raw_reads.66.raw_reads.73.N2.S raw_reads.66.raw_reads.73.C3.S raw_reads.66.raw_reads.73.N3.S
LAsort -v raw_reads.67.raw_reads.73.C0 raw_reads.67.raw_reads.73.N0 raw_reads.67.raw_reads.73.C1 raw_reads.67.raw_reads.73.N1 raw_reads.67.raw_reads.73.C2 raw_reads.67.raw_reads.73.N2 raw_reads.67.raw_reads.73.C3 raw_reads.67.raw_reads.73.N3 && LAmerge -v L1.67.73 raw_reads.67.raw_reads.73.C0.S raw_reads.67.raw_reads.73.N0.S raw_reads.67.raw_reads.73.C1.S raw_reads.67.raw_reads.73.N1.S raw_reads.67.raw_reads.73.C2.S raw_reads.67.raw_reads.73.N2.S raw_reads.67.raw_reads.73.C3.S raw_reads.67.raw_reads.73.N3.S
LAsort -v raw_reads.68.raw_reads.73.C0 raw_reads.68.raw_reads.73.N0 raw_reads.68.raw_reads.73.C1 raw_reads.68.raw_reads.73.N1 raw_reads.68.raw_reads.73.C2 raw_reads.68.raw_reads.73.N2 raw_reads.68.raw_reads.73.C3 raw_reads.68.raw_reads.73.N3 && LAmerge -v L1.68.73 raw_reads.68.raw_reads.73.C0.S raw_reads.68.raw_reads.73.N0.S raw_reads.68.raw_reads.73.C1.S raw_reads.68.raw_reads.73.N1.S raw_reads.68.raw_reads.73.C2.S raw_reads.68.raw_reads.73.N2.S raw_reads.68.raw_reads.73.C3.S raw_reads.68.raw_reads.73.N3.S
LAsort -v raw_reads.69.raw_reads.73.C0 raw_reads.69.raw_reads.73.N0 raw_reads.69.raw_reads.73.C1 raw_reads.69.raw_reads.73.N1 raw_reads.69.raw_reads.73.C2 raw_reads.69.raw_reads.73.N2 raw_reads.69.raw_reads.73.C3 raw_reads.69.raw_reads.73.N3 && LAmerge -v L1.69.73 raw_reads.69.raw_reads.73.C0.S raw_reads.69.raw_reads.73.N0.S raw_reads.69.raw_reads.73.C1.S raw_reads.69.raw_reads.73.N1.S raw_reads.69.raw_reads.73.C2.S raw_reads.69.raw_reads.73.N2.S raw_reads.69.raw_reads.73.C3.S raw_reads.69.raw_reads.73.N3.S
LAsort -v raw_reads.70.raw_reads.73.C0 raw_reads.70.raw_reads.73.N0 raw_reads.70.raw_reads.73.C1 raw_reads.70.raw_reads.73.N1 raw_reads.70.raw_reads.73.C2 raw_reads.70.raw_reads.73.N2 raw_reads.70.raw_reads.73.C3 raw_reads.70.raw_reads.73.N3 && LAmerge -v L1.70.73 raw_reads.70.raw_reads.73.C0.S raw_reads.70.raw_reads.73.N0.S raw_reads.70.raw_reads.73.C1.S raw_reads.70.raw_reads.73.N1.S raw_reads.70.raw_reads.73.C2.S raw_reads.70.raw_reads.73.N2.S raw_reads.70.raw_reads.73.C3.S raw_reads.70.raw_reads.73.N3.S
LAsort -v raw_reads.71.raw_reads.73.C0 raw_reads.71.raw_reads.73.N0 raw_reads.71.raw_reads.73.C1 raw_reads.71.raw_reads.73.N1 raw_reads.71.raw_reads.73.C2 raw_reads.71.raw_reads.73.N2 raw_reads.71.raw_reads.73.C3 raw_reads.71.raw_reads.73.N3 && LAmerge -v L1.71.73 raw_reads.71.raw_reads.73.C0.S raw_reads.71.raw_reads.73.N0.S raw_reads.71.raw_reads.73.C1.S raw_reads.71.raw_reads.73.N1.S raw_reads.71.raw_reads.73.C2.S raw_reads.71.raw_reads.73.N2.S raw_reads.71.raw_reads.73.C3.S raw_reads.71.raw_reads.73.N3.S
LAsort -v raw_reads.72.raw_reads.73.C0 raw_reads.72.raw_reads.73.N0 raw_reads.72.raw_reads.73.C1 raw_reads.72.raw_reads.73.N1 raw_reads.72.raw_reads.73.C2 raw_reads.72.raw_reads.73.N2 raw_reads.72.raw_reads.73.C3 raw_reads.72.raw_reads.73.N3 && LAmerge -v L1.72.73 raw_reads.72.raw_reads.73.C0.S raw_reads.72.raw_reads.73.N0.S raw_reads.72.raw_reads.73.C1.S raw_reads.72.raw_reads.73.N1.S raw_reads.72.raw_reads.73.C2.S raw_reads.72.raw_reads.73.N2.S raw_reads.72.raw_reads.73.C3.S raw_reads.72.raw_reads.73.N3.S
LAsort -v raw_reads.73.raw_reads.55.C0 raw_reads.73.raw_reads.55.N0 raw_reads.73.raw_reads.55.C1 raw_reads.73.raw_reads.55.N1 raw_reads.73.raw_reads.55.C2 raw_reads.73.raw_reads.55.N2 raw_reads.73.raw_reads.55.C3 raw_reads.73.raw_reads.55.N3 && LAmerge -v L1.73.55 raw_reads.73.raw_reads.55.C0.S raw_reads.73.raw_reads.55.N0.S raw_reads.73.raw_reads.55.C1.S raw_reads.73.raw_reads.55.N1.S raw_reads.73.raw_reads.55.C2.S raw_reads.73.raw_reads.55.N2.S raw_reads.73.raw_reads.55.C3.S raw_reads.73.raw_reads.55.N3.S
LAsort -v raw_reads.73.raw_reads.56.C0 raw_reads.73.raw_reads.56.N0 raw_reads.73.raw_reads.56.C1 raw_reads.73.raw_reads.56.N1 raw_reads.73.raw_reads.56.C2 raw_reads.73.raw_reads.56.N2 raw_reads.73.raw_reads.56.C3 raw_reads.73.raw_reads.56.N3 && LAmerge -v L1.73.56 raw_reads.73.raw_reads.56.C0.S raw_reads.73.raw_reads.56.N0.S raw_reads.73.raw_reads.56.C1.S raw_reads.73.raw_reads.56.N1.S raw_reads.73.raw_reads.56.C2.S raw_reads.73.raw_reads.56.N2.S raw_reads.73.raw_reads.56.C3.S raw_reads.73.raw_reads.56.N3.S
LAsort -v raw_reads.73.raw_reads.57.C0 raw_reads.73.raw_reads.57.N0 raw_reads.73.raw_reads.57.C1 raw_reads.73.raw_reads.57.N1 raw_reads.73.raw_reads.57.C2 raw_reads.73.raw_reads.57.N2 raw_reads.73.raw_reads.57.C3 raw_reads.73.raw_reads.57.N3 && LAmerge -v L1.73.57 raw_reads.73.raw_reads.57.C0.S raw_reads.73.raw_reads.57.N0.S raw_reads.73.raw_reads.57.C1.S raw_reads.73.raw_reads.57.N1.S raw_reads.73.raw_reads.57.C2.S raw_reads.73.raw_reads.57.N2.S raw_reads.73.raw_reads.57.C3.S raw_reads.73.raw_reads.57.N3.S
LAsort -v raw_reads.73.raw_reads.58.C0 raw_reads.73.raw_reads.58.N0 raw_reads.73.raw_reads.58.C1 raw_reads.73.raw_reads.58.N1 raw_reads.73.raw_reads.58.C2 raw_reads.73.raw_reads.58.N2 raw_reads.73.raw_reads.58.C3 raw_reads.73.raw_reads.58.N3 && LAmerge -v L1.73.58 raw_reads.73.raw_reads.58.C0.S raw_reads.73.raw_reads.58.N0.S raw_reads.73.raw_reads.58.C1.S raw_reads.73.raw_reads.58.N1.S raw_reads.73.raw_reads.58.C2.S raw_reads.73.raw_reads.58.N2.S raw_reads.73.raw_reads.58.C3.S raw_reads.73.raw_reads.58.N3.S
LAsort -v raw_reads.73.raw_reads.59.C0 raw_reads.73.raw_reads.59.N0 raw_reads.73.raw_reads.59.C1 raw_reads.73.raw_reads.59.N1 raw_reads.73.raw_reads.59.C2 raw_reads.73.raw_reads.59.N2 raw_reads.73.raw_reads.59.C3 raw_reads.73.raw_reads.59.N3 && LAmerge -v L1.73.59 raw_reads.73.raw_reads.59.C0.S raw_reads.73.raw_reads.59.N0.S raw_reads.73.raw_reads.59.C1.S raw_reads.73.raw_reads.59.N1.S raw_reads.73.raw_reads.59.C2.S raw_reads.73.raw_reads.59.N2.S raw_reads.73.raw_reads.59.C3.S raw_reads.73.raw_reads.59.N3.S
LAsort -v raw_reads.73.raw_reads.60.C0 raw_reads.73.raw_reads.60.N0 raw_reads.73.raw_reads.60.C1 raw_reads.73.raw_reads.60.N1 raw_reads.73.raw_reads.60.C2 raw_reads.73.raw_reads.60.N2 raw_reads.73.raw_reads.60.C3 raw_reads.73.raw_reads.60.N3 && LAmerge -v L1.73.60 raw_reads.73.raw_reads.60.C0.S raw_reads.73.raw_reads.60.N0.S raw_reads.73.raw_reads.60.C1.S raw_reads.73.raw_reads.60.N1.S raw_reads.73.raw_reads.60.C2.S raw_reads.73.raw_reads.60.N2.S raw_reads.73.raw_reads.60.C3.S raw_reads.73.raw_reads.60.N3.S
LAsort -v raw_reads.73.raw_reads.61.C0 raw_reads.73.raw_reads.61.N0 raw_reads.73.raw_reads.61.C1 raw_reads.73.raw_reads.61.N1 raw_reads.73.raw_reads.61.C2 raw_reads.73.raw_reads.61.N2 raw_reads.73.raw_reads.61.C3 raw_reads.73.raw_reads.61.N3 && LAmerge -v L1.73.61 raw_reads.73.raw_reads.61.C0.S raw_reads.73.raw_reads.61.N0.S raw_reads.73.raw_reads.61.C1.S raw_reads.73.raw_reads.61.N1.S raw_reads.73.raw_reads.61.C2.S raw_reads.73.raw_reads.61.N2.S raw_reads.73.raw_reads.61.C3.S raw_reads.73.raw_reads.61.N3.S
LAsort -v raw_reads.73.raw_reads.62.C0 raw_reads.73.raw_reads.62.N0 raw_reads.73.raw_reads.62.C1 raw_reads.73.raw_reads.62.N1 raw_reads.73.raw_reads.62.C2 raw_reads.73.raw_reads.62.N2 raw_reads.73.raw_reads.62.C3 raw_reads.73.raw_reads.62.N3 && LAmerge -v L1.73.62 raw_reads.73.raw_reads.62.C0.S raw_reads.73.raw_reads.62.N0.S raw_reads.73.raw_reads.62.C1.S raw_reads.73.raw_reads.62.N1.S raw_reads.73.raw_reads.62.C2.S raw_reads.73.raw_reads.62.N2.S raw_reads.73.raw_reads.62.C3.S raw_reads.73.raw_reads.62.N3.S
LAsort -v raw_reads.73.raw_reads.63.C0 raw_reads.73.raw_reads.63.N0 raw_reads.73.raw_reads.63.C1 raw_reads.73.raw_reads.63.N1 raw_reads.73.raw_reads.63.C2 raw_reads.73.raw_reads.63.N2 raw_reads.73.raw_reads.63.C3 raw_reads.73.raw_reads.63.N3 && LAmerge -v L1.73.63 raw_reads.73.raw_reads.63.C0.S raw_reads.73.raw_reads.63.N0.S raw_reads.73.raw_reads.63.C1.S raw_reads.73.raw_reads.63.N1.S raw_reads.73.raw_reads.63.C2.S raw_reads.73.raw_reads.63.N2.S raw_reads.73.raw_reads.63.C3.S raw_reads.73.raw_reads.63.N3.S
LAsort -v raw_reads.73.raw_reads.64.C0 raw_reads.73.raw_reads.64.N0 raw_reads.73.raw_reads.64.C1 raw_reads.73.raw_reads.64.N1 raw_reads.73.raw_reads.64.C2 raw_reads.73.raw_reads.64.N2 raw_reads.73.raw_reads.64.C3 raw_reads.73.raw_reads.64.N3 && LAmerge -v L1.73.64 raw_reads.73.raw_reads.64.C0.S raw_reads.73.raw_reads.64.N0.S raw_reads.73.raw_reads.64.C1.S raw_reads.73.raw_reads.64.N1.S raw_reads.73.raw_reads.64.C2.S raw_reads.73.raw_reads.64.N2.S raw_reads.73.raw_reads.64.C3.S raw_reads.73.raw_reads.64.N3.S
LAsort -v raw_reads.73.raw_reads.65.C0 raw_reads.73.raw_reads.65.N0 raw_reads.73.raw_reads.65.C1 raw_reads.73.raw_reads.65.N1 raw_reads.73.raw_reads.65.C2 raw_reads.73.raw_reads.65.N2 raw_reads.73.raw_reads.65.C3 raw_reads.73.raw_reads.65.N3 && LAmerge -v L1.73.65 raw_reads.73.raw_reads.65.C0.S raw_reads.73.raw_reads.65.N0.S raw_reads.73.raw_reads.65.C1.S raw_reads.73.raw_reads.65.N1.S raw_reads.73.raw_reads.65.C2.S raw_reads.73.raw_reads.65.N2.S raw_reads.73.raw_reads.65.C3.S raw_reads.73.raw_reads.65.N3.S
LAsort -v raw_reads.73.raw_reads.66.C0 raw_reads.73.raw_reads.66.N0 raw_reads.73.raw_reads.66.C1 raw_reads.73.raw_reads.66.N1 raw_reads.73.raw_reads.66.C2 raw_reads.73.raw_reads.66.N2 raw_reads.73.raw_reads.66.C3 raw_reads.73.raw_reads.66.N3 && LAmerge -v L1.73.66 raw_reads.73.raw_reads.66.C0.S raw_reads.73.raw_reads.66.N0.S raw_reads.73.raw_reads.66.C1.S raw_reads.73.raw_reads.66.N1.S raw_reads.73.raw_reads.66.C2.S raw_reads.73.raw_reads.66.N2.S raw_reads.73.raw_reads.66.C3.S raw_reads.73.raw_reads.66.N3.S
LAsort -v raw_reads.73.raw_reads.67.C0 raw_reads.73.raw_reads.67.N0 raw_reads.73.raw_reads.67.C1 raw_reads.73.raw_reads.67.N1 raw_reads.73.raw_reads.67.C2 raw_reads.73.raw_reads.67.N2 raw_reads.73.raw_reads.67.C3 raw_reads.73.raw_reads.67.N3 && LAmerge -v L1.73.67 raw_reads.73.raw_reads.67.C0.S raw_reads.73.raw_reads.67.N0.S raw_reads.73.raw_reads.67.C1.S raw_reads.73.raw_reads.67.N1.S raw_reads.73.raw_reads.67.C2.S raw_reads.73.raw_reads.67.N2.S raw_reads.73.raw_reads.67.C3.S raw_reads.73.raw_reads.67.N3.S
LAsort -v raw_reads.73.raw_reads.68.C0 raw_reads.73.raw_reads.68.N0 raw_reads.73.raw_reads.68.C1 raw_reads.73.raw_reads.68.N1 raw_reads.73.raw_reads.68.C2 raw_reads.73.raw_reads.68.N2 raw_reads.73.raw_reads.68.C3 raw_reads.73.raw_reads.68.N3 && LAmerge -v L1.73.68 raw_reads.73.raw_reads.68.C0.S raw_reads.73.raw_reads.68.N0.S raw_reads.73.raw_reads.68.C1.S raw_reads.73.raw_reads.68.N1.S raw_reads.73.raw_reads.68.C2.S raw_reads.73.raw_reads.68.N2.S raw_reads.73.raw_reads.68.C3.S raw_reads.73.raw_reads.68.N3.S
LAsort -v raw_reads.73.raw_reads.69.C0 raw_reads.73.raw_reads.69.N0 raw_reads.73.raw_reads.69.C1 raw_reads.73.raw_reads.69.N1 raw_reads.73.raw_reads.69.C2 raw_reads.73.raw_reads.69.N2 raw_reads.73.raw_reads.69.C3 raw_reads.73.raw_reads.69.N3 && LAmerge -v L1.73.69 raw_reads.73.raw_reads.69.C0.S raw_reads.73.raw_reads.69.N0.S raw_reads.73.raw_reads.69.C1.S raw_reads.73.raw_reads.69.N1.S raw_reads.73.raw_reads.69.C2.S raw_reads.73.raw_reads.69.N2.S raw_reads.73.raw_reads.69.C3.S raw_reads.73.raw_reads.69.N3.S
LAsort -v raw_reads.73.raw_reads.70.C0 raw_reads.73.raw_reads.70.N0 raw_reads.73.raw_reads.70.C1 raw_reads.73.raw_reads.70.N1 raw_reads.73.raw_reads.70.C2 raw_reads.73.raw_reads.70.N2 raw_reads.73.raw_reads.70.C3 raw_reads.73.raw_reads.70.N3 && LAmerge -v L1.73.70 raw_reads.73.raw_reads.70.C0.S raw_reads.73.raw_reads.70.N0.S raw_reads.73.raw_reads.70.C1.S raw_reads.73.raw_reads.70.N1.S raw_reads.73.raw_reads.70.C2.S raw_reads.73.raw_reads.70.N2.S raw_reads.73.raw_reads.70.C3.S raw_reads.73.raw_reads.70.N3.S
LAsort -v raw_reads.73.raw_reads.71.C0 raw_reads.73.raw_reads.71.N0 raw_reads.73.raw_reads.71.C1 raw_reads.73.raw_reads.71.N1 raw_reads.73.raw_reads.71.C2 raw_reads.73.raw_reads.71.N2 raw_reads.73.raw_reads.71.C3 raw_reads.73.raw_reads.71.N3 && LAmerge -v L1.73.71 raw_reads.73.raw_reads.71.C0.S raw_reads.73.raw_reads.71.N0.S raw_reads.73.raw_reads.71.C1.S raw_reads.73.raw_reads.71.N1.S raw_reads.73.raw_reads.71.C2.S raw_reads.73.raw_reads.71.N2.S raw_reads.73.raw_reads.71.C3.S raw_reads.73.raw_reads.71.N3.S
LAsort -v raw_reads.73.raw_reads.72.C0 raw_reads.73.raw_reads.72.N0 raw_reads.73.raw_reads.72.C1 raw_reads.73.raw_reads.72.N1 raw_reads.73.raw_reads.72.C2 raw_reads.73.raw_reads.72.N2 raw_reads.73.raw_reads.72.C3 raw_reads.73.raw_reads.72.N3 && LAmerge -v L1.73.72 raw_reads.73.raw_reads.72.C0.S raw_reads.73.raw_reads.72.N0.S raw_reads.73.raw_reads.72.C1.S raw_reads.73.raw_reads.72.N1.S raw_reads.73.raw_reads.72.C2.S raw_reads.73.raw_reads.72.N2.S raw_reads.73.raw_reads.72.C3.S raw_reads.73.raw_reads.72.N3.S
LAsort -v raw_reads.73.raw_reads.73.C0 raw_reads.73.raw_reads.73.N0 raw_reads.73.raw_reads.73.C1 raw_reads.73.raw_reads.73.N1 raw_reads.73.raw_reads.73.C2 raw_reads.73.raw_reads.73.N2 raw_reads.73.raw_reads.73.C3 raw_reads.73.raw_reads.73.N3 && LAmerge -v L1.73.73 raw_reads.73.raw_reads.73.C0.S raw_reads.73.raw_reads.73.N0.S raw_reads.73.raw_reads.73.C1.S raw_reads.73.raw_reads.73.N1.S raw_reads.73.raw_reads.73.C2.S raw_reads.73.raw_reads.73.N2.S raw_reads.73.raw_reads.73.C3.S raw_reads.73.raw_reads.73.N3.S
LAcheck -vS raw_reads L1.55.73
LAcheck -vS raw_reads L1.56.73
LAcheck -vS raw_reads L1.57.73
LAcheck -vS raw_reads L1.58.73
LAcheck -vS raw_reads L1.59.73
LAcheck -vS raw_reads L1.60.73
LAcheck -vS raw_reads L1.61.73
LAcheck -vS raw_reads L1.62.73
LAcheck -vS raw_reads L1.63.73
LAcheck -vS raw_reads L1.64.73
LAcheck -vS raw_reads L1.65.73
LAcheck -vS raw_reads L1.66.73
LAcheck -vS raw_reads L1.67.73
LAcheck -vS raw_reads L1.68.73
LAcheck -vS raw_reads L1.69.73
LAcheck -vS raw_reads L1.70.73
LAcheck -vS raw_reads L1.71.73
LAcheck -vS raw_reads L1.72.73
LAcheck -vS raw_reads L1.73.55
LAcheck -vS raw_reads L1.73.56
LAcheck -vS raw_reads L1.73.57
LAcheck -vS raw_reads L1.73.58
LAcheck -vS raw_reads L1.73.59
LAcheck -vS raw_reads L1.73.60
LAcheck -vS raw_reads L1.73.61
LAcheck -vS raw_reads L1.73.62
LAcheck -vS raw_reads L1.73.63
LAcheck -vS raw_reads L1.73.64
LAcheck -vS raw_reads L1.73.65
LAcheck -vS raw_reads L1.73.66
LAcheck -vS raw_reads L1.73.67
LAcheck -vS raw_reads L1.73.68
LAcheck -vS raw_reads L1.73.69
LAcheck -vS raw_reads L1.73.70
LAcheck -vS raw_reads L1.73.71
LAcheck -vS raw_reads L1.73.72
LAcheck -vS raw_reads L1.73.73
